Package: units
Version: 1.0-1
Title: Measurement Units for R Vectors
Authors@R: c(person("Edzer", "Pebesma", role = c("aut", "cre"), email = "edzer.pebesma@uni-muenster.de", comment = c(ORCID = "0000-0001-8049-7069")),
    person("Thomas", "Mailund", role = "aut", email = "mailund@birc.au.dk"),
    person("Tomasz", "Kalinowski", role = "aut"),
    person("James", "Hiebert", role = "ctb"),
    person("Iñaki", "Ucar", role = "aut", email = "iucar@fedoraproject.org", comment = c(ORCID = "0000-0001-6403-5550")),
    person("Thomas Lin", "Pedersen", role = "ctb")
    )
Depends: R (>= 3.5.0)
Imports: Rcpp
LinkingTo: Rcpp (>= 1.1.0)
Suggests: NISTunits, measurements, xml2, magrittr, pillar (>= 1.3.0),
        dplyr (>= 1.0.0), vctrs (>= 0.3.1), ggplot2 (> 3.2.1), testthat
        (>= 3.0.0), vdiffr, knitr, rvest, rmarkdown
VignetteBuilder: knitr
Description: Support for measurement units in R vectors, matrices
    and arrays: automatic propagation, conversion, derivation
    and simplification of units; raising errors in case of unit
    incompatibility. Compatible with the POSIXct, Date and difftime 
    classes. Uses the UNIDATA udunits library and unit database for 
    unit compatibility checking and conversion.
    Documentation about 'units' is provided in the paper by Pebesma, Mailund &
    Hiebert (2016, <doi:10.32614/RJ-2016-061>), included in this package as a
    vignette; see 'citation("units")' for details.
SystemRequirements: udunits-2
License: GPL-2
URL: https://r-quantities.github.io/units/,
        https://github.com/r-quantities/units
BugReports: https://github.com/r-quantities/units/issues
RoxygenNote: 7.3.3
Encoding: UTF-8
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-03-09 14:34:42 UTC; edzer
Author: Edzer Pebesma [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8049-7069>),
  Thomas Mailund [aut],
  Tomasz Kalinowski [aut],
  James Hiebert [ctb],
  Iñaki Ucar [aut] (ORCID: <https://orcid.org/0000-0001-6403-5550>),
  Thomas Lin Pedersen [ctb]
Maintainer: Edzer Pebesma <edzer.pebesma@uni-muenster.de>
Repository: CRAN
Date/Publication: 2026-03-11 11:20:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 08:46:52 UTC; unix
Archs: units.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",mixed_units)
S3method("[",units)
S3method("[<-",units)
S3method("[[",units)
S3method("[[<-",units)
S3method("units<-",logical)
S3method("units<-",mixed_units)
S3method("units<-",numeric)
S3method("units<-",units)
S3method(Math,units)
S3method(Ops,mixed_units)
S3method(Ops,symbolic_units)
S3method(Ops,units)
S3method(Summary,units)
S3method(all.equal,units)
S3method(anyDuplicated,units)
S3method(as.Date,units)
S3method(as.POSIXct,units)
S3method(as.character,mixed_symbolic_units)
S3method(as.character,symbolic_units)
S3method(as.data.frame,mixed_units)
S3method(as.data.frame,units)
S3method(as.list,units)
S3method(as_units,Date)
S3method(as_units,POSIXt)
S3method(as_units,call)
S3method(as_units,character)
S3method(as_units,default)
S3method(as_units,difftime)
S3method(as_units,expression)
S3method(as_units,mixed_units)
S3method(as_units,name)
S3method(as_units,symbolic_units)
S3method(as_units,units)
S3method(boxplot,units)
S3method(c,mixed_units)
S3method(c,units)
S3method(cbind,units)
S3method(diff,units)
S3method(drop_units,data.frame)
S3method(drop_units,mixed_units)
S3method(drop_units,units)
S3method(duplicated,units)
S3method(format,mixed_units)
S3method(format,units)
S3method(hist,units)
S3method(log10,units)
S3method(log2,units)
S3method(mean,units)
S3method(median,units)
S3method(mixed_units,numeric)
S3method(mixed_units,units)
S3method(plot,units)
S3method(print,mixed_units)
S3method(print,units)
S3method(quantile,units)
S3method(rbind,units)
S3method(rep,units)
S3method(seq,units)
S3method(set_units,logical)
S3method(set_units,mixed_units)
S3method(set_units,numeric)
S3method(set_units,units)
S3method(str,mixed_units)
S3method(str,units)
S3method(summary,units)
S3method(unique,mixed_units)
S3method(unique,units)
S3method(units,mixed_units)
S3method(units,symbolic_units)
S3method(units,units)
S3method(weighted.mean,units)
export(as_difftime)
export(as_units)
export(convert_to_base)
export(deparse_unit)
export(drop_units)
export(install_unit)
export(keep_units)
export(load_units_xml)
export(make_unit_label)
export(make_units)
export(mixed_units)
export(remove_unit)
export(scale_alpha_units)
export(scale_color_units)
export(scale_colour_units)
export(scale_fill_units)
export(scale_linewidth_units)
export(scale_radius_units)
export(scale_size_area_units)
export(scale_size_units)
export(scale_x_units)
export(scale_y_units)
export(set_units)
export(ud_are_convertible)
export(ud_convert)
export(unitless)
export(units_options)
export(valid_udunits)
export(valid_udunits_prefixes)
import(graphics)
import(stats)
import(utils)
importFrom(Rcpp,evalCpp)
useDynLib(units)

When updating from 1.x to 2.x:

The name of the personal units file has changed from $HOME/units.dat
to $HOME/.units ($HOME/unitdef.units under Windows).

The format for nonlinear unit definitions has changed.  Run 'units -c'
and add the "units=" keyword in front of any bracketed unit
specifications.  

------------------------------------------------------------------

GNU `units' converts between different systems of units.  It can
handle multiplicative scale changes.  It can also handle nonlinear
conversions such as Celsius to Fahrenheit (which may appear to be
linear but is actually affine).  

General installation instructions appear in the file `INSTALL'.  You
should be able to run `./configure' followed by `make'.  If you give
no options to configure, it will compile units to look for the units
data file in a standard location (probably /usr/local/share).  If you
try to use the program without installing you will need to use the
`-f' option.  If you don't want to commit to an installation location,
you can invoke configure by typing `./configure --enable-path-search'.
Then no path name will be compiled into `units' and it will search the
current directory and the directories listed in your PATH environment
variable to find the units data file.

For full functionality you should have the GNU readline library
installed to provide history and editing of data entry.  You can
update currency conversions using the units_cur script which requires
Python and the unidecode module.  

The documentation is available in texinfo, roff, and text format.  The
man page is generated automatically from the texinfo documentation.
This man page produces readable results when run through nroff, but it
should probably not be printed with troff or groff---no effort has
been made to ensure that it prints out reasonably.  To generate a
printed manual, use `units.dvi' instead.

The distribution includes three simple icons that may be useful for
installation in a GUI.  Use the icotool command to extract the png
files from the .ico files.  The icon_ms.png file is suitable for use
as a small button.

This program has the following incompatibilties with unix `units':
  * The '-' character is a subtraction operator rather than a multiply
      operator by default.
  * Exponentiation in numbers requires an `e', so you must write 2.5e-2
      instead of 2.5-2.
  * Prefixes are listed in the units file.
  * GNU `units' tries the -s, -es, and -ies plural forms.
  * The default output format is slightly different.
  * The units database is much larger and more informative, but with some
      differences. (e.g. `g' is for gravity in unix `units' and grams in
      GNU `units'.)  The comment character has been changed to `#'.

GNU `units' includes the following extensions:
  * Multiplication can be written with a `*' if desired.
  * Exponents can be written with `^' or `**' in units.
  * Exponents can be larger than 9 if written with `^' or `**'.
  * Sums of units can be converted.
  * The units data file is extensively commented. 
  * Units which measure reciprocal dimensions can be converted.
  * Parentheses for grouping are supported.
  * Funtions such as sin, cos, and log are supported.
  * Roots of units and rational exponents can be computed.
  * Nonlinear units conversions are supported. 
  * Conversion to lists of units (e.g. feet and inches) is supported

----------------------------------------------------------------------

Ports

A Windows version

A binary distribution for Windows is available at
http://ftp.gnu.org/gnu/units/windows/ The executable was built with
Microsoft Visual C/C++ Express using Makefile.Win and the same source
files included in the source distribution.  There is currently no
support for UTF-8 or readline; however, command history and intraline
editing are available via the standard Windows console facilities
described in the documentation for doskey.

A port of units 1.87 to Windows is available from the
http://gnuwin32.sourceforge.net/packages/units.htm 
This port includes readline support.

A Java version of units by Roman Redziejowski <roman.redz@swipnet.se>
is available on SourceForge at http://units-in-java.sourceforge.net/

Two versions are available for Android.  Steve Pomeroy has a version
based on the the above Java version that you can obtain at 
http://staticfree.info/projects/units/ and Keith Flowers has compiled
the C code for Android: http://apps.keithflower.org/?page_id=6

A Perl version was written by Bob Walton <bob@bwalton.com> and can be
accessed either as a units converting web form or as perl source code
from:  http://bwalton.com/cgi-bin/myunitscgi.pl

Several years ago the program was ported to Pocket PC and given the
name QuickUnits. This version appears to remain available here:
http://www.freewareppc.com/calculator/quickunits.shtml

-----------------------------------------------------------------------------

Ideas the future (may or may not happen):

  * Bundle up the units conversion stuff into a library. 
  * Inflation adjusted currency?
  * Allow multiple definitions of the same unit and resolve the
      correct definition by a conformability check.  (This has
      exponential growth behavior in the number of units typed in!)
  * When a nonconformable units error is given list units the user
      might have meant (e.g. britainpound for pound) by a
      conformability check and string pattern match of some sort. 
      "spelling advice"
  * Allow some way of having units like '$' that don't require a trailing
      space so you can write '$5'.  This could be handled by having a 
      command in the units database that specifies units which automatically
      get a space inserted  after their name.   
  * Have a metacommand in the units datafile that specifies how plurals should
      be tried for this file.  This would allow expansion into other 
      languages.  (Of course, the real work of expanding into other languages
      is writing a units file that is appropriate for the language in question
      and includes local units.  It's not just a translation task.)
      Another thing that could be accomplished here would be translation of
      English words like "cubic" and "per" into their symbolic meanings.
      A command in the units file could indicate that "per" should be 
      substituted into a '/' and "cubic" means the cube the next unit.
      As it stands, "per" is hard coded into the parser.
  * Represent uncertainties in values in the database.

-----------------------------------------------------------------------------

Acknowledgements

    This program owes a lot to Jeff Conrad who made many helpful suggestions,
    found numerous bugs, and helped me to find the definitions of obscure
    units.  Chris Madsen also made some valuable contributions. 

    The documentation has greatly benefited from the suggestions made by
    Robert Chassell who kindly read several drafts.

    The following people have been particularly helpful in fixing portability
    problems: Kaveh Ghazi, Eric Backus, and Marcus Daniels.

Bug reports and suggestions for improvements should be sent to the author:
Adrian Mariano (adrianm@gnu.org).  

Overview
========
GNU 'units' program converts quantities expressed in various systems of
measurement to their equivalents in other systems of measurement.  Like
many similar programs, it can handle multiplicative scale changes; but
it can also handle nonlinear conversions such as Fahrenheit to Celsius
(which may appear to be linear but is actually affine).  The program can
also perform conversions from and to sums of units, such as converting
between meters and feet plus inches.

Basic operation is simple: at the 'You have:' prompt, enter the unit
from which you want to convert; at the subsequent 'You want:' prompt,
enter the unit to which you want to convert. For example,

  You have: ft
  You want: m
          * 0.3048
          / 3.2808399

In other words, 1 foot is equal to 0.3048 meter (exactly), and 1 meter
is equal to approximately 3.2808339 feet.

To quit the program, enter either 'quit' or 'exit' at either the 'You
have:' or 'You want:' prompt.  You can also quit by entering Ctrl-C at
any time.

The program's features are described in detail in the user manual. 

Building and Installation
=========================
General installation instructions appear in the file 'INSTALL'.  You
should be able to run './configure' followed by 'make'.  If you give
no options to configure, it will compile units to look for the units
data file in a standard location (probably /usr/local/share) and the
currency file in /usr/local/com.

To check that currency conversion update script works, and to verify
that units works properly, run 'make check'.

To install the program, run 'make install'.  

If you want to use the program without installing, you will need to use
the '-f' option; if you plan to use the program frequently without
installing, you can avoid having to do this by setting
'UNITSFILE=definitions.units' in the environment.  You can also avoid
having to do this by building for a relocatable installation, described
below.

If you wish to change the locations of the files you can invoke
'./configure' with the option '--prefix=<your_prefix>' to set a
different installation location.  You may also choose to use
'--sharedstatedir=<directory>' to change the location for the currency
file.  

Building for a Relocatable Installation
---------------------------------------
By default, the location of the units data file is compiled into the
program as an absolute path name, so its location cannot be changed.  If
you invoke configure by typing './configure --enable-relocation', the
data file location will be compiled into the program as
'definitions.units', and units will search for the data file in the
following places:

 *  In the directory that contains the units executable, typical for
    installations on Windows.

 *  In <anypath>/bin/../share/units, where <anypath>/bin is the
    directory that contains the units executable, typical for
    installations on Unix-like systems.  For example, the default
    installation typically places the executable in /usr/local/bin and
    the data file in /usr/local/share/units; if you wished, you could
    change these locations to /usr/me/bin and /usr/me/share/units
    without recompiling the program.

If configure is run with '--enable-relocation', you can run units
without installing without giving the '-f' option.

If units is invoked with the '-f' option or the environment variable
UNITSFILE is set, that path is used, and it makes no difference whether
configure was run with the '--enable-relocation' option.

GNU readline Library
--------------------
For full functionality you should have the GNU readline library
installed to provide history and editing of data entry.

Note that MacOS comes with editline, which is similar to GNU readline,
but not similar enough.  When compiling under MacOS the configure
script looks for GNU readline in /opt/local.  If you have installed it
somewhere else then invoke configure as follows:
       ./configure -I<path_to_includes> -L<path_to_libs>
Then the configure script should detect readline and compile it in.        

If you are installing on Windows, readline may not be available;
however, command history and intraline editing are available via the
standard Windows console facilities described in the documentation for
doskey.

Currency Conversion Updates
---------------------------
You can update currency conversions using the units_cur Python script;
see the user manual for details.

Building on Windows
===================
units can be built from the Windows command prompt using Visual Studio;
see UnitsWin.pdf for details.

units can also be built from the MKS Korn shell using Visual Studio;
see UnitsMKS.pdf for details.

Documentation
=============
The documentation is provided in texinfo, roff, and text format.  The
roff manual page source units.man is generated automatically from the
texinfo documentation; this produces a readable man page when run
through nroff, but most equations are not included.  The manual page
formats well for printing or as PDF with groff.  A printed manual can
also be generated using 'units.dvi'; use this if you encounter problems
with groff.

Icons
=====
The distribution includes three icons that may be useful for
installation in a GUI.  Use the icotool command to extract the png files
from the .ico files.  The icon_ms.png file is suitable for use as a
small button.

Incompatibilities with Unix 'units'
===================================
This program has the following incompatibilities with unix 'units':
  * The '-' character is a subtraction operator rather than a multiply
      operator by default.
  * Exponentiation in numbers requires an 'e', so you must write 2.5e-2
      instead of 2.5-2.
  * Prefixes are listed in the units file.
  * GNU 'units' tries the -s, -es, and -ies plural forms.
  * The default output format is slightly different.
  * The units database is much larger and more informative, but with some
      differences. (e.g. 'g' is for gravity in unix 'units' and grams in
      GNU 'units'.)  The comment character has been changed to '#'.

GNU 'units' Extensions
======================
GNU 'units' includes the following extensions:
  * Multiplication can be written with a '*' if desired.
  * Exponents can be written with '^' or '**' in units.
  * Exponents can be larger than 9 if written with '^' or '**'.
  * Sums of units can be converted.
  * The units data file is extensively commented. 
  * Units which measure reciprocal dimensions can be converted.
  * Parentheses for grouping are supported.
  * Functions such as sin, cos, and log are supported.
  * Roots of units and rational exponents can be computed.
  * Nonlinear units conversions are supported. 
  * Conversion to lists of units (e.g. feet and inches) is supported

Windows Binary Distribution
===========================
A binary distribution for Windows is available at
http://ftp.gnu.org/gnu/units/windows/. The executable was built with
Microsoft Visual Studio using Makefile.Win and the same source
files included in the source distribution.  The binary version is
usually the same as that of the current source distribution.  There is
currently no support for UTF-8 or readline; however, command history and
intraline editing are available via the standard Windows console
facilities described in the documentation for doskey.

Other Ports
===========
A port of units 1.87 to Windows is available from the
http://gnuwin32.sourceforge.net/packages/units.htm 
This port includes readline support.

A Java version of units by Roman Redziejowski <roman.redz@swipnet.se>
is available on SourceForge at http://units-in-java.sourceforge.net/

Two versions are available for Android.  Steve Pomeroy has a version
based on the above Java version that you can obtain at http://
staticfree.info/projects/units/ and Keith Flowers has compiled the C
code for Android: http://apps.keithflower.org/?page_id=6

A Perl version was written by Bob Walton <bob@bwalton.com> and can be
accessed either as a units converting web form or as perl source code
from:  http://bwalton.com/cgi-bin/myunitscgi.pl

A project called Frink uses a (modified) version of the units
database to supply a units-aware programming language.
https://futureboy.us/frinkdocs/

Jillian England has created a units definition file that changes
energy to mass and seconds to meters:
https://github.com/NinerXrayBravoTwoTwo/MassEnergyUnits

-----------------------------------------------------------------------------

When updating from 1.x to 2.x:

The name of the personal units file has changed from $HOME/units.dat
to $HOME/.units ($HOME/unitdef.units under Windows).

The format for nonlinear unit definitions has changed.  Run 'units -c'
and add the "units=" keyword in front of any bracketed unit
specifications.  

-----------------------------------------------------------------------------
 
Ideas the future (may or may not happen):

  * Bundle up the units conversion stuff into a library. 
  * Inflation adjusted currency?
  * Allow multiple definitions of the same unit and resolve the
      correct definition by a conformability check.  (This has
      exponential growth behavior in the number of units typed in!)
  * When a nonconformable units error is given list units the user
      might have meant (e.g. britainpound for pound) by a
      conformability check and string pattern match of some sort. 
      "spelling advice"
  * Allow some way of having units like '$' that don't require a trailing
      space so you can write '$5'.  This could be handled by having a 
      command in the units database that specifies units which automatically
      get a space inserted  after their name.   
  * Have a metacommand in the units datafile that specifies how plurals should
      be tried for this file.  This would allow expansion into other 
      languages.  (Of course, the real work of expanding into other languages
      is writing a units file that is appropriate for the language in question
      and includes local units.  It's not just a translation task.)
      Another thing that could be accomplished here would be translation of
      English words like "cubic" and "per" into their symbolic meanings.
      A command in the units file could indicate that "per" should be 
      substituted into a '/' and "cubic" means the cube the next unit.
      As it stands, "per" is hard coded into the parser.
  * Represent uncertainties in values in the database.

Acknowledgements
================
    This program owes a lot to Jeff Conrad who made many helpful suggestions,
    found numerous bugs, and helped me to find the definitions of obscure
    units.  Chris Madsen also made some valuable contributions. 

    The documentation has greatly benefited from the suggestions made by
    Robert Chassell who kindly read several drafts.

    The following people have been particularly helpful in fixing portability
    problems: Kaveh Ghazi, Eric Backus, and Marcus Daniels.

Bug reports and suggestions for improvements should be sent to the author:
Adrian Mariano (adrianm@gnu.org).  

GNU 'units' converts between different systems of units.  It can
handle multiplicative scale changes.  It can also handle nonlinear
conversions such as Celsius to Fahrenheit (which may appear to be
linear but is actually affine).  

------------------------------------------------------------------

Installation

General installation instructions appear in the file 'INSTALL'.  You
should be able to run './configure' followed by 'make'.  If you give
no options to configure, it will compile units to look for the units
data file in a standard location (probably /usr/local/share) and the
currency file in /usr/llocal/com.  If you try to use the program
without installing you will need to use the '-f' option.  If you don't
want to commit to an installation location, you can invoke configure
by typing './configure --enable-path-search'.  Then no path name will
be compiled into 'units' and it will search the current directory and
the directories listed in your PATH environment variable to find the
units data file.

If you wish to change the locations of the files you can invoke
'./configure' with the option '--prefix=<your_prefix>' to set a
different installation location.  You may also choose to use
'--sharedstatedir=<directory>' to change the location for the currency
file.  

If you are building on Windows with the MKS Toolkit, see the file
'UnitsMKS.pdf'.

If you are building with Microsoft Visual Studio from the Windows
command prompt, see the file 'UnitsWin'.

For full functionality you should have the GNU readline library
installed to provide history and editing of data entry.  You can
update currency conversions using the units_cur Python script.

The documentation is available in texinfo, roff, and text format.  The
man page is generated automatically from the texinfo documentation.
This man page produces readable results when run through nroff, but it
should probably not be printed with troff or groff--no effort has
been made to ensure that it prints out reasonably.  To generate a
printed manual, use 'units.dvi' instead.

The distribution includes three icons that may be useful for
installation in a GUI.  Use the icotool command to extract the png files
from the .ico files.  The icon_ms.png file is suitable for use as a
small button.

This program has the following incompatibilties with unix 'units':
  * The '-' character is a subtraction operator rather than a multiply
      operator by default.
  * Exponentiation in numbers requires an 'e', so you must write 2.5e-2
      instead of 2.5-2.
  * Prefixes are listed in the units file.
  * GNU 'units' tries the -s, -es, and -ies plural forms.
  * The default output format is slightly different.
  * The units database is much larger and more informative, but with some
      differences. (e.g. 'g' is for gravity in unix 'units' and grams in
      GNU 'units'.)  The comment character has been changed to '#'.

GNU 'units' includes the following extensions:
  * Multiplication can be written with a '*' if desired.
  * Exponents can be written with '^' or '**' in units.
  * Exponents can be larger than 9 if written with '^' or '**'.
  * Sums of units can be converted.
  * The units data file is extensively commented. 
  * Units which measure reciprocal dimensions can be converted.
  * Parentheses for grouping are supported.
  * Funtions such as sin, cos, and log are supported.
  * Roots of units and rational exponents can be computed.
  * Nonlinear units conversions are supported. 
  * Conversion to lists of units (e.g. feet and inches) is supported

----------------------------------------------------------------------

Doron Behar <doron.behar@gmail.com> has written a script for downloading
currency data from openexchangerates which may include additional,
currency rates for cryptocurrencies and even black market currencies as
well. In order to use this script, you have to provide it an API key
which can be obtained after you register at their website. A copy of
Behar's script is included with the units distribution under the name
"units_cur_openexchangerates".  The latest version can be found at

https://gitlab.com/doronbehar/units-openexchangerates

Any concerns or bug reports for this script should be sent to him.

----------------------------------------------------------------------

Ports

Windows: Building with Microsoft Visual Studio

units can be built from the Windows command prompt using Visual Studio;
see UnitsWin for details.

units can also be built from the the MKS Korn shell using Visual Studio;
see UnitsMKS for details.

Windows: Binary Distribution

A binary distribution for Windows is available at
http://ftp.gnu.org/gnu/units/windows/. The executable was built with
Microsoft Visual Studio using Makefile.Win and the same source
files included in the source distribution.  There is currently no
support for UTF-8 or readline; however, command history and intraline
editing are available via the standard Windows console facilities
described in the documentation for doskey.

A port of units 1.87 to Windows is available from the
http://gnuwin32.sourceforge.net/packages/units.htm 
This port includes readline support.

A Java version of units by Roman Redziejowski <roman.redz@swipnet.se>
is available on SourceForge at http://units-in-java.sourceforge.net/

Two versions are available for Android.  Steve Pomeroy has a version
based on the the above Java version that you can obtain at 
http://staticfree.info/projects/units/ and Keith Flowers has compiled
the C code for Android: http://apps.keithflower.org/?page_id=6

A Perl version was written by Bob Walton <bob@bwalton.com> and can be
accessed either as a units converting web form or as perl source code
from:  http://bwalton.com/cgi-bin/myunitscgi.pl

A project called Frink uses a (modified) version of the units
database to supply a units-aware programming language.
https://futureboy.us/frinkdocs/

Jillian England has created a units definition file that changes
energy to mass and seconds to meters:
https://github.com/NinerXrayBravoTwoTwo/MassEnergyUnits

-----------------------------------------------------------------------------

When updating from 1.x to 2.x:

The name of the personal units file has changed from $HOME/units.dat
to $HOME/.units ($HOME/unitdef.units under Windows).

The format for nonlinear unit definitions has changed.  Run 'units -c'
and add the "units=" keyword in front of any bracketed unit
specifications.  

-----------------------------------------------------------------------------
 
Ideas the future (may or may not happen):

  * Bundle up the units conversion stuff into a library. 
  * Inflation adjusted currency?
  * Allow multiple definitions of the same unit and resolve the
      correct definition by a conformability check.  (This has
      exponential growth behavior in the number of units typed in!)
  * When a nonconformable units error is given list units the user
      might have meant (e.g. britainpound for pound) by a
      conformability check and string pattern match of some sort. 
      "spelling advice"
  * Allow some way of having units like '$' that don't require a trailing
      space so you can write '$5'.  This could be handled by having a 
      command in the units database that specifies units which automatically
      get a space inserted  after their name.   
  * Have a metacommand in the units datafile that specifies how plurals should
      be tried for this file.  This would allow expansion into other 
      languages.  (Of course, the real work of expanding into other languages
      is writing a units file that is appropriate for the language in question
      and includes local units.  It's not just a translation task.)
      Another thing that could be accomplished here would be translation of
      English words like "cubic" and "per" into their symbolic meanings.
      A command in the units file could indicate that "per" should be 
      substituted into a '/' and "cubic" means the cube the next unit.
      As it stands, "per" is hard coded into the parser.
  * Represent uncertainties in values in the database.

-----------------------------------------------------------------------------

Acknowledgements

    This program owes a lot to Jeff Conrad who made many helpful suggestions,
    found numerous bugs, and helped me to find the definitions of obscure
    units.  Chris Madsen also made some valuable contributions. 

    The documentation has greatly benefited from the suggestions made by
    Robert Chassell who kindly read several drafts.

    The following people have been particularly helpful in fixing portability
    problems: Kaveh Ghazi, Eric Backus, and Marcus Daniels.

Bug reports and suggestions for improvements should be sent to the author:
Adrian Mariano (adrianm@gnu.org).  

			 units 2.25 for Windows
			     January 2026

Introduction
============
The units program converts quantities expressed in various systems of
measurement to their equivalents in other systems of measurement.  Like
many similar programs, it can handle multiplicative scale changes; but
it can also handle nonlinear conversions such as Fahrenheit to Celsius.
The program can also perform conversions from and to sums of units, such
as converting between meters and feet plus inches.

Basic operation is simple: at the 'You have:' prompt, enter the unit
from which you want to convert; at the subsequent 'You want:' prompt,
enter the unit to which you want to convert. For example,

  You have: ft
  You want: m
          * 0.3048
          / 3.2808399

In other words, 1 foot is equal to 0.3048 meter (exactly), and 1 meter
is equal to approximately 3.2808339 feet.

To quit the program, press Ctrl-C or Ctrl-Z; with the latter, you may
need to press 'Enter'.

Running units
=============
You can run units from the command line, or via Start Menu or Desktop
shortcuts if you chose to create them during installation.  The 'units'
shortcut runs units in the traditional Windows Console Host.  If you
have Windows Terminal installed, you should have additional 'units WT'
shortcut that runs units in Windows Terminal with Unicode support.

Documentation
=============
The program's features are briefly described in UnitsForWindows, and are
described in detail in the user manual.  Unless you declined to create a
Start Menu folder, you can access the user manual from the Start Menu.
From the Start Menu, select 'Programs', position the cursor over 'units'
(or whatever name you specified during installation), and select 'Units
User Manual' from the submenu that appears.  The manual is PDF; to view
it, you will need the free Adobe Reader or a similar application.  Adobe
Reader is available from Adobe at http://www.adobe.com.

Updating Currency Definitions
=============================
The script units_cur.py can be used to update currency definitions (if
your system hides file extensions, this script will display as
"units_cur").  The script requires Python (available from
https://www.python.org/).

If you want to use the currency updater, install Python
You then should be able to run units_cur.py using the shortcut on the
Start Menu, or if you have added the units installation directory to
your PATH, from a command window.

Updating currency definitions is described in more detail in
UnitsForWindows.

library(rvest)
r = read_html(
 "http://cfconventions.org/Data/cf-standard-names/34/build/cf-standard-name-table.html")
tb = html_table(r, fill = TRUE)
library(units)
cf_units = tb[[3]][2][[1]]
u = lapply(cf_units, as_units, force_single_symbol=TRUE, check_is_valid=FALSE)
names(u) = substr(tb[[3]][1][[1]], 1, 50)
# does udunits2 understand them?
all(sapply(u, units::ud_is_parseable))

library(ggplot2)
library(units)
mtcars$consumption <- set_units(mtcars$mpg, mi / gallon)
mtcars$power <- set_units(mtcars$hp, hp)
ggplot(mtcars) + geom_point(aes(power, consumption))
ggplot(mtcars) + geom_point(aes(1/power, 1/consumption))

# to generate figure 1 in vignette:

#pdf("fig1.pdf", width = 4, height = 4)
library(units)
units_options(negative_power = TRUE)
mtcars$consumption <- set_units(mtcars$mpg, mi / gallon)
mtcars$displacement <- set_units(mtcars$disp, "in^3")
units(mtcars$displacement) <- "cm^3"
units(mtcars$consumption) <- "km/l"
m = par("mar")
m[3] = 0.1
par(mar = m + c(0, .3, 0, 0))
with(mtcars, plot(1/displacement, 1/consumption))
#pdf("fig2.pdf", width = 4, height = 4)
library(ggplot2)
ggplot(mtcars) + geom_point(aes(1/displacement, 1/consumption))

# demo illustrating the different year lengths, as available in
# NISTunits:
library(NISTunits)
options(digits=7)
NISTyearTOsec(1) / (24 * 3600) # number of mean solar days days in 1 common Gregorian year
NISTyearTropicalTOsec(1) / (24 * 3600) # number of mean solar days days in 1 tropical year
NISTyearSiderealTOsec(1) / (24 * 3600) # number of mean solar days days in 1 sidereal year
NISTyearTropicalTOsec(1) / (24 * 3600) # number of mean solar days days in 1 tropical year
NISTyearSiderealTOsec(1) / (24 * 3600) # number of mean solar days days in 1 sidereal year

# ... and in udunits2:
library(units)
as_units("common_year")
as_units("leap_year")
as_units("Gregorian_year")
as_units("Julian_year")
as_units("yr") # yr = tropical:

.startup_msg <- function(warn = FALSE) {
  paste("udunits database from" , default_units_xml(warn))
}

default_units_xml <- function(warn = FALSE) {
  paths = c(
    Sys.getenv("UDUNITS2_XML_PATH"),
    "/usr/local/share/udunits/udunits2.xml",
    "/usr/share/xml/udunits/udunits2.xml",
    "/usr/share/udunits/udunits2.xml")

  fallback = system.file("share/udunits/udunits2.xml", package="units")

  w = which(file.exists(paths))
  if (length(w) > 1 && warn)
    warning(paste("multiple udunits databases present:\n - "),
            paste(paths[w], collapse = "\n - "), call.=FALSE)
  p = if (length(w) == 0) fallback else paths[w[1]]

  if (!file.exists(p))
    stop("Failed to identify udunits database: units will not work properly.",
         "\nPlease set the UDUNITS2_XML_PATH environment variable before ",
         "attempting to read the units database")
  p
}

#' Load a unit system
#'
#' Load an XML database containing a unit system compatible with UDUNITS2.
#'
#' A unit system comprises a root \code{<unit-system>} and a number of children
#' defining prefixes (\code{<prefix>}) or units (\code{<unit>}).
#' See the contents of
#'
#' \code{system.file("share/udunits", package="units")}
#'
#' for examples.
#'
#' @param path a path to a valid unit system in XML format.
#'
#' @examples
#' # load a new unit system
#' load_units_xml(system.file("share/udunits/udunits2-base.xml", package="units"))
#' \dontrun{
#' set_units(1, rad) # doesn'twork
#' }
#'
#' # reload the default unit system
#' load_units_xml()
#' set_units(1, rad) # works again
#'
#' @export
load_units_xml <- function(path = default_units_xml()) {
  if (.Platform$OS.type == "windows")
    path <- shortPathName(path)
  ud_init(path)
}

.read_ud_db <- function(path = default_units_xml(), type) {
  if (! requireNamespace("xml2", quietly = TRUE))
    stop("package 'xml2' is required for this functionality", call.=FALSE)

  db <- xml2::read_xml(path)

  # resolve imports
  imports <- xml2::xml_find_all(db, "import")
  files <- sapply(imports, xml2::xml_text)
  for (i in files) {
    import <- xml2::read_xml(file.path(dirname(path), i))
    for (child in xml2::xml_children(import)) {
      source <- gsub("udunits2-|\\.xml", "", i)
      source <- paste0("<doc><source>", source, "</source></doc>")
      source <- xml2::xml_child(xml2::read_xml(source))
      xml2::xml_add_child(child, source)
      xml2::xml_add_child(db, child)
    }
  }
  xml2::xml_remove(imports)

  # filter by type
  if (!missing(type)) {
    children <- xml2::xml_children(db)
    xml2::xml_remove(children[xml2::xml_name(children) != type])
  }

  db
}

type_sum.units <- function(x, ...) {
  gr = units_options("group")
  # see https://github.com/r-lib/pillar/issues/73 : currently the [ and ] mess up things.
  structure(paste0(gr[1], as.character(units(x)), gr[2]),
            class = "type_sum_units")
}

type_sum.mixed_units <- function(x, ...) {
  "mixed_units"
}

pillar_shaft.units <- function(x, ...) {
  pillar::pillar_shaft(unclass(x), ...)
}

pillar_shaft.mixed_units <- function(x, ...) {
  u_char <- pillar::style_subtle(sapply(x, function(unit) format(unit[0])))
  out <- paste0(format(unclass(x), ...), u_char)
  pillar::new_pillar_shaft_simple(out, align = "right", min_width = 6)
}

format_type_sum.type_sum_units <- function(x, width, ...) {
  pillar::style_subtle(x)
}


# vctrs proxying and restoration -------------------------------------

vec_proxy.units = function(x, ...) {
  x
}
vec_restore.units = function(x, to, ...) {
  restore_units(x, to)
}
restore_units <- function(x, to) {
  out <- .as.units(x, units(to))
  attr(out, "pillar") <- attr(to, "pillar")
  out
}


# vctrs coercion -----------------------------------------------------

vec_ptype2.units.units = function(x, y, ..., x_arg = "", y_arg = "") {
  x_units = units(x)
  y_units = units(y)

  if (!ud_are_convertible(x_units, y_units))
    vctrs::stop_incompatible_type(x, y, x_arg = x_arg, y_arg = y_arg)

  x_bare = drop_units(x)
  y_bare = drop_units(y)
  common = vctrs::vec_ptype2(x_bare, y_bare, ..., x_arg = x_arg, y_arg = y_arg)

  # Use left-hand side units
  set_units(common, x_units, mode = "standard")
}

vec_cast.units.units = function(x, to, ..., x_arg = "", to_arg = "") {
  x_units = units(x)
  to_units = units(to)

  if (!ud_are_convertible(x_units, to_units))
    vctrs::stop_incompatible_cast(x, to, x_arg = x_arg, to_arg = to_arg)

  # Convert to target units before converting base type. Unit
  # conversion might change the type and so must happen first.
  out = set_units(x, to_units, mode = "standard")

  out_bare = drop_units(out)
  to_bare = drop_units(to)
  out = vctrs::vec_cast(out_bare, to_bare, ..., x_arg = x_arg, to_arg = to_arg)

  # Set target units again
  set_units(out, to_units, mode = "standard")
}


#nocov start
register_all_s3_methods <- function() {
  register_s3_method("pillar::type_sum", "units")
  register_s3_method("pillar::type_sum", "mixed_units")
  register_s3_method("pillar::pillar_shaft", "units")
  register_s3_method("pillar::pillar_shaft", "mixed_units")
  register_s3_method("pillar::format_type_sum", "type_sum_units")
  register_s3_method("vctrs::vec_proxy", "units")
  register_s3_method("vctrs::vec_restore", "units")
  register_s3_method("vctrs::vec_ptype2", "units.units")
  register_s3_method("vctrs::vec_cast", "units.units")
  register_s3_method("ggplot2::scale_type", "units")
}

register_s3_method <- function(generic, class, fun=NULL) {
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  pieces <- strsplit(generic, "::")[[1]]
  stopifnot(length(pieces) == 2)
  package <- pieces[[1]]
  generic <- pieces[[2]]

  if (is.null(fun))
    fun <- get(paste0(generic, ".", class), envir=parent.frame())
  stopifnot(is.function(fun))

  if (package %in% loadedNamespaces())
    registerS3method(generic, class, fun, envir=asNamespace(package))

  # Always register hook in case package is later unloaded & reloaded
  setHook(packageEvent(package, "onLoad"), function(...)
    registerS3method(generic, class, fun, envir=asNamespace(package)))
}
# nocov end

.units_options <- new.env(FALSE, parent = globalenv())

.default_options <- list(
  sep = c("~", "~"),
  group = c("[", "]"),
  negative_power = FALSE,
  parse = TRUE,
  set_units_mode = "symbols",
  auto_convert_names_to_symbols = TRUE,
  simplify = NA,
  allow_mixed = FALSE,
  unitless_symbol = "1",
  define_bel = TRUE
)

.setopt <- function(option) {
  name <- paste0(".units.", substitute(option))
  last <- mget(name, envir = .units_options, ifnotfound = NA)[[1]]
  assign(name, option, envir = .units_options)
  last
}

#' set one or more units global options
#'
#' set units global options, mostly related how units are printed and plotted
#' @param ... named options (character) for which the value is queried
#' @param sep character length two; default \code{c("~", "~")}; space separator between variable and units, and space separator between two different units
#' @param group character length two; start and end group, may be two empty strings, a parenthesis pair, or square brackets; default: square brackets.
#' @param negative_power logical, default \code{FALSE}; should denominators have negative power, or follow a division symbol?
#' @param parse logical, default \code{TRUE}; should the units be made into an expression (so we get subscripts)? Setting to \code{FALSE} may be useful if \link{parse} fails, e.g. if the unit contains symbols that assume a particular encoding
#' @param set_units_mode character; either \code{"symbols"} or \code{"standard"}; see \link{set_units}; default is \code{"symbols"}
#' @param auto_convert_names_to_symbols logical, default \code{TRUE}: should names, such as \code{degree_C} be converted to their usual symbol?
#' @param simplify logical, default \code{NA}; simplify units in expressions?
#' @param allow_mixed logical; if \code{TRUE}, combining mixed units creates a \code{mixed_units} object, if \code{FALSE} it generates an error
#' @param unitless_symbol character; set the symbol to use for unitless (1) units
#' @param define_bel logical; if \code{TRUE}, define the unit \code{B} (i.e., the \emph{bel}, widely used with the \emph{deci-} prefix as \code{dB}, \emph{decibel}) as an alias of \code{lg(re 1)}. \code{TRUE} by default, unless \code{B} is already defined in the existing XML database.
#' @details This sets or gets units options. Set them by using named arguments, get them by passing the option name.
#'
#' The default \code{NA} value for \code{simplify} means units are not simplified in \link{set_units} or \link{as_units}, but are simplified in arithmetical expressions.
#' @return in case options are set, invisibly a named list with the option values that are being set; if an option is queried, the current option value.
#' @examples
#' old = units_options(sep = c("~~~", "~"), group = c("", "")) # more space, parenthesis
#' old
#' ## set back to defaults:
#' units_options(sep = c("~", "~"), group = c("[", "]"), negative_power = FALSE, parse = TRUE)
#' units_options("group")
#' @export
units_options = function(..., sep, group, negative_power, parse, set_units_mode, auto_convert_names_to_symbols, simplify,
		allow_mixed, unitless_symbol, define_bel) {
	# op = as.list(units:::.units_options)
	ret = list()
	if (!missing(sep)) {
		stopifnot(is.character(sep) && length(sep) == 2)
		ret$sep = .setopt(sep)
	}
	if (!missing(group)) {
		stopifnot(is.character(group) && length(group) == 2)
		ret$group = .setopt(group)
	}
	if (!missing(negative_power)) {
		stopifnot(is.logical(negative_power))
		ret$negative_power = .setopt(negative_power)
	}
	if (!missing(parse)) {
		stopifnot(is.logical(parse))
		ret$parse = .setopt(parse)
	}
	if (!missing(set_units_mode)) {
		stopifnot(is.character(set_units_mode) && length(set_units_mode) == 1)
		ret$set_units_mode = .setopt(set_units_mode)
	}
	if (!missing(auto_convert_names_to_symbols)) {
		stopifnot(is.logical(auto_convert_names_to_symbols))
		ret$auto_convert_names_to_symbols = .setopt(auto_convert_names_to_symbols)
	}
	if (!missing(simplify)) {
		stopifnot(is.logical(simplify))
		ret$simplify = .setopt(simplify)
	}
	if (!missing(allow_mixed)) {
		stopifnot(is.logical(allow_mixed))
		ret$allow_mixed = .setopt(allow_mixed)
	}
	if (!missing(unitless_symbol)) {
		stopifnot(is.character(unitless_symbol), length(unitless_symbol) == 1)
		ret$unitless_symbol = .setopt(unitless_symbol)
	}
	if (!missing(define_bel)) {
	  stopifnot(is.logical(define_bel))
	  ret$define_bel = .setopt(define_bel)
	  if (!identical(ret$define_bel, define_bel)) {
	    if (!is.na(ret$define_bel))
	      remove_unit("B", "bel")
	    if (define_bel)
	      install_unit("B", "lg(re 1)", "bel")
	  }
	}

	dots = list(...)
	if (length(dots)) {
		if (length(ret) > 0)
			stop("either set, or get units_option, but don't try to do both in one function call")
		if (is.list(dots[[1]]))
			do.call(units_options, dots[[1]])
		else
			get(paste0(".units.", dots[[1]]), envir = .units_options)
	} else
		invisible(ret)
}

.units.simplify = function() {
  get(".units.simplify", envir = .units_options)
}

.as.mixed_units = function(x) {
	stopifnot(is.list(unclass(x)))
	structure(x, class = c("mixed_units", "list"))
}

# constructor function:
#' Create or convert to a mixed units list-column
#' @param x numeric, or vector of class \code{units}
#' @param values character vector with units encodings, or list with symbolic units of class \code{mixed_symbolic_units}
#' @param value see values
#' @param ... ignored
#' @details if \code{x} is of class \code{units}, \code{values} should be missing or of class \code{mixed_symbolic_units}; if \code{x} is numeric, \code{values} should be a character vector the length of \code{x}.
#' @examples
#' a <- 1:4
#' u <- c("m/s", "km/h", "mg/L", "g")
#' mixed_units(a, u)
#' units(a) = as_units("m/s")
#' mixed_units(a) # converts to mixed representation
#' @export
mixed_units <- function(x, values, ...) UseMethod("mixed_units")


#' @export
mixed_units.units = function(x, values, ...) {
	stopifnot(missing(values))
	u = as.character(units(x))
	mixed_units(unclass(x), rep(u, length(x)))
}


#' @export
mixed_units.numeric = function(x, values, ...) {
	#stopifnot(length(x) == length(values), is.character(values), is.numeric(x))
	stopifnot(is.character(values), is.numeric(x))
	.as.mixed_units(mapply(set_units, x, values, mode = "standard", SIMPLIFY = FALSE))
}

#' @export
#' @name mixed_units
`units<-.mixed_units` = function(x, value) {
	set_units(x, value)
}


#' @export
format.mixed_units = function(x, ...) {
	sapply(x, format, ...)
}

#' @export
`[.mixed_units` = function(x, i, ...) {
	.as.mixed_units(unclass(x)[i])
}

#' @export
c.mixed_units = function(...) {
	args = list(...)
	.as.mixed_units(do.call(c, lapply(args, unclass)))
}

#' @export
set_units.mixed_units = function(x, value, ..., mode = "standard") {
	if (! is.character(value))
		stop("use character string to denote target unit") # FIXME: rlang::quo stuff needed here?

  # conversion data frame and split
  cv <- data.frame(
    val = as.numeric(x), from = I(units(x)), to = value, idx = seq_along(x),
    stringsAsFactors = FALSE)
  sp <- paste(cv$from, cv$to, sep=".")
  sp <- factor(sp, levels=unique(sp))

  # grouped conversion
  cv <- do.call(rbind, unname(by(cv, sp, function(x) {
    x$val <- set_units(x$val, x$from[[1]], mode="standard")
    x$val <- mixed_units(set_units(x$val, x$to[1], mode="standard"))
    x
  }, simplify=FALSE)))

  # reordering
  cv[order(cv$idx), ]$val
}

#' @export
as_units.mixed_units = function(x, ...) {
	set_units(do.call(c, x), value = units(x[[1]]), mode = "standard")
}


#' @export
units.mixed_units = function(x) {
  u = lapply(x, function(i) if (inherits(i, "units")) units(i) else NULL)
	structure(u, class = "mixed_symbolic_units")
}

#' @export
as.character.mixed_symbolic_units = function(x, ...) {
	sapply(x, function(i) if (!is.null(i)) as.character(i) else "NULL")
}

.cat_units_table <- function(x) {
  cat("Mixed units: ")
  if (!length(x)) return()

  tbl = table(as.character(units(x)))
  tbl = paste(names(tbl), " (", as.numeric(tbl), ")", sep = "")
  cat(paste(tbl, collapse = ", "), "\n")
}

#' @export
print.mixed_units = function(x, ...) {
  .cat_units_table(x)
	cat(paste(format(x, ...), collapse = ", "), "\n")
}

#' @export
str.mixed_units = function(object, ...) {
  .cat_units_table(object)
  cat(capture.output(str(unclass(object), ...))[-1], sep="\n")
}

#' @name drop_units
#' @export
drop_units.mixed_units = function(x) {
	sapply(x, drop_units)
}

#' @export
Ops.mixed_units = function(e1, e2) {
	if (inherits(e2, "units"))
		e2 = mixed_units(e2)
    ret = switch(.Generic,
			"==" = mapply(function(x, y) { x == y }, e1, e2, SIMPLIFY = TRUE),
			"!=" = mapply(function(x, y) { x != y }, e1, e2, SIMPLIFY = TRUE),
			"*"  = mapply(function(x, y) { x * y  }, e1, e2, SIMPLIFY = FALSE),
			"/"  = mapply(function(x, y) { x / y  }, e1, e2, SIMPLIFY = FALSE),
			"+"  = mapply(function(x, y) { x + y  }, e1, e2, SIMPLIFY = FALSE),
			"-"  = mapply(function(x, y) { x - y  }, e1, e2, SIMPLIFY = FALSE),
			stop(paste("operation", .Generic, "not supported"))
		)
	if (is.list(ret))
		ret = .as.mixed_units(ret)
	ret
}

#' @export
unique.mixed_units <- function(x, incomparables = FALSE, ...) {
  .as.mixed_units(NextMethod())
}

#' @export
as.data.frame.mixed_units <- as.data.frame.AsIs

.symbolic_units <- function(numerator, denominator = vector("character")) {
  structure(list(numerator = numerator,
                 denominator = denominator),
            class = "symbolic_units")
}

.multiply_symbolic_units <- function(value, e1, e2) {
  numerator <- sort(c(e1$numerator, e2$numerator))
  denominator <- sort(c(e1$denominator, e2$denominator))
  .simplify_units(value, .symbolic_units(numerator, denominator))
}

.invert_symbolic_units <- function(e) {
  .symbolic_units(e$denominator, e$numerator)
}

.divide_symbolic_units <- function(value, e1, e2) {
  .multiply_symbolic_units(value, e1, .invert_symbolic_units(e2))
}

.same_units <- function(e1, e2) {
  identical(e1$numerator, e2$numerator) && identical(e1$denominator, e2$denominator)
}

# Inside the group generic functions we do have .Generic even if the diagnostics
# think we do not.
# !diagnostics suppress=.Generic
#' @export
Ops.symbolic_units <- function(e1, e2) {
  if (nargs() == 1)
    stop(paste("unary", .Generic, "not defined for \"units\" objects"))

  eq <- switch(.Generic, "==" = , "!=" = TRUE, FALSE)
  if (eq) {
    if (.Generic == "==")
      .same_units(e1, e2)
    else
      !.same_units(e1, e2)
  } else
    stop(paste("operation", .Generic, "not allowed for symbolic operators")) # nocov
}

#' The "unit" type for vectors that are actually dimension-less.
#' @export
unitless <- .symbolic_units(vector("character"), vector("character"))

.pretty_print_sequence <- function(terms, op, neg_power = FALSE, sep = "") {
  pwr_op <- if (op == " ") "" else "^"
  sym <- unique(terms)
  pwr <- tabulate(factor(terms, sym))
  if (neg_power) pwr <- pwr * -1

  for (i in seq_along(sym)) if (pwr[i] != 1)
    sym[i] <- paste(sym[i], pwr[i], sep = pwr_op)

  s <- matrix(c(sym, rep(paste0(op, sep), length(sym)-1), ""), nrow=length(sym))
  if (!units_options("strict_tokenizer"))
    s[sym >= "0" & sym <= "9", 2] <- ""
  paste0(t(s), collapse="")
}

#' @export
as.character.symbolic_units <- function(x, ...,
                                        neg_power = units_options("negative_power"),
                                        escape_units = FALSE, prod_sep = "*", plot_sep = "") {
  sep <- plot_sep

  numerator <- x$numerator[x$numerator != "1"]
  denominator <- x$denominator[x$denominator != "1"]
  if (escape_units) {
    numerator <- unlist(Map(function(name) paste0("`", name, "`", sep = ""), numerator))
    denominator <- unlist(Map(function(name) paste0("`", name, "`", sep = ""), denominator))
  }

  if (x == unitless) { # xxx
    u <- if (escape_units)
      unlist(Map(function(name) paste0("`", name, "`", sep = ""),
                 units_options("unitless_symbol")))
    else
      units_options("unitless_symbol")
    return(u)
  }

  num_str <- denom_str <- character(0)

  num_str <- if (length(numerator) > 0)
    .pretty_print_sequence(numerator, prod_sep, FALSE, plot_sep)
  else if (! neg_power) "1" # only denominator: 1/(cm^2*h)

  if (length(denominator) > 0) {
    sep <- if (neg_power)
      paste0(prod_sep, plot_sep) else "/"
    denom_str <- .pretty_print_sequence(denominator, prod_sep, neg_power, plot_sep)
    if (!neg_power && length(unique(denominator)) > 1)
      denom_str <- paste0("(", denom_str, ")")
  }

  if (length(num_str) == 0)
    denom_str
  else if (length(denom_str) == 0)
    num_str
  else
    paste(num_str, denom_str, sep = sep)
}

.simplify_units <- function(value, sym_units) {
  # from R >= 3.5
  isFALSE <- function(x) is.logical(x) && length(x) == 1L && !is.na(x) && !x

  if (isFALSE(.units.simplify())) {
    value = unclass(value)
    units(value) = sym_units
    return(value)
  }

  # This is just a brute force implementation that takes each element in the
  # numerator and tries to find a value in the denominator that can be converted
  # to the same unit. It modifies "value" to rescale the nominator to the denominator
  # before removing matching units.

  drop_ones = function(u) u[ u != "1" ]
  class(value) <- "units"

  new_numerator <- drop_ones(sym_units$numerator)
  new_denominator <- drop_ones(sym_units$denominator)
  delete_num <- c()
  for (i in seq_along(new_numerator)) {
    str1 <- new_numerator[i]

    for (j in seq_along(new_denominator)) {
      str2 <- new_denominator[j]

      if (ud_are_same(str1, str2)) {
        attr(value, "units") <- units(as_units(str1))
        units(value) <- str2
        delete_num <- c(delete_num, i)
        new_denominator <- new_denominator[-j]
        break
      }

    }
  }
  if (length(delete_num) > 0)
    new_numerator <- new_numerator[-delete_num]

  as_units(drop_units(value), .symbolic_units(new_numerator, new_denominator))
}

#' Convert units to their base units
#'
#' Convert the units of a \code{units} object to their base units, as defined by
#' the udunits database (SI units).
#'
#' @param x object of class \code{units}.
#' @param simplify logical; if TRUE (default), the resulting units are simplified.
#' @param keep_fraction logical; if TRUE (default), the result is kept as a fraction.
#'
#' @return object of class \code{units} with units converted to base units.
#' @export
#'
#' @examples
#' x <- set_units(32, mJ/g)
#' convert_to_base(x)
#' convert_to_base(x, keep_fraction=FALSE)
#' convert_to_base(x, simplify=FALSE)
#' convert_to_base(x, simplify=FALSE, keep_fraction=FALSE)
convert_to_base <- function(x, simplify = TRUE, keep_fraction = TRUE) {
  stopifnot(inherits(x, "units"))

  u_strBase <- function(u_str, simplify) {
    u_new <- ud_parse(u_str, names=FALSE, definition=TRUE, ascii=TRUE)
    u_new <- strsplit(x = u_new, split = " @ ")[[1]][1]
    u_new <- strsplit(x = u_new, split = " ")[[1]]
    u_new <- u_new[length(u_new)]

    if (simplify)
      u_new <- ud_parse(u_new, names=FALSE, definition=FALSE, ascii=TRUE)
    gsub(".", " ", u_new, fixed = TRUE)
  }

  u <- vapply(units(x), paste0, character(1L), collapse = "*", recycle0=TRUE)
  u[u == ""] <- "1"

  u["numerator"]   <- sprintf("(%s)", u["numerator"])
  u["denominator"] <- sprintf("(%s)", u["denominator"])

  if (!keep_fraction) u <- paste(u, collapse = "/")

  u_base <- vapply(u, u_strBase, character(1L), simplify = simplify)

  if (keep_fraction) {
    is_unitless <- u_base == "1"

    u_base["numerator"]   <- sprintf("(%s)", u_base["numerator"])
    u_base["denominator"] <- sprintf("(%s)-1", u_base["denominator"])

    u_base <- paste(u_base[!is_unitless], collapse = " ")
  }

  set_units(x, u_base, mode = "standard")
}

test_that("we can make symbolic units", {
  m <- units:::.symbolic_units("m")
  expect_true(inherits(m, "symbolic_units"))
  expect_equal(as.character(m), "m")
})

test_that("we can multiply and divide symbolic units", {
  m <- units:::.symbolic_units("m")
  s <- units:::.symbolic_units("s")
  expect_equal(as.character(units(units:::.multiply_symbolic_units(1, m, m))), "m^2")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, m, s))), "m/s")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.divide_symbolic_units(1, m, s)), s))), "m/s^2")
})

test_that("we can simplify basic units", {
  m <- units:::.symbolic_units("m")
  s <- units:::.symbolic_units("s")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, m, m))), units_options("unitless_symbol"))
  opt = units_options(unitless_symbol = "-")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, m, m))), "-")
  units_options(unitless_symbol = opt$unitless_symbol)
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, m, m))), units_options("unitless_symbol"))
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.divide_symbolic_units(1, m, s)), m))), "1/s")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.multiply_symbolic_units(1, m, m)), m))), "m")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.divide_symbolic_units(1, s, m)), s))), "1/m")
})

test_that("we can simplify basic units", {
  m <- units:::.symbolic_units("m")
  s <- units:::.symbolic_units("s")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, m, m))), units_options("unitless_symbol"))
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.divide_symbolic_units(1, m, s)), m))), "1/s")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.multiply_symbolic_units(1, m, m)), m))), "m")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.divide_symbolic_units(1, s, m)), s))), "1/m")
})

test_that("we can simplify basic units with conversion", {
  m <- units:::.symbolic_units("m")
  s <- units:::.symbolic_units("s")
  km <- units:::.symbolic_units("km")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, m, m))), units_options("unitless_symbol"))
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, m, km))), units_options("unitless_symbol"))
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.divide_symbolic_units(1, m, s)), km))), "1/s")
  expect_equal(as.character(units(units:::.divide_symbolic_units(1, units(units:::.multiply_symbolic_units(1, m, m)), km))), "m")
})

test_that("we can compare units", {
  g <- as_units("g")
  m <- as_units("m")
  s <- as_units("s")
  d <- as_units("day")
  e1 <- 1:10 * g * m / s / d
  e2 <- 1:10 * m / s
  expect_equal(units:::.same_units(units(e1),units(e1)), TRUE)
  expect_equal(units:::.same_units(units(e1),units(e2)), FALSE)
  expect_equal(units(e1) == units(e1), TRUE)
  expect_equal(units(e1) != units(e1), FALSE)
  expect_equal(units(e1) == units(e2), FALSE)
  expect_equal(units(e1) != units(e2), TRUE)
})

test_that("symbolic_unit works", {
  expect_error(units:::symbolic_unit("qqq"))
  expect_silent(units:::symbolic_unit("m"))
})

#' set_units
#'
#' A pipe friendly version of \code{units<-}
#'
#' @param x a numeric to be assigned units, or a units object to have units
#'   converted
#'
#' @param value a \code{units} object, or something coercible to one with
#'   \code{as_units}. Depending on \code{mode}, the unit is constructed from the
#'   supplied bare expression or from the supplied value via standard
#'   evaluation.
#'
#' @param ... passed on to \code{as_units}
#'
#' @param mode if \code{"symbols"} (the default), then unit is constructed from
#'   the expression supplied. Otherwise, if\code{mode = "standard"},
#'   standard evaluation is used for the supplied value This argument can be set
#'   via a global option \code{units_options(set_units_mode = "standard")}
#'
#' @export
#' @rdname set_units
#' @seealso \code{\link{as_units}}
set_units <- function(x, value, ..., mode = units_options("set_units_mode"))
  UseMethod("set_units")

#' @export
set_units.numeric <- function(x, value, ..., mode = units_options("set_units_mode")) {
  
  if (missing(value))
    value <- unitless
  else if (mode == "symbols") {
    value <- substitute(value)
    
    if(is.numeric(value) && !identical(value, 1) && !identical(value, 1L))
      stop("The only valid number defining a unit is '1', signifying a unitless unit")
  }
  
  units(x) <- as_units(value, ...)
  x
}

#' @export
set_units.logical <- set_units.numeric

#' @export
set_units.units <- set_units.numeric

test_that("we can convert from udunits time/date to R", {
  #skip("set_units doesn't accept strings")
  # rewrite this to 
	expect_equal(as.Date("1999-10-31"), 
	             as.Date(as_units("days since 1999-10-30")))
	expect_equal(as.POSIXct("1999-10-31 01:00:00", tz = "UTC"), 
		as.POSIXct(set_units(25, as_units("hours since 1999-10-30"), mode = "standard")))
})

test_that("we can convert from R's time/date to udunits", {
  #skip("set_units doesn't accept strings")
	expect_equal(
		as_units(as.Date("1970-01-10")),
		set_units(9, as_units("days since 1970-01-01"), mode = "standard"))
	expect_equal(
		as_units(as.Date("1999-10-31"), "days since 1999-10-30"),
		set_units(1, as_units("days since 1999-10-30"), mode = "standard"))
	expect_equal(
		as_units(as.POSIXct("1970-01-01 01:00:00", tz = "UTC")), 
		set_units(3600, as_units("seconds since 1970-01-01 00:00:00 +00:00"), mode = "standard"))
	expect_equal(
		as_units(as.POSIXct("1999-10-31 01:00:00", tz = "UTC"), "hours since 1999-10-30"), 
		set_units(25, as_units("hours since 1999-10-30"), mode = "standard"))
})

library(units)
library(testthat)

ref = set_units(1, m)

# character input
expect_equal(set_units(1, "m", mode = "standard"), ref)
# units input
expect_equal(set_units(1, as_units("m"), mode = "standard"), ref)
# symbolic_units input
expect_equal(set_units(1, units(ref), mode = "standard"), ref)

set_units(1, m/s)

# use object from parent frame, non-ambiguous:
foo = ref
expect_equal(set_units(1, foo, mode = "standard"), ref)
foo = "m"
expect_equal(set_units(1, foo, mode = "standard"), ref)
foo = as_units("m")
expect_equal(set_units(1, foo, mode = "standard"), ref)

# ambiguous, should warn:
m = "u"
set_units(1, m) # m by NSE
m = as_units("u")
set_units(1, m, mode = "standard") # u by units
set_units(1, m) # m by NSE

degree_C = "m"
set_units(1, degree_C)

degree_C = as_units("m")

set_units(1, degree_C)

# OK:
set_units(1, "bar", mode = "standard")
(tenbar <- set_units(10, bar))
(set_units(1, units(tenbar), mode = "standard"))
(set_units(1, tenbar, mode = "standard"))

set_units(10, uu <- degree_C, mode = "standard")

test_that("pillar methods are available for units objects", {
  skip_if_not_installed("pillar")

  x = set_units(1:3, km)
  m = c(x, set_units(4:6, g), allow_mixed = TRUE)

  expect_equal(unclass(pillar::type_sum(x)), "[km]")
  expect_s3_class(pillar::type_sum(x), "type_sum_units")
  expect_equal(pillar::type_sum(m), "mixed_units")

  expect_snapshot({
    pillar::pillar(x[1])
    pillar::pillar(m[1])
  })
})

test_that("units have coercion methods", {
  skip_if_not_installed("vctrs", "0.3.1")

  x = set_units(1:3, "cm")
  y = set_units(4.0, "m")
  z = set_units(10, "celsius")

  expect_error(vctrs::vec_ptype_common(y, x, z), class = "vctrs_error_incompatible_type")
  expect_error(vctrs::vec_cast_common(y, x, z), class = "vctrs_error_incompatible_type")

  expect_identical(vctrs::vec_ptype_common(x, y, x), set_units(double(), "cm"))
  expect_identical(vctrs::vec_ptype_common(x, x), set_units(integer(), "cm"))
  expect_identical(vctrs::vec_ptype_common(y, x, x), set_units(double(), "m"))

  expect_identical(
    vctrs::vec_cast_common(x, y),
    list(set_units(c(1, 2, 3), "cm"), set_units(400, "cm"))
  )
  expect_identical(
    vctrs::vec_cast_common(y, x),
    list(set_units(4, "m"), set_units(c(0.01, 0.02, 0.03), "m"))
  )
})

test_that("can combine units vectors", {
  skip_if_not_installed("vctrs", "0.3.1")
  skip_if_not_installed("dplyr", "1.0.0")

  x <- set_units(1:3, "cm")
  y <- set_units(4, "m")

  exp = set_units(c(1, 2, 3, 400), "cm")
  expect_identical(vctrs::vec_c(x, y), exp)

  # Recursive case
  df1 = dplyr::tibble(x = dplyr::tibble(x = x))
  df2 = dplyr::tibble(x = dplyr::tibble(x = y))
  df_exp = dplyr::tibble(x = dplyr::tibble(x = exp))
  expect_identical(vctrs::vec_c(df1, df2), df_exp)
})

test_that("can slice units vectors", {
  skip_if_not_installed("vctrs", "0.3.1")
  skip_if_not_installed("dplyr", "1.0.0")

  x = set_units(1:3, "cm")
  exp = list(set_units(1L, "cm"), set_units(2L, "cm"), set_units(3L, "cm"))
  expect_identical(vctrs::vec_chop(x), exp)

  # Recursive case
  df = dplyr::tibble(dplyr::tibble(x = x))
  exp = list(
    dplyr::tibble(x = set_units(1L, "cm")),
    dplyr::tibble(x = set_units(2L, "cm")),
    dplyr::tibble(x = set_units(3L, "cm"))
  )
  expect_identical(vctrs::vec_chop(df), exp)
})

test_that("split-apply-combine with dplyr and base agree", {
  skip_if_not_installed("vctrs", "0.3.1")
  skip_if_not_installed("dplyr", "1.0.0")

  `%>%` <- dplyr::`%>%`
  iris2 <- iris
  for (i in 1:4)
    units(iris2[,i]) <- "cm"

  out <- iris2 %>%
    dplyr::group_by(Species) %>%
    dplyr::summarise(dplyr::across(where(is.numeric), mean))

  # Transform to list of lists
  out <- vctrs::vec_chop(out[2:5]) %>%
    stats::setNames(out$Species) %>%
    lapply(as.list)

  exp <- lapply(split(iris2[1:4], iris2$Species), lapply, mean)
  expect_equal(out, exp)
})

test_that("storage mode is uniformly set to double (see #324)", {
  skip_if_not_installed("vctrs", "0.3.1")

  x <- set_units(1:10, cm)
  br <- set_units(2:4, `in`)
  expect_silent(vctrs::vec_cast_common(x, br))
})

test_that("keep_units restores units", {
  x <- set_units(1:5, m)

  expect_identical(x, keep_units(drop_units, x))
})

test_that("keep_units warns when no units are provided", {
  x <- 1:5

  expect_warning(keep_units(sum, x))

  expect_identical(suppressWarnings(keep_units(sum, x)), sum(x))
})

test_that("keep_units sets user-provided units", {
  rate <- set_units(3, "1/min")
  x <- keep_units(rexp, 3, rate, unit=units(1/rate))
  expect_identical(units(x), units(1/rate))

})

# Inside the group generic functions we do have .Generic even if the diagnostics
# think we do not.
# !diagnostics suppress=.Generic

#' S3 Ops Group Generic Functions for units objects
#'
#' Ops functions for units objects, including comparison, product and divide,
#' add, subtract.
#'
#' Users are advised against performing arithmetical operations with
#' temperatures in different units. The \pkg{units} package ensure that results
#' 1) are arithmetically correct, and 2) satisfy dimensional analysis, but could
#' never ensure that results are physically meaningful. Temperature units are
#' special because there is an absolute unit, Kelvin, and relative ones, Celsius
#' and Fahrenheit degrees. Arithmetic operations between them are meaningless
#' from the physical standpoint. Users are thus advised to convert all
#' temperatures to Kelvin before operating.
#'
#' @param e1 object of class \code{units},
#'        or something that can be coerced to it by \code{as_units(e1)}
#' @param e2 object of class \code{units},
#'        or something that can be coerced to it by \code{as_units(e2)},
#'        or in case of power a number (integer n or 1/n)
#'
#' @return object of class \code{units}
#' @export
#'
#' @examples
#' a <- set_units(1:3, m/s)
#' b <- set_units(1:3, m/s)
#' a + b
#' a * b
#' a / b
#' a <- as_units("kg m-3")
#' b <- set_units(1, kg/m/m/m)
#' a + b
#' a = set_units(1:5, m)
#' a %/% a
#' a %/% set_units(2)
#' set_units(1:5, m^2) %/% set_units(2, m)
#' a %% a
#' a %% set_units(2)
Ops.units <- function(e1, e2) {
  if (missing(e2))
    return(NextMethod())

  cmp <- .Generic %in% c("==", "!=", "<", ">", "<=", ">=") # comparison-type
  div <- .Generic %in% c("/", "%/%")                       # division-type
  mul <- .Generic %in% "*"                                 # multiplication-only
  prd <- .Generic %in% c("*", "/", "%/%", "%%")            # product-type
  pw  <- .Generic %in% c("**", "^")                        # power-type
  mod <- .Generic %in% c("%/%", "%%")                      # modulo-type
  pm  <- .Generic %in% c("+", "-")                         # addition-type

  if (! any(cmp, pm, prd, pw, mod))
    stop(paste("operation", .Generic, "not allowed"))

  e1_inherits_units <- inherits(e1, "units")
  e2_inherits_units <- inherits(e2, "units")
  both_inherit_units <- e1_inherits_units && e2_inherits_units

  if ((pm || cmp) && !both_inherit_units)
    stop("both operands of the expression should be \"units\" objects") # nocov

  if (cmp) {
    if (!ud_are_convertible(units(e1), units(e2)))
      stop("cannot compare non-convertible units")

    if (length(e1) >= length(e2)) {
      e1 <- ud_compare(e1, e2, as.character(units(e1)), as.character(units(e2)))
      attr <- attributes(e2)
      e2 <- rep(0, length(e2))
      attributes(e2) <- attr
    } else {
      e2 <- ud_compare(e2, e1, as.character(units(e2)), as.character(units(e1)))
      attr <- attributes(e1)
      e1 <- rep(0, length(e1))
      attributes(e1) <- attr
    }
    return(NextMethod())
  }

  if (!isFALSE(.units.simplify())) { # see #355
    # Special cases for identical unit division and inverse unit multiplication
    # which may not be otherwise divisible by udunits (see #310)
    identical_units <-
      both_inherit_units &&
      identical(units(e1), units(e2))

    inverse_units <-
      both_inherit_units &&
      identical(units(e1)$numerator, units(e2)$denominator) &&
      identical(units(e1)$denominator, units(e2)$numerator)

    if ((div && identical_units) || (mul && inverse_units)) {
      e1 <- drop_units(e1)
      e2 <- drop_units(e2)
      return(set_units(NextMethod(), 1))
    }
  }

  if (mod) {
    div <- e1 / e2
    int <- round(div)
    if (.Generic == "%/%") {
      return(int)
    } else {
      return(e1 - int*e2)
    }
  } else if (prd) {
    if (!e1_inherits_units)
      e1 <- set_units(e1, 1) # TODO: or warn?

    if (!e2_inherits_units)
      e2 <- set_units(e2, 1) # TODO: or warn?

    ve1 <- unclass(e1) ; ue1 <- units(e1)
    ve2 <- unclass(e2) ; ue2 <- units(e2)

    if (.Generic == "*") {
      numerator <- sort(c(ue1$numerator, ue2$numerator))
      denominator <- sort(c(ue1$denominator, ue2$denominator))
    } else {
      numerator <- sort(c(ue1$numerator, ue2$denominator))
      denominator <- sort(c(ue1$denominator, ue2$numerator))
    }
    return(.simplify_units(NextMethod(), .symbolic_units(numerator, denominator)))

  } else if (pw) {
    if (e2_inherits_units) {
      if (e1_inherits_units && identical(units(e1), units(as_units(1)))) {
        e1 <- drop_units(e1)
      } else if (e1_inherits_units) {
        stop("power operation only allowed with numeric power")
      }

      # code to manage things like exp(log(...)) and 10^log10(...) follows
      # this is not supported in udunits2, so we are on our own
      u <- units(e2)
      invalid_lengths <- length(u$numerator) > 1 || length(u$denominator)
      if (u == unitless || invalid_lengths || !grepl("\\(re", u$numerator))
        stop("power operation only allowed with logarithmic unit")

      sp <- strsplit(u$numerator, "\\(re ")[[1]]

      # only works for the same base
      type <- strsplit(sp[1], " ")[[1]]
      fc <- if (length(type) == 1) 1 else as.numeric(type[1])
      base <- switch(type[length(type)], lb=2, lg=10, ln=exp(1/fc))
      if (!isTRUE(all.equal(e1, base)))
        stop("wrong base in power operation")

      # this is the next unit after "undoing" the outer logarithm
      nxt_u <- paste(sp[-1], collapse="(re ")   # next unit
      nxt_u <- substr(nxt_u, 1, nchar(nxt_u)-1) # remove last parenthesis

      if (length(sp) > 2) { # another logarithm!
        mult <- 1
        attr(e2, "units")$numerator <- sub("^1 ", "", nxt_u)
      } else {
        nxt_u <- strsplit(nxt_u, " ")[[1]]
        mult <- as.numeric(nxt_u[1])
        nxt_u <- as_units(nxt_u[2])
        attr(e2, "units") <- units(nxt_u)
      }
      return(mult * NextMethod())
    }

    if (length(e2) > 1L) {
      if (length(unique(e2)) == 1) # all identical
        e2 = e2[1]
      else # return mixed units:
        return(.as.mixed_units(mapply("^", e1, e2, SIMPLIFY=FALSE)))
    }

    # we repeat each unit the number of times given by e2. They are already
    # sorted so they will remain sorted. We need to flip numerator and denominator
    # when the power is negative and we have a special case when it is zero where
    # units should be removed.
    if (e2 == 0) {
      u <- units(as_units(1))
    } else {
      if (any((table(units(e1)$denominator)*e2) %% 1 != 0) ||
          any((table(units(e1)$numerator)*e2)   %% 1 != 0))
            stop("powers not divisible") # work on wording
      if (e2 > 0)
        u <- .symbolic_units(
          rep(unique(units(e1)$numerator),table(units(e1)$numerator)*e2),
          rep(unique(units(e1)$denominator),table(units(e1)$denominator)*e2))
      else
        u <- .symbolic_units(
          rep(unique(units(e1)$denominator),table(units(e1)$denominator)*abs(e2)),
          rep(unique(units(e1)$numerator),table(units(e1)$numerator)*abs(e2)))
    }
  } else { # pm:
    units(e2) <- units(e1)
    u <- units(e1)
  }

  .as.units(NextMethod(), u)
}

#' #' matrix multiplication
#' #' @name matmult
#' #' @param x numeric matrix or vector
#' #' @param y numeric matrix or vector
#' #' @export
#' #' @details see \code{"\link[base]{\%*\%}"} for the base function, reimplemented
#' #'   as default method
#' `%*%` = function(x, y) UseMethod("%*%")
#'
#' #' @name matmult
#' #' @export
#' `%*%.default` = function(x, y) {
#' 	if (inherits(y, "units"))
#' 		`%*%.units`(x, y)
#' 	else
#' 		base::`%*%`(x, y)
#' }
#'
#' #' @name matmult
#' #' @export
#' #' @examples
#' #' a = set_units(1:5, m)
#' #' a %*% a
#' #' a %*% t(a)
#' #' a %*% set_units(1:5, 1)
#' #' set_units(1:5, 1) %*% a
#' `%*%.units` = function(x, y) {
#' 	ret = `%*%.default`(unclass(x), unclass(y))
#' 	units(ret) = .multiply_symbolic_units(1, units(x), units(y))
#' 	ret
#' }

#' @export
`[<-.units` <- function(x, ..., value) {
  units(value) <- units(x)
  NextMethod()
}

#' @export
`[[<-.units` <- `[<-.units`

#' @export
c.units <- function(..., recursive = FALSE, allow_mixed = units_options("allow_mixed")) {
  args <- list(...)
  args[sapply(args, is.null)] <- NULL # remove NULLs
  u <- units(args[[1]])

  if (length(args) == 1)
    return(.as.units(NextMethod(), u))

  dup <- vapply(args, function(x) identical(units(x), u), TRUE)

  if (all(dup))
    return(.as.units(do.call(c, lapply(args, drop_units)), u))

  if (.units_are_convertible(args[!dup], u)) {
    args[!dup] <- lapply(args[!dup], set_units, u, mode = "standard")
    return(.as.units(do.call(c, lapply(args, drop_units)), u))
  }

  if (allow_mixed)
    return(do.call(c, lapply(args, mixed_units)))

  stop("units are not convertible, and cannot be mixed; try setting units_options(allow_mixed = TRUE)?")
}

.units_are_convertible = function(x, u) {
  u <- ud_char(u)
	for (i in seq_along(x))
		if (! ud_are_convertible(units(x[[i]]), u))
			return(FALSE)
	TRUE
}

#' @export
diff.units = function(x, ...) {
  u = units(x)
  # units(x) = u will not work here, as units(x) is NULL!
  .as.units(NextMethod(), u)
}

#' @export
rep.units = function(x, ...) {
  u = units(x)
  .as.units(NextMethod(), u)
}


#' deparse unit to string in product power form (e.g. km m-2 s-1)
#'
#' deparse unit to string in product power form (e.g. km m-2 s-1)
#' @param x object of class units
#' @return length one character vector
#' @examples
#' u = as_units("kg m-2 s-1")
#' u
#' deparse_unit(u)
#' @export
deparse_unit = function(x) {
  stopifnot(inherits(x, "units"))
  as.character(units(x), neg_power=TRUE, prod_sep=" ")
}
# This should perhaps be an option in format.symbolic_units

#' @method all.equal units
#' @export
all.equal.units = function(target, current, ...) {
  if (inherits(current, "units")) {
    units(current) <- units(target)
    target <- drop_units(target)
    current <- drop_units(current)
  }
  NextMethod()
}

#' seq method for units objects
#' @param from see \link[base]{seq}
#' @param to see \link[base]{seq}
#' @param by see \link[base]{seq}
#' @param length.out see \link[base]{seq}
#' @param along.with see \link[base]{seq}
#' @param ... see \link[base]{seq}
#' @details arguments with units are converted to have units of the first argument (which is either \code{from} or \code{to})
#' @export
#' @examples
#' seq(to = set_units(10, m), by = set_units(1, m), length.out = 5)
#' seq(set_units(10, m), by = set_units(1, m), length.out = 5)
#' seq(set_units(10, m), set_units(19, m))
#' seq(set_units(10, m), set_units(.1, km), set_units(10000, mm))
seq.units = function(from, to, by = ((to - from)/(length.out - 1)),
         length.out = NULL, along.with = NULL, ...) {
  mf = missing(from)
  mt = missing(to)
  uuu = if (mf)
      units(to)
  	else
      units(from)
  if (! mf)
    from = as.numeric(from)
  if (! mt)
  	to = as.numeric(set_units(to, uuu, mode = "standard"))
  if (! missing(by))
    by = as.numeric(set_units(by, uuu, mode = "standard"))
  set_units(NextMethod(), uuu, mode = "standard")
}

#' @export
str.units = function(object, ...) {
  gr <- units_options("group")
  unit_string <- paste0(gr[1], as.character(attr(object, "units")), gr[2])
  cat(paste0(" Units: ", unit_string))
  str(drop_units(object), ...)
}

#' @export
duplicated.units <- function(x, incomparables=FALSE, ...) {
  if (is.null(dim(x)))
    NextMethod() else duplicated.array(x, incomparables, ...)
}

#' @export
anyDuplicated.units <- function(x, incomparables=FALSE, ...) {
  if (is.null(dim(x)))
    NextMethod() else anyDuplicated.array(x, incomparables, ...)
}

#' @export
unique.units <- function(x, incomparables = FALSE, ...) {
  xx <- if (is.null(dim(x)))
    NextMethod() else unique.array(x, incomparables, ...)
  .as.units(xx, units(x))
}

#' Combine R Objects by Rows or Columns
#'
#' S3 methods for \code{units} objects (see \code{\link[base]{cbind}}).
#'
#' @inheritParams base::cbind
#' @name cbind.units
#'
#' @examples
#' x <- set_units(1, m/s)
#' y <- set_units(1:3, m/s)
#' z <- set_units(8:10, m/s)
#' (m <- cbind(x, y)) # the '1' (= shorter vector) is recycled
#' (m <- cbind(m, z)[, c(1, 3, 2)]) # insert a column
#' (m <- rbind(m, z)) # insert a row
#' @export
cbind.units <- function(..., deparse.level = 1) {
  dots <- .deparse(list(...), substitute(list(...)), deparse.level)
  units_first_arg <- units(dots[[1]])
  class_first_arg <- class(dots[[1]])
  dots <- lapply(dots, function(x) {
    dots_unified <- set_units(x, units_first_arg, mode = "standard")
    ret <- drop_units(dots_unified)
    return(ret)
  })
  call <- as.character(match.call()[[1]])
  value <- do.call(call, c(dots, deparse.level=deparse.level))
  attr(value, "units") <- units_first_arg
  class(value) <- class_first_arg
  return(value)
}

.deparse <- function(dots, symarg, deparse.level) {
  deparse.level <- as.integer(deparse.level)
  if (identical(deparse.level, -1L)) deparse.level <- 0L # R Core's hack
  stopifnot(0 <= deparse.level, deparse.level <= 2)

  nm <- c( ## 0:
    function(i) NULL,
    ## 1:
    function(i) if(is.symbol(s <- symarg[[i]])) deparse(s) else NULL,
    ## 2:
    function(i) deparse(symarg[[i]])[[1L]])[[ 1L + deparse.level ]]
  Nms <- function(i) { if(!is.null(s <- names(symarg)[i]) && nzchar(s)) s else nm(i) }

  symarg <- as.list(symarg)[-1L]
  dnames <- sapply(seq_along(dots), Nms)
  if (!all(sapply(dnames, is.null)))
    names(dots) <- dnames
  dots
}

#' @rdname cbind.units
#' @export
rbind.units <- cbind.units

#' Defunct functions in \pkg{units}
#'
#' These functions are no longer available.
#'
#' \itemize{
#'  \item \code{ud_units}: Use \code{\link{as_units}} instead.
#'  \item \code{as.units}: Use \code{\link{as_units}} instead.
#'  \item \code{make_unit}: Use \code{\link{as_units}} instead.
#'  \item \code{parse_unit}: Use \code{\link{as_units}} instead.
#'  \item \code{as_cf}: Use \code{\link{deparse_unit}} instead.
#'  \item \code{install_symbolic_unit}: Use \code{\link{install_unit}} instead.
#'  \item \code{remove_symbolic_unit}: Use \code{\link{remove_unit}} instead.
#'  \item \code{install_conversion_constant}: Use \code{\link{install_unit}} instead.
#'  \item \code{install_conversion_offset}: Use \code{\link{install_unit}} instead.
#' }
#'
#' @name units-defunct
NULL

#' Deprecated functions
#' 
#' The following functions are deprecated and will be removed in a future release.
#' 
#' @param chr length 1 character string
#' 
#' @export
#' @rdname deprecated
make_unit <- function(chr) {
  .Deprecated("as_units", msg = "make_unit() is deprecated. Please use as_units()")
  if (!ud_is_parseable(chr))
    warning(paste(sQuote(chr), "is not a valid unit symbol recognized by udunits"), call.=FALSE)
  as_units.character(chr, force_single_symbol = TRUE, check_is_valid = FALSE)
}

#' @export
#' @rdname deprecated
parse_unit <- function(chr) {
  .Deprecated("as_units", msg = "parse_unit() is deprecated. Please use as_units()")
  as_units.character(chr, implicit_exponents = TRUE)
}

#' @export
#' @name deprecated
#' @param x a numeric
#' @param value a units object, by default, unitless
as.units <- function(x, value = unitless) {
  .Deprecated("as_units")    # nocov
  as_units(x, value = value) # nocov
}

# Inside the group generic functions we do have .Generic even if the diagnostics
# think we do not.
# !diagnostics suppress=.Generic

#' Mathematical operations for units objects
#'
#' @param x object of class units
#' @param ... parameters passed on to the Math functions
#'
#' @details Logarithms receive a special treatment by the underlying \pkg{udunits2}
#' library. If a natural logarithm is applied to some \code{unit}, the result is
#' \code{ln(re 1 unit)}, which means \emph{natural logarithm referenced to
#' \code{1 unit}}. For base 2 and base 10 logarithms, the output \code{lb(...)}
#' and \code{lg(...)} respectively instead of \code{ln(...)}.
#'
#' This is particularly important for some units that are typically expressed in
#' a logarithmic scale (i.e., \emph{bels}, or, more commonly, \emph{decibels}),
#' such as Watts or Volts. For some of these units, the default \pkg{udunits2}
#' database contains aliases: e.g., \code{BW} (bel-Watts) is an alias of
#' \code{lg(re 1 W)}; \code{Bm} (bel-milliWatts) is an alias of
#' \code{lg(re 0.001 W)}; \code{BV} is an alias of \code{lg(re 1 V)} (bel-Volts),
#' and so on and so forth (see the output of \code{valid_udunits()} for further
#' reference).
#'
#' Additionally, the \pkg{units} package defines \code{B}, the \emph{bel}, by
#' default (because it is not defined by \pkg{udunits2}) as an alias of
#' \code{lg(re 1)}, unless a user-provided XML database already contains a
#' definition of \code{B}, or the \code{define_bel} option is set to \code{FALSE}
#' (see \code{help(units_options)}).
#'
#' @export
#'
#' @examples
#' # roundings, cummulative functions
#' x <- set_units(sqrt(1:10), m/s)
#' signif(x, 2)
#' cumsum(x)
#'
#' # trigonometry
#' sin(x) # not meaningful
#' x <- set_units(sqrt(1:10), rad)
#' sin(x)
#' cos(x)
#' x <- set_units(seq(0, 1, 0.1), 1)
#' asin(x)
#' acos(x)
#'
#' # logarithms
#' x <- set_units(sqrt(1:10), W)
#' log(x) # base exp(1)
#' log(x, base = 3)
#' log2(x)
#' log10(x)
#' set_units(x, dBW) # decibel-watts
#' set_units(x, dBm) # decibel-milliwatts
Math.units = function(x, ...) {
  if (.Generic == "sqrt")
    return(x^0.5)

  if (.Generic == "exp")
    return(exp(1)^x)

  if (.Generic == "expm1") {
    out <- exp(1)^x
    mult <- drop_units(out) / exp(1)^drop_units(x)
    return(out - set_units(mult, units(out), mode="standard"))
  }

  if (.Generic == "sign")
    return(as.numeric(NextMethod()))

  OK <- switch(.Generic, "abs" = , "sign" = , "floor" = , "ceiling" = , "log" = ,
               "log1p" =, "trunc" = , "round" = , "signif" = , "cumsum" = ,
               "cummax" = , "cummin" = TRUE, FALSE)

  rad = units(as_units("rad"))
  deg = units(as_units("degree"))
  if (!OK && (units(x) == rad || units(x) == deg || units(x) == unitless)) {
    OK <- switch(.Generic, sin =, cos =, tan =, sinpi =, cospi =, tanpi = TRUE, FALSE)
    if (OK) {
      units(x) <- "rad" # convert deg -> rad
      x <- set_units(x) # result has unit 1
    }
  }
  if (!OK && units(x) == unitless) {
    OK <- switch(.Generic, asin =, acos =, atan = TRUE, FALSE)
    if (OK)
      units(x) <- "rad" # unit of the answer (also unitless)
  }

  if (!OK) {
    warning(paste("Operation", .Generic, "not meaningful for units"))
    x <- drop_units(x)
    NextMethod()
  } else {
    if (.Generic %in% c("log", "log1p")) {
      base <- if (missing(...)) exp(1) else c(...)[1]
      uptr <- R_ut_parse(as.character(units(x)))
      u <- R_ut_format(R_ut_log(uptr, base), ascii=TRUE)
      .as.units(NextMethod(), units(symbolic_unit(u)))
    } else
      .as.units(NextMethod(), units(x))
  }
}

#' @export
#' @method log10 units
log10.units <- function(x) log(x, 10)

#' @export
#' @method log2 units
log2.units <- function(x) log(x, 2)

#' Apply a function keeping units
#'
#' Helper function to apply a function to a \code{units} object and then restore
#' the original units.
#'
#' Provided for incompatible functions that do not preserve units. The user is
#' responsible for ensuring the correctness of the output.
#'
#' If \code{x} is not a \code{units} object
#' and \code{unit} is not provided by the user,
#' a warning is issued, and the output will also have no units
#' (see examples).
#'
#' @param FUN the function to be applied.
#' @param x first argument of \code{FUN}, of class \code{units}.
#' @param ... optional arguments to \code{FUN}.
#' @param unit symbolic unit to restore after \code{FUN}.
#'
#' @return An object of class \code{units}.
#'
#' @examples
#' x <- set_units(1:5, m)
#' keep_units(drop_units, x)
#'
#' # An example use case is with random number generating functions:
#' mu <- as_units(10, "years")
#' keep_units(rnorm, n = 1, x = mu)
#'
#' # units can be directly specified if needed; for example, with
#' # `rexp()`, the units of the rate parameter are the inverse of
#' # the units of the output:
#' rate <- as_units(3, "1/year")
#' keep_units(rexp, n = 1, x = rate, unit = units(1/rate))
#'
#' # if `x` does not actually have units, a warning is issued,
#' # and the output has no units:
#' rate2 <- 3
#' keep_units(rexp, n = 1, x = rate2)
#'
#' @export
keep_units <- function(FUN, x, ..., unit=units(x)) {
  if (inherits(try(unit, silent = TRUE), "symbolic_units")) {
    set_units(do.call(FUN, list(x, ...)), unit, mode = "standard")
  } else {
    warning("wrong `unit` specification.")
    do.call(FUN, list(x, ...))
  }
}

dfapply <- function(X, FUN, ...) {
  attrs <- attributes(X)
  X <- lapply(X, FUN, ...)
  attributes(X) <- attrs
  X
}

#' Handle measurement units
#'
#' A number of functions are provided for handling unit objects.
#' \itemize{
#'     \item \code{`units<-`} and \code{units} are the basic functions to set
#'     and retrieve units.
#'     \item \code{as_units}, a generic with methods for a
#'     character string and for quoted language. Note, direct usage of this function
#'     by users is typically not necessary, as coercion via \code{as_units} is
#'     automatically done with \code{`units<-`} and \code{set_units}.
#'     \item \code{make_units}, constructs units from bare expressions.
#'     \code{make_units(m/s)} is equivalent to \code{as_units(quote(m/s))}.
#'     \item \code{set_units}, a pipe-friendly version of \code{`units<-`}. By
#'     default it operates with bare expressions, but this
#'     behavior can be disabled by a specifying \code{mode = "standard"} or setting
#'     \code{units_options(set_units_mode = "standard")}.
#'     If \code{value} is missing or set to \code{1}, the object becomes unitless.
#' }
#'
#' @param x numeric vector, or object of class \code{units}.
#' @param value object of class \code{units} or \code{symbolic_units}, or in the
#' case of \code{set_units} expression with symbols (see examples).
#'
#' @return An object of class \code{units}.
#'
#' @details
#' If \code{value} is of class \code{units} and has a value unequal to 1, this
#' value is ignored unless \code{units_options("simplifiy")} is \code{TRUE}. If
#' \code{simplify} is \code{TRUE}, \code{x} is multiplied by this value.
#'
#' @export
#' @name units
#'
#' @examples
#' x = 1:3
#' class(x)
#' units(x) <- as_units("m/s")
#' class(x)
#' y = 2:5
`units<-.numeric` <- function(x, value) {
  if(!length(value))
    return(x)

  storage.mode(x) <- "double" # issues/324

  if(!inherits(value, "units") && !inherits(value, "symbolic_units"))
    value <- as_units(value)

  if (inherits(value, "units")) {
	if (any(is.na(value)))
	  stop("a missing value for units is not allowed")
  if (isTRUE(.units.simplify()))
    x <- x * unclass(value)
	else if (any(unclass(value) != 1.0))
	  warning(paste("numeric value", unclass(value), "is ignored in unit assignment"))
    value <- units(value)
  }

  attr(x, "units") = value
  class(x) <- "units"
  x
}

#' @name units
#' @export
#'
#' @examples
#' a <- set_units(1:3, m/s)
#' units(a) <- make_units(km/h)
#' a
#' # convert to a mixed_units object:
#' units(a) <- c("m/s", "km/h", "km/h")
#' a
`units<-.units` <- function(x, value) {
  if(!length(value))
    return(drop_units(x))

  if(!inherits(value, "units") && !inherits(value, "symbolic_units")) {
    if ((is.character(value) && length(value) > 1))
      return(set_units(mixed_units(x), value))
    value <- as_units(value)
  }

  dimx = dim(x)
  if (inherits(value, "units")) {
    if (!identical(as.numeric(value), 1))
      x <- .as.units(unclass(x) * unclass(value), units(x))
    value <- units(value)
  }

  if (identical(units(x), value)) # do nothing; possibly user-defined units:
    return(x)

  str1 <- as.character(units(x))
  str2 <- as.character(value)

  if (ud_are_convertible(str1, str2))
    .as.units(ud_convert(unclass(x), str1, str2), value, dim = dimx)
  else
    stop(paste("cannot convert", units(x), "into", value), call. = FALSE)
}

#unit_ambiguous = function(value) {
#  msg = paste("ambiguous argument:", value, "is interpreted by its name, not by its value")
#  warning(msg, call. = FALSE)
#}


#' @name units
#' @export
`units<-.logical` <- function(x, value) {
  if (!all(is.na(x)))
    stop("x must be numeric, non-NA logical not supported")

  x <- as.numeric(x)
  units(x) <- value
  x
}

#' @return The \code{units} method retrieves the units attribute, which is of
#' class \code{symbolic_units}.
#'
#' @name units
#' @export
units.units <- function(x) {
  attr(x, "units")
}

#' @name units
#' @export
units.symbolic_units <- function(x) {
  x
}

#' @export
as.data.frame.units <- function(x, row.names = NULL, optional = FALSE, ...) {
	df = as.data.frame(unclass(x), row.names, optional, ...)
	if (!optional && ncol(df) == 1)
	  colnames(df) <- deparse(substitute(x))
	for (i in seq_along(df))
		units(df[[i]]) = units(x)
	df
}

#' @export
as.list.units <- function(x, ...)
  lapply(NextMethod(), set_units, units(x), mode="standard")

#' convert units object into difftime object
#'
#' @param x object of class \code{units}
#'
#' @export
#' @examples
#'
#' t1 = Sys.time()
#' t2 = t1 + 3600
#' d = t2 - t1
#' du <- as_units(d)
#' dt = as_difftime(du)
#' class(dt)
#' dt
as_difftime <- function(x) {
  stopifnot(inherits(x, "units"))
  u <- as.character(units(x))
  if (u == "s")
    as.difftime(x, units = "secs")
  else if (u == "min")
    as.difftime(x, units = "mins")
  else if (u == "h")
    as.difftime(x, units = "hours")
  else if (u == "d")
    as.difftime(x, units = "days")
  else
    stop(paste("cannot convert unit", u, "to difftime object"))
}

# #' Convert units to hms
# #'
# #' Convert units to hms
# #' @param x object of class units
# #' @param ... passed on to as.hms.difftime
# #' @return object of class hms
# #' @examples
# #' if (require(hms)) {
# #'  as.hms(1:10 * make_units(s))
# #'  as.hms(1:10 * make_units(min))
# #'  as.hms(1:10 * make_units(h))
# #'  as.hms(1:10 * make_units(d))
# #' }
# #' @export
# as.hms.units = function(x, ...) {
# 	hms::as.hms(as_difftime(x), ...)
# }


#' @export
`[.units` <- function(x, i, j, ..., drop = TRUE) {
  restore_units(NextMethod(), x)
}

#' @export
`[[.units` <- function(x, i, j, ...) {
  restore_units(NextMethod(), x)
}

#' @export
as.POSIXct.units = function (x, tz = "UTC", ...) {
	units(x) = symbolic_unit("seconds since 1970-01-01 00:00:00 +00:00")
	as.POSIXct.numeric(as.numeric(x), tz = tz, origin = as.POSIXct("1970-01-01 00:00:00", tz = "UTC"))
}

#' @method as.Date units
#' @export
as.Date.units = function (x, ...) {
	units(x) = symbolic_unit("days since 1970-01-01")
	as.Date(as.numeric(x), origin = as.Date("1970-01-01 00:00:00"))
}

#' @param ... passed on to other methods.
#' @param mode if \code{"symbols"} (the default), then unit is constructed from
#'   the expression supplied. Otherwise, if\code{mode = "standard"},
#'   standard evaluation is used for the supplied value This argument can be set
#'   via a global option \code{units_options(set_units_mode = "standard")}
#'
#' @name units
#' @export
set_units <- function(x, value, ..., mode = units_options("set_units_mode"))
  UseMethod("set_units")

#' @export
set_units.numeric <- function(x, value, ..., mode = units_options("set_units_mode")) {
  if (missing(value))
    value <- unitless
  else if (mode == "symbols") {
    value <- substitute(value)

    if(is.numeric(value) && !identical(value, 1) && !identical(value, 1L))
      stop("The only valid number defining a unit is '1', signifying a unitless unit")
  }

  units(x) <- as_units(value, ...)
  x
}

#' @export
set_units.logical <- set_units.numeric

#' @export
set_units.units <- set_units.numeric

#' Drop Units
#'
#' Drop units attribute and class.
#'
#' @param x an object with units metadata.
#'
#' @return the numeric without any units attributes, while preserving other
#' attributes like dimensions or other classes.
#'
#' @details Equivalent to \code{units(x) <- NULL}, or the pipe-friendly version
#' \code{set_units(x, NULL)}, but \code{drop_units} will fail if the object has
#' no units metadata. Use the alternatives if you want this operation to succeed
#' regardless of the object type.
#'
#' A \code{data.frame} method is also provided, which checks every column and
#' drops units if any.
#'
#' @export
#' @examples
#' x <- 1
#' y <- set_units(x, m/s)
#'
#' # this succeeds
#' drop_units(y)
#' set_units(y, NULL)
#' set_units(x, NULL)
#'
#' \dontrun{
#' # this fails
#' drop_units(x)
#' }
#'
#' df <- data.frame(x=x, y=y)
#' df
#' drop_units(df)
#'
drop_units <- function(x) UseMethod("drop_units")

#' @name drop_units
#' @export
drop_units.units <- function(x) {
  class(x) <- setdiff(class(x), "units")
  attr(x, "units") <- NULL
  x
}

#' @name drop_units
#' @export
drop_units.data.frame <- function(x) {
  for (i in seq_along(x)) {
    if (inherits(x[[i]], "units"))
      x[[i]] <- drop_units(x[[i]])
  }
  x
}

test_that("mixed units work", {
   m = c(set_units(1:3, km), set_units(4:6, g), allow_mixed = TRUE)

   # select a subset
   expect_s3_class(m[3:4], "mixed_units")
   expect_s3_class(m[3], "mixed_units")

# select a single units object:
   expect_s3_class(m[[3]], "units")

   m <- set_units(m, c(rep(c("m", "kg"), each = 3)))
   expect_s3_class(m, "mixed_units")
   units(m) = rep(c("mm", "mg"), each = 3)
   expect_s3_class(m, "mixed_units")
   # does the value get recycled?
   expect_s3_class(set_units(m[1:3], "m"), "mixed_units")

   # convert to units:
   expect_s3_class(as_units( m[1:3] ), "units")

# round-trip via units:
   m0 <- mixed_units(as_units(m[1:3]))
   expect_identical(m[1:3], m0)

# Ops using by single unit: needs to be explicitly coerced to mixed_units:
   expect_s3_class(m[1:3] * mixed_units(set_units(1, mm)), "mixed_units")
   expect_s3_class(m[1:3] / mixed_units(set_units(1, mm)), "mixed_units")
   expect_s3_class(m[1:3] + mixed_units(set_units(1, mm)), "mixed_units")
   expect_s3_class(m[1:3] - mixed_units(set_units(1, mm)), "mixed_units")
   expect_type(m[1:3] == mixed_units(set_units(1, mm)), "logical")
   expect_type(m[1:3] != mixed_units(set_units(1, mm)), "logical")
   expect_error(m[1:3] ^ mixed_units(set_units(1, mm)))

   # FIXME: Ops.mixed_units and Ops.units must be the same method
   # to avoid the warning and the error.
   # We can discriminate by switchpatching.
   expect_error(expect_warning(m[1:3] * set_units(1, mm)))

   expect_s3_class(units(m), "mixed_symbolic_units")
   expect_type(format(m), "character")
   expect_type(as.character(units(m)), "character")
   expect_equal(drop_units(m), sapply(m, as.numeric))

   units_options(allow_mixed = TRUE)
   m = c(set_units(1:3, km), set_units(4:6, g))
   expect_s3_class(m, "mixed_units")
   expect_equal(m, mixed_units(1:6, c(rep("km", 3), rep("g", 3))))
   units_options(allow_mixed = FALSE)
})

test_that("order is preserved", {
   x <- 1:10
   u <- rep(c("m", "l"), 5)
   m <- mixed_units(x, u)
   m <- set_units(m, paste0("k", u), mode = "standard")

   expect_equal(as.numeric(m), x / 1000)
})

test_that("unique.mixed_units works", {
   x <- c(set_units(c(1, 1, 2), kg), set_units(c(4, 4, 5), s), allow_mixed = TRUE)
   expect_equal(unique(x), c(set_units(c(1, 2), kg), set_units(c(4, 5), s), allow_mixed = TRUE))

   y <- c(set_units(c(1, 1, 1), m/s), set_units(c(1, 1, 1), kg/s), allow_mixed = TRUE)
   expect_equal(unique(y), c(set_units(1, m/s), set_units(1, kg/s), allow_mixed = TRUE))

   z <- c(set_units(c(1, 2), kg), set_units(c(3, 4), s), set_units(c(2, 3), kg), allow_mixed = TRUE)
   expect_equal(unique(z), c(set_units(c(1, 2), kg), set_units(c(3, 4), s), set_units(3, kg), allow_mixed = TRUE))
})

test_that("as.data.frame.mixed_units works", {
  x <- 1:3
  x_units <- set_units(x, "m")
  x_mixed <- mixed_units(x_units)

  df1 <- data.frame(x)
  df1$x_units <- x_units
  df1$x_mixed <- x_mixed
  df2 <- data.frame(x, x_units, x_mixed)

  expect_equal(df1, df2)
})

test_that("we can compare vectors with equal units", {
  x <- 1:4 * as_units("m")
  y <- 1:4 * as_units("m")
  z <- 2 * y

  expect_true(all(x == y))
  expect_true(all(x <= y))
  expect_true(all(x >= y))

  expect_true(all(x < z))
  expect_true(all(x <= z))
  expect_true(all(z > x))
  expect_true(all(z >= x))

  expect_false(any(x > y))
  expect_false(any(x != y))
  expect_true(all(x != z))

  expect_error(x == set_units(1, "kg"))
})

test_that("comparing special values gives correct results", {
  x <- c(-Inf, NaN, NA, Inf) * as_units("m")
  expect_equal(x == x, drop_units(x) == drop_units(x))
  expect_equal(x > x, drop_units(x) > drop_units(x))
  expect_equal(x < x, drop_units(x) < drop_units(x))
  expect_equal(x == rev(x), drop_units(x) == rev(drop_units(x)))
  expect_equal(x > rev(x), drop_units(x) > rev(drop_units(x)))
  expect_equal(x < rev(x), drop_units(x) < rev(drop_units(x)))
})

test_that("vectors are correctly recycled in comparisons", {
  x <- 0:3 * as_units("m")
  y <- 0:1 * as_units("m")
  res <- drop_units(x) == drop_units(y)
  expect_equal(x == y, res)
  expect_equal(y == x, res)

  y <- 0:2 * as_units("m")
  expect_warning(res <- drop_units(x) == drop_units(y))
  expect_warning(expect_equal(x == y, res))
  expect_warning(expect_equal(y == x, res))
})

test_that("aliases are correctly handled in comparisons (#339)", {
  expect_true(as_units("foot") == as_units("feet"))
  expect_true(as_units("foot") == as_units("ft"))
})

test_that("we can scale units with scalars", {
  x <- 1:4
  ux <- x * as_units("m")

  expect_equal(as.numeric(10 * ux), 10 * x)
  expect_equal(as.numeric(ux / 10), x / 10)
})

test_that("we can multiply and divide units", {
  x <- 1:4 ; y <- 5:8
  m <- x * as_units("m")
  s <- y * as_units("s")

  expect_equal(as.numeric(m * s), x * y)
  expect_equal(as.numeric(m / s), x / y)

  # FIXME: There ought to be a test that the expressions get the right units
  # but I am not entirely sure how that should be wrapped. Just checking string
  # equality would give problems if units are equivalent but needs to be converted
  # first...
})

test_that("we can take powers of units", {
  x <- 1:4
  ux <- x * as_units("m")

  expect_equal(as.numeric(ux ** 2), x ** 2)
  expect_equal(as.numeric(ux ^ 2), x ^ 2)
  expect_equal(as.character(units(ux ** 2)), "m^2")
  expect_equal(as.character(units(ux ^ 2)), "m^2")

  expect_error(ux ^ 1.3)
  expect_error(ux ^ 0.3)
  expect_error(ux ^ ux)
  # expect_error(ux ^ x) --> see below: gives mixed_units
  expect_silent(ux ^ x)

  expect_equal(as.numeric(ux ** -2), x ** -2)
  expect_equal(as.numeric(ux ^ -2), x ^ -2)
  expect_equal(as.character(units(ux ** -2)), "1/m^2")
  expect_equal(as.character(units(ux ^ -2)), "1/m^2")

  expect_equal(as.numeric(ux ** 0), x ** 0)
  expect_equal(as.numeric(ux ^ 0), x ^ 0)
  expect_identical(units(ux ** 0), units(as_units(1)))
  expect_identical(units(ux ^ 0), units(as_units(1)))
})

test_that("we support unary +/-", {
  expect_equal(-set_units(10, m), set_units(-10, m))
  expect_equal(+set_units(10, m), set_units(10, m))
})

test_that("we can convert units and simplify after multiplication", {
  x <- 1:4
  y <- 1:4
  z <- 1:4
  m <- as_units("m")
  s <- as_units("s")
  km <- as_units("km")
  ux <- x * m
  uy <- y * s
  uz <- z * km

  expect_equal(as.numeric(ux/ux), x/x)
  expect_equal(as.character(units(ux/ux)), units_options("unitless_symbol"))

  expect_equal(as.numeric(ux*uy), x*y)
  expect_equal(as.character(units(ux*uy)), "m*s")
  expect_equal(as.numeric(ux*uz), x*z)
  expect_equal(as.character(units(ux*uz)), "km*m")
  expect_equal(as.numeric(set_units(ux*uz, km * km)), (x/1000)*z)
  expect_equal(as.character(units(set_units(ux*uz, km * km))), "km^2")

  expect_equal(as.numeric(ux/ux), x/x)
  expect_equal(as.character(units(ux/ux)), units_options("unitless_symbol"))
  expect_equal(as.numeric(ux/uy), x/y)
  expect_equal(as.character(units(ux/uy)), "m/s")
  expect_equal(as.numeric(ux/uz), x/(1000*z))
  expect_equal(as.character(units(ux/uz)), units_options("unitless_symbol"))
  expect_equal(as.numeric(ux/uy/uz), x/y/z/1000)
  expect_equal(as.character(units(ux/uy/uz)), "1/s")
})

test_that("unit one is handled correctly", {
  one <- set_units(1)
  onem <- set_units(1, m)

  expect_equal(one * one, one)
  expect_equal(onem * one, onem)
  expect_equal(one * one * one, one)
  expect_equal(onem * one * one, onem)
  expect_equal(one / one, one)
  expect_equal(onem / one, onem)
  expect_equal(one ^ 3, one)
#  expect_equal(one ^ pi, one)
#  expect_equal(one ^ -pi, one)
})

test_that("we can compute powers if the result is an integer", {
  expect_equal(set_units(1:10, m^0), set_units(1:10))
  expect_equal(as.numeric(set_units(1:10, m^2) ^ 0.5), (1:10) ^ .5)
  expect_equal(as.numeric(set_units(1:10, m^2) ^ 3/2), (1:10) ^ 3/2)
  expect_error(as.numeric(set_units(1:10, m^3) ^ 0.5), "powers not divisible")
  expect_error(as.numeric(set_units(1:10, m^3) ^ 1.5))
  expect_equal(as.numeric(set_units(1:10, m^-3) ^ (-1/3)), (1:10) ^ (-1/3))
  expect_equal(as.numeric(set_units(1:10, m^-2) ^ (-1/2)), (1:10) ^ (-1/2))
  expect_equal(as.numeric(set_units(1:10, m^-2) ^ (1/2)),  (1:10) ^ (1/2))
  expect_error(set_units(1:10, m^2) ^ (1/pi), "powers not divisible")
  expect_error(set_units(1:10, m^3) ^ -1.5, "powers not divisible")
})

test_that("we can undo logatithms", {
  x <- set_units(1:5, cm^2)
  y <- set_units(1e6*x, dam^2)

  expect_equal(exp(10^log10(log(x))), set_units(x, m^2))
  expect_equal(exp(set_units(10)^log10(log(x))), set_units(x, m^2))
  expect_equal(exp(10^log10(log(y))), set_units(y, m^2))
  expect_equal(expm1(3^log(log1p(x), base=3)), set_units(x, m^2))
  expect_equal(expm1(3^log(log1p(y), base=3)), set_units(y, m^2))
  expect_error(exp(log10(x)), "wrong base in power operation")
  expect_error(exp(x), "only allowed with logarithmic unit")
  expect_error(exp(set_units(1, 1)), "only allowed with logarithmic unit")
})

test_that("%/% and %% work", {
  x <- set_units(3, m^2)
  y <- set_units(1.4, foot)

  expect_true(all.equal(x, y * (x %/% y) + x %% y))

  z <- set_units(drop_units(x) %% drop_units(set_units(y, m)), m^2)
  expect_equal(x %% y, z)
  expect_equal(x %% set_units(y, m), z)
  expect_equal(x %% drop_units(set_units(y, m)), z)
  expect_equal(x %% set_units(drop_units(set_units(y, m)), 1), z)
  expect_equal(x %% set_units(drop_units(set_units(y, m)), 1), z)
  expect_error(drop_units(x) %% set_units(y, m))

  z <- set_units(drop_units(set_units(x, foot^2)) %% drop_units(y), foot^2)
  expect_equal(set_units(x, foot^2) %% y, z)
  expect_equal(set_units(x, foot^2) %% set_units(y, m), z)
  expect_equal(set_units(x, foot^2) %% drop_units(y), z)
  expect_equal(set_units(x, foot^2) %% set_units(drop_units(y), 1), z)
  expect_error(drop_units(set_units(x, foot^2)) %% y)
})

#test_that("%*% work", {
  # mat = set_units(matrix(1:5, 1), m)
  # expect_equal(mat %*% t(mat), set_units(matrix(55), m^2))
  # expect_equal(t(mat) %*% mat, set_units(t(unclass(mat)) %*% unclass(mat), m^2))
#})

test_that("The order of multiplication for basic units is commutative", {
  a <- set_units(1:4, m^-3)
  b <- set_units(1:4, mm)

  expect_equal(a * b, b * a)
  expect_true(units(a * b) == units(b * a))

  a <- set_units(1:4, m)
  b <- set_units(1:4, mm)
  expect_equal(a * b, b * a)
  expect_true(units(a * b) == units(b * a))
})

test_that("Division gets the right scaling and units", {
  a <- set_units(1:4, m^-3)
  b <- set_units(1:4, mm)

  expect_true(units(a / b) != units(b / a))
  expect_equal(a / b, 1/(b / a))
})

test_that("we obtain mixed units when taking powers of multiple integers", {
  a = set_units(1:4, m)
  p = 4:1
  expect_equal(a ^ p, c(set_units(1, m^4), set_units(8, m^3), set_units(9, m^2), set_units(4, m),  allow_mixed=TRUE))
})

test_that("identical units can always be divided and return unitless (#310)", {
  x1 <- log(set_units(100, "g"))
  x2 <- log(set_units(4, "g"))
  expect_equal(
    x1/x2,
    set_units(log(100)/log(4), "")
  )
  expect_equal(
    x1 %/% x2,
    set_units(log(100) %/% log(4), "")
  )
})

test_that("inverse units can always be multiplied and return unitless (related to #310)", {
  x1 <- log(set_units(100, "g"))
  x2 <- log(set_units(4, "g"))
  expect_equal(
    x1*(1/x2),
    set_units(log(100)/log(4), "")
  )
})

test_that("identical unit multiplication and division respect 'simplify' option (#355)", {
  initial_unit_simplify <- units_options("simplify")
  on.exit(units_options(simplify=initial_unit_simplify))

  gg <- structure(list(numerator="g", denominator="g"), class="symbolic_units")
  x_gg <- set_units(1, gg, mode="standard")
  x_ul <- set_units(1, unitless, mode="standard")

  units_options(simplify = FALSE)
  expect_equal(set_units(1, "g/g"), x_gg)
  expect_equal(set_units(1, "g") / set_units(1, "g"), x_gg)

  units_options(simplify = TRUE)
  expect_equal(set_units(1, "g/g"), x_ul)
  expect_equal(set_units(1, "g") / set_units(1, "g"), x_ul)

  units_options(simplify = NA)
  expect_equal(set_units(1, "g/g"), x_gg)
  expect_equal(set_units(1, "g") / set_units(1, "g"), x_ul)
})

test_that("We can replace parts if they have an equivalent unit", {
  x <- 1:4 * as_units("km")
  y <- c(3000, 4000) * as_units("m")
  z <- 5000 * as_units("m")
  x[3:4] <- y
  x[[5]] <- z

  expect_length(x, 5)
  expect_equal(as.numeric(x), 1:5)
  expect_equal(as.character(units(x)), "km")
})

test_that("We can't replace parts if they have different unit", {
  x <- 1:4 * as_units("km")
  y <- c(3000, 4000) * as_units("s")
  z <- 5000 * as_units("s")

  expect_error(x[3:4] <- y)
  expect_error(x[[5]] <- z)
})

test_that("We can concatenate units if they have the same unit", {
  x <- 1:4 * as_units("m")
  y <- 5:8 * as_units("m")
  z <- c(x, y)

  expect_equal(length(z), length(x) + length(y))
  expect_equal(x, z[1:4])
  expect_equal(y, z[1:4 + 4])
})

test_that("We can't concatenate units if they have different units", {
  x <- 1:4 * as_units("m")
  y <- 5:8 * as_units("s")
  expect_error(c(x, y))
})

test_that("We can concatenate units if their units can be converted", {
  x <- 1:4 * as_units("m")
  y <- 5:8 * as_units("km")
  z <- c(x, y)

  expect_equal(length(z), length(x) + length(y))
  expect_equal(as.character(units(z)), "m")
  expect_equal(x, z[1:4])
  expect_equal(set_units(y, units(as_units("m")), mode = "standard"), z[1:4 + 4])
})

test_that("We can use diff on a units object", {
  x = 1:10 * as_units("m")
  y = rep(1,9) * as_units("m")
  expect_equal(diff(x), y)
})

test_that("parse_unit works", {
  kg = as_units("kg")
  m = as_units("m")
  s = as_units("s")
  u0 = kg/m/m/s
  u = as_units("kg m-2 s-1", implicit_exponents = TRUE)
  expect_equal(u, u0)
  J = as_units("J")
  u0 = make_units(kg*kg*kg*m*m*J/s/s/s/s/s)
  u = as_units("kg3 m2 s-5 J", implicit_exponents = TRUE)
  expect_equal(u, u0)
})

test_that("deparse_unit works", {
  str = "kg m-2 s-1"
  u = as_units(str, implicit_exponents = TRUE)
  str0 = deparse_unit(u)
  expect_equal(str, str0)
})

test_that("set_units(x, u) is a short form for x * make_units(u)", {
  expect_equal(set_units(1:10, m/s), 1:10 * make_units(m/s))
  x = set_units(1:5, m/s)
  y = x
  units(y) = set_units(1, km/h)
  expect_identical(y, set_units(x, km/h))
})

test_that("rep.units works", {
  expect_equal(rep(set_units(1:2, m/s), 2), set_units(c(1,2,1,2), m/s))
})

test_that("set_units works with symbols in character data, and resolves names", {
  skip_on_cran()
  deg = paste0(enc2native(intToUtf8(176)), "C")
  expect_equal(set_units(1:2, deg, mode = "standard"),        set_units(1:2, "degree_C", mode = "standard"))
  expect_equal(set_units(1:2, deg, mode = "standard"),        set_units(1:2, "degree_Celsius", mode = "standard"))
  expect_equal(set_units(1:2, "degree_C", mode = "standard"), set_units(1:2, "degree_Celsius", mode = "standard"))
  expect_equal(set_units(1:2, degree_C),                      set_units(1:2, degree_Celsius))
  expect_equal(set_units(1:2, deg, mode = "standard"),        set_units(1:2, degree_Celsius))
  x = set_units(1:3, km)
  y <- set_units(x, "meter", mode = "standard")
  expect_equal(y, set_units(c(1000,2000,3000), m))
})

test_that("all.equal works", {
  expect_false(isTRUE(all.equal(set_units(1, m), 1)))

  expect_true(all.equal(set_units(1, m/s), set_units(3.6, km/h)))
  expect_true(all.equal(set_units(3.6, km/h), set_units(1, m/s)))

  expect_true(set_units(1, m/s) == set_units(3.6, km/h))
  expect_false(set_units(3.6, km/h) == set_units(1, m/s)) # see R FAQ 7.31
  expect_true(set_units(1e-20, g) < set_units(1e-10, g))  # see #351
})

test_that("seq works", {
  expect_equal(
    seq(to = set_units(10, m), by = set_units(1, m), length.out = 5),
    set_units(seq(to = 10, by = 1, length.out = 5), m)
  )
  expect_equal(
    seq(set_units(10, m), by = set_units(1, m), length.out = 5),
    set_units(seq(10, by = 1, length.out = 5), m)
  )
  expect_equal(
    seq(set_units(10, m), set_units(19, m)),
    set_units(seq(10, 19), m)
  )
  expect_equal(
    seq(set_units(10, m), set_units(.02, km)),
    set_units(seq(10, 20), m)
  )
})

test_that("str works", {
  expect_snapshot({
    str(set_units(1/1:3, m/s))
  })
})

test_that("subsetting keeps pillar attribute (#275)", {
  x <- set_units(1:3, m)
  attr(x, "pillar") <- "bar"

  expect_equal(attr(x[1:2], "pillar"), "bar")
  expect_equal(attr(x[[1]], "pillar"), "bar")
})

test_that("duplicated-related methods work as expected", {
  x <- set_units(1:4, m)
  expect_equal(duplicated(x), duplicated(drop_units(x)))
  expect_equal(anyDuplicated(x), anyDuplicated(drop_units(x)))
  expect_equal(unique(x), x)

  x <- set_units(rep(c(1, 2), 2), m)
  expect_equal(duplicated(x), duplicated(drop_units(x)))
  expect_equal(anyDuplicated(x), anyDuplicated(drop_units(x)))
  expect_equal(unique(x), x[1:2])

  x <- set_units(matrix(rep(c(1, 2), 2), 2, byrow=TRUE), m)
  expect_equal(duplicated(x), duplicated(drop_units(x)))
  expect_equal(anyDuplicated(x), anyDuplicated(drop_units(x)))
  expect_equal(unique(x), x[1, , drop=FALSE])
})

test_that("we can convert numbers to unit-less units", {
  x <- as_units(1:4)
  expect_equal(length(x), 4)
  expect_equal(class(x), "units")
  expect_equal(as.numeric(x), 1:4)

  y <- 1:4
  units(y) <- unitless
  expect_equal(x, y)
})

test_that("we can convert numbers to physical units", {
  m <- as_units("m")
  x <- 1:4 * m
  expect_equal(length(x), 4)
  expect_equal(class(x), "units")
  expect_equal(as.character(units(x)), "m")
  expect_equal(as.numeric(x), 1:4)

  y <- 1:4
  units(y) <- m
  expect_equal(x, y)

  z <- 1:4 / m
  expect_equal(length(z), 4)
  expect_equal(class(z), "units")
  expect_equal(as.character(units(z)), "1/m")
  expect_equal(as.numeric(z), 1:4)
})

test_that("we can convert NA values to physical units", {
  m <- as_units("m")
  x <- NA * m
  expect_equal(class(x), "units")
  expect_equal(as.character(units(x)), "m")
  expect_equal(as.numeric(x), as.numeric(NA))

  x <- set_units(NA,m/s)
  expect_equal(as.character(units(x)), "m/s")
  expect_equal(x + set_units(5,m/s), set_units(NA,m/s))
  expect_error(x + set_units(5,m))

  x <- NA
  units(x) <- m
  expect_equal(as.character(units(x)), 'm')

  x <- rep(NA,5)
  s <- as_units("s")
  units(x) <- make_units(m/s)
  expect_equal(length(x),5)
  expect_equal(units(x),units(m/s))
  expect_equal(x,5 * x)
  expect_error(x + 1)
})

test_that("we can convert between two units that can be converted", {
  m <- as_units("m")
  km <- as_units("km")
  x <- y <- 1:4 * m
  units(x) <- km
  expect_equal(as.numeric(y), 1000 * as.numeric(x))
  skip_if_not_installed("magrittr")
  `%>%` <- magrittr::`%>%`
  y %>% set_units(km) -> z
  expect_equal(x, z)
})

test_that("we can't convert between two units that can't be converted", {
  m <- as_units("m")
  s <- as_units("s")
  expect_error(units(m) <- s)
})

test_that("we can convert difftime objects to units", {
  s <- Sys.time()
  d <- s - (s + 1)
  x <- as_units(d)
  expect_equal(as.numeric(x), as.numeric(d))

  week <- as.difftime(1, units = "weeks")
  units_week <- as_units(week)
  expect_equal(as.character(units(units_week)), "d")
  expect_equal(as.numeric(units_week), 7)
})

test_that("we can convert units objects to difftime objects", {
  s <- Sys.time()
  d <- s - (s + 1)
  x <- as_units(d)
  y <- as_difftime(x)

  expect_equal(d, y)
})

#test_that("we can convert units objects to and from hms objects", {
#  s <- Sys.time()
#  library(hms)
#  d <- as.hms(s - (s + 1))
#  x <- as_units(d)
#  y <- as.hms(x)
#
#  expect_equal(d, y)
#})

test_that("we can subscript units", {
  x <- 1:4
  y <- x * as_units("m")
  expect_equal(as.numeric(y[1]), x[1])
  expect_equal(class(y[1]), class(y))
  expect_equal(as.numeric(y[[1]]), x[[1]])
  expect_equal(class(y[[1]]), class(y))
})

test_that("m + m*s is an error", {
  m <- as_units("m")
  s <- as_units("s")
  expect_error(m + m * s)
})

test_that("we can convert between units that are not simply a scalar from each other", {
  m <- 0 * as_units("degC")
  units(m) <- as_units("degK")
  expect_equal(as.numeric(m), units:::ud_convert(0, "degC", "degK"))
  expect_equal(as.character(units(m)), "K")

  temp <- 75 * as_units('degF')
  units(temp) <- as_units('degK')
  result <- temp / as_units('degF')
  expect_equal(as.numeric(result), 75)
  expect_equal(units(result), unitless)
})

test_that("dim propagates", {
  y = x = set_units(matrix(1:4,2), m)
  units(y) = as_units("mm")
  expect_equal(dim(x), dim(y))
})

test_that("conversion of g/kg to dimensionless is not the default", {
	a_orig <- a <- 1:10
	units(a) = as_units("mg/kg")
	expect_equal(as.numeric(a), a_orig)
})

test_that("conversion to dimensionless with prefix works (g/kg) if simplify=TRUE", {
	a_orig <- a <- 1:10
	units_options(simplify = TRUE)
	units(a) = as_units("mg/kg")
	expect_equal(as.numeric(a), a_orig/1e6)
	units(a) = as_units("kg/mg")
	expect_equal(a, set_units(a_orig, 1))
	units(a) = as_units("g/g")
	expect_equal(a, set_units(a_orig, 1))
	units(a) = as_units("kg/g")
	expect_equal(a, set_units(a_orig * 1000, 1))
	units_options(simplify = NA)
})

test_that("a NULL value returns NULL", {
  expect_null(as_units(NULL))
})

test_that("as.data.frame.units works", {
  x <- matrix(1:9, 3)

  df1 <- as.data.frame(x)
  for (col in names(df1))
    df1[[col]] <- set_units(df1[[col]], m)
  df2 <- as.data.frame(set_units(x, m))

  expect_equal(df1, df2)
})

test_that("units.symbolic_units works", {
  m = set_units(1, m)
  expect_equal(units(m), units(units(m)))
})

test_that("new base units work", {
  install_unit("person")
  expect_equal(set_units(1, person) + set_units(1, kperson), set_units(1001, person))
  expect_error(set_units(1, person) + set_units(1, rad), "cannot convert")

  # restore
  remove_unit("person")
})

test_that("errors are correctly coerced to a data frame", {
  a <- 1:10
  b <- a * as_units("m")

  expect_equal(as.data.frame(b)$b, b)
  x <- data.frame(a, b)
  expect_equal(x$a, a)
  expect_equal(x$b, b)
  x <- cbind(x, a, data.frame(b))
  expect_equal(x[[3]], a)
  expect_equal(x[[4]], b)
  x <- rbind(x, a[1:4], x[1,])
  expect_equal(x[[1]], c(a, 1, 1))
  expect_equal(x[[2]], c(b, c(2, 1) * as_units("m")))
  expect_equal(x[[3]], c(a, 3, 1))
  expect_equal(x[[4]], c(b, c(4, 1) * as_units("m")))
})

test_that("units are correctly coerced to a list", {
  x <- 1:10 * as_units("m")
  y <- as.list(x)
  expect_type(y, "list")
  expect_true(all(sapply(seq_along(y), function(i) all.equal(y[[i]], x[i]))))
})

test_that("NA as units generate warnings", {
  expect_error(set_units(NA_real_, NA_character_, mode="standard"), "a missing value for units is not allowed")
  expect_error(set_units(NA_real_, NA, mode="standard"), "a missing value for units is not allowed")
})

test_that("ud_are_convertible return the expected value", {
  x <- 1:10 * as_units("m")
  expect_type(ud_are_convertible("m", "km"), "logical")
  expect_true(ud_are_convertible("m", "km"))
  expect_true(ud_are_convertible(units(x), "km"))
  expect_false(ud_are_convertible("s", "kg"))
})

test_that("set_units keeps names even with unit conversion (#305)", {
  expect_named(set_units(c(a=1), "m"), "a")
  expect_named(
    set_units(set_units(c(a=1), "m"), "km"),
    "a"
  )
})

test_that("conversion to mixed_units", {
  a <- set_units(1:3, m/s)
  expect_s3_class(a, "units")
  units(a) <- c("m/s", "km/h", "km/h")
  expect_s3_class(a, "mixed_units")
  expect_equal(
    drop_units(a),
    c(1, 0.002*3600, 0.003*3600)
  )
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

ud_exit <- function() {
    invisible(.Call('_units_ud_exit', PACKAGE = 'units'))
}

ud_init <- function(path) {
    invisible(.Call('_units_ud_init', PACKAGE = 'units', path))
}

ud_set_encoding <- function(enc_str) {
    invisible(.Call('_units_ud_set_encoding', PACKAGE = 'units', enc_str))
}

ud_compare <- function(x, y, xn, yn) {
    .Call('_units_ud_compare', PACKAGE = 'units', x, y, xn, yn)
}

ud_convert <- function(val, from, to) {
    .Call('_units_ud_convert', PACKAGE = 'units', val, from, to)
}

ud_map_names <- function(names, inunit) {
    invisible(.Call('_units_ud_map_names', PACKAGE = 'units', names, inunit))
}

ud_unmap_names <- function(names) {
    invisible(.Call('_units_ud_unmap_names', PACKAGE = 'units', names))
}

ud_map_symbols <- function(symbols, inunit) {
    invisible(.Call('_units_ud_map_symbols', PACKAGE = 'units', symbols, inunit))
}

ud_unmap_symbols <- function(symbols) {
    invisible(.Call('_units_ud_unmap_symbols', PACKAGE = 'units', symbols))
}

R_ut_get_dimensionless_unit_one <- function() {
    .Call('_units_R_ut_get_dimensionless_unit_one', PACKAGE = 'units')
}

R_ut_new_base_unit <- function() {
    .Call('_units_R_ut_new_base_unit', PACKAGE = 'units')
}

R_ut_new_dimensionless_unit <- function() {
    .Call('_units_R_ut_new_dimensionless_unit', PACKAGE = 'units')
}

R_ut_get_name <- function(unit) {
    .Call('_units_R_ut_get_name', PACKAGE = 'units', unit)
}

R_ut_get_symbol <- function(unit) {
    .Call('_units_R_ut_get_symbol', PACKAGE = 'units', unit)
}

R_ut_are_convertible <- function(a, b) {
    .Call('_units_R_ut_are_convertible', PACKAGE = 'units', a, b)
}

R_ut_scale <- function(unit, factor) {
    .Call('_units_R_ut_scale', PACKAGE = 'units', unit, factor)
}

R_ut_offset <- function(unit, origin) {
    .Call('_units_R_ut_offset', PACKAGE = 'units', unit, origin)
}

R_ut_multiply <- function(a, b) {
    .Call('_units_R_ut_multiply', PACKAGE = 'units', a, b)
}

R_ut_invert <- function(a) {
    .Call('_units_R_ut_invert', PACKAGE = 'units', a)
}

R_ut_divide <- function(numer, denom) {
    .Call('_units_R_ut_divide', PACKAGE = 'units', numer, denom)
}

R_ut_raise <- function(a, i) {
    .Call('_units_R_ut_raise', PACKAGE = 'units', a, i)
}

R_ut_root <- function(a, i) {
    .Call('_units_R_ut_root', PACKAGE = 'units', a, i)
}

R_ut_log <- function(a, base) {
    .Call('_units_R_ut_log', PACKAGE = 'units', a, base)
}

R_ut_parse <- function(name) {
    .Call('_units_R_ut_parse', PACKAGE = 'units', name)
}

R_ut_format <- function(p, names = FALSE, definition = FALSE, ascii = FALSE) {
    .Call('_units_R_ut_format', PACKAGE = 'units', p, names, definition, ascii)
}

# Inside the group generic functions we do have .Generic even if the diagnostics
# think we do not.
# !diagnostics suppress=.Generic

#' @export
Summary.units = function(..., na.rm = FALSE) {
  OK <- switch(.Generic, "sum" = , "min" = , "max" = , "range" = TRUE, FALSE)
  if (! OK)
    stop(paste("Summary operation", .Generic, "not allowed"))

  args <- list(...)
  if (length(args) > 1) {
    args <- do.call(c, args) # turns args into a single units vector
    do.call(.Generic, c(list(args), na.rm = na.rm)) # concatenate, convert if necessary, re-call with length 1 arg
  } else
	.as.units(NextMethod(), units(args[[1]]))
}


#' @export
print.units = function (x, ...) { # nocov start
  gr = units_options("group")
  if (is.array(x) || length(x) > 1L) {
    cat("Units: ", paste0(gr[1], as.character(attr(x, "units")), gr[2]), "\n", sep = "")
    x <- drop_units(x)
    attr(x, "pillar") <- NULL
    NextMethod()
  } else {
    cat(format(x, ...), "\n", sep="")
    invisible(x)
  }
} # nocov end

#' @export
mean.units = function(x, ...) {
  .as.units(NextMethod(), units(x))
}

#' @export
weighted.mean.units = function(x, w, ...) {
  if (!missing(w) && inherits(w, "units"))
    w = drop_units(w)
  .as.units(NextMethod(), units(x))
}

#' @export
median.units = mean.units

#' @export
quantile.units = function(x, ...) {
  .as.units(quantile(unclass(x), ...), units(x))
}

#' @export
format.units = function(x, ...) {
  gr = units_options("group")
  u = paste0(gr[1], units(x), gr[2])
  x <- drop_units(x)
  setNames(paste(NextMethod(), u), names(x))
}

#' @export
summary.units = function(object, ...) {
  summary(unclass(object), ...)
}

# Build against mingw-w64 build of udunits
if(!file.exists("../windows/udunits-2.2.28/include/udunits.h")){
  if(getRversion() < "3.3.0") setInternet2()
  download.file("https://github.com/rwinlib/udunits/archive/v2.2.28.zip", "lib.zip", quiet = TRUE)
  dir.create("../windows", showWarnings = FALSE)
  unzip("lib.zip", exdir = "../windows")
  unlink("lib.zip")
}

`%|%` <- function(x, y) ifelse(is.na(x), y, x)

`%empty%` <-  function(x, y) if(length(x)==0) y else x

pcc <- function(...) paste0(..., collapse = ", ")

#' Get information about valid units
#'
#' These functions require the \pkg{xml2} package, and return data frames with
#' complete information about pre-defined units from UDUNITS2. Inspect this data
#' frames to determine what inputs are accepted by \code{as_units} (and the
#' other functions it powers: \code{as_units}, \code{set_units} , \code{units<-}).
#'
#' Any entry listed under \code{symbol} , \code{symbol_aliases} , \code{
#' name_singular} , \code{name_singular_aliases} , \code{name_plural} , or
#' \code{name_plural_aliases} is valid. Additionally, any entry under
#' \code{symbol} or \code{symbol_aliases} may can also contain a valid prefix,
#' as specified by \code{valid_udunits_prefixes()} .
#'
#' Note, this is primarily intended for interactive use, the exact format of the
#' returned data frames may change in the future.
#'
#' @param quiet logical, defaults \code{TRUE} to give a message about the location of
#'   the udunits database being read.
#'
#' @return a data frame with columns \code{symbol} , \code{symbol_aliases} ,
#'   \code{name_singular} , \code{name_singular_aliases} , \code{name_plural} ,
#'   or \code{name_plural_aliases} , \code{def} , \code{definition} ,
#'   \code{comment} , \code{dimensionless} and \code{source_xml}
#'
#' @export
#'
#' @name valid_udunits
#' @examples
#' if (requireNamespace("xml2", quietly = TRUE)) {
#'   valid_udunits()
#'   valid_udunits_prefixes()
#'   if(interactive())
#'     View(valid_udunits())
#' }
valid_udunits <- function(quiet = FALSE) {
  db <- .read_ud_db(type="unit")
  if (!quiet) message(.startup_msg())

  xml_nodes <- xml2::xml_children(db)

  l <- lapply(seq_len(xml2::xml_length(db)), function(i) {
    unit <- xml_nodes[[i]]

    source_xml <- xml2::xml_text(xml2::xml_find_first(unit, ".//source"))

    symbols <- xml2::xml_find_all(unit, ".//symbol")
    symbols <- xml2::xml_text(symbols) %empty% ""
    symbol         <-     symbols[ 1]
    symbol_aliases <- pcc(symbols[-1])

    unit_names <- xml2::xml_find_all(unit, ".//name")

    all_names <- unlist(lapply(unit_names, function(.x)
      xml2::xml_text(xml2::xml_children(.x))))

    singular <- xml2::xml_find_all(unit_names, ".//singular")
    singular <- xml2::xml_text(singular)

    plural   <- xml2::xml_find_all(unit_names, ".//plural")
    plural   <- xml2::xml_text(plural)

    name_singular         <-     singular[ 1]  %|% ""
    name_singular_aliases <- pcc(singular[-1]) %|% ""

    name_plural         <-     plural[ 1]  %|% ""
    name_plural_aliases <- pcc(plural[-1]) %|% ""

    def <- xml2::xml_find_all(unit, ".//def")
    def <- xml2::xml_text(def) %empty% ""

    definition <-  xml2::xml_find_all(unit, ".//definition")
    definition <- xml2::xml_text(definition) %empty% ""

    comment <- xml2::xml_find_all(unit, ".//comment")
    comment <- xml2::xml_text(comment) %empty% ""

    dimensionless <- xml2::xml_find_all(unit, ".//dimensionless")
    dimensionless <- as.logical(length(dimensionless))

    # all node names that might be in a unit node
    # db %>% xml_children() %>% map(~xml_children(.x) %>% xml_name()) %>%
    # unique() %>% unlist() %>% unique()
    # [1] "base"          "name"          "symbol"
    # [4] "aliases"       "definition"    "def"
    # [7] "comment"       "dimensionless"

    # rest_xml <- unit %>% xml_children()
    # rest <- map(rest_xml, xml_text)
    # names(rest) <- rest_xml %>% xml_name()
    # rest <- list(rest)

    data.frame(symbol, symbol_aliases,
               name_singular, name_singular_aliases,
               name_plural, name_plural_aliases,
               def, definition, comment, dimensionless, source_xml, #, rest
               stringsAsFactors = FALSE)

  })

  df <- do.call(rbind, c(l, stringsAsFactors=FALSE, make.row.names=FALSE))
  class(df) <- c( "tbl_df", "tbl", "data.frame")
  df
}

#' @name valid_udunits
#' @export
valid_udunits_prefixes <- function(quiet = FALSE) {
  db <- .read_ud_db(type="prefix")
  if (!quiet) message(.startup_msg())

  # all prefix valid names
  # db %>% xml_children() %>% map(~xml_children(.x) %>% xml_name()) %>%
  #   unlist() %>% unique()
  # "value"  "name"   "symbol"

  l <- lapply( xml2::xml_children(db), function(prefix) {

    symbols <- xml2::xml_find_all(prefix, ".//symbol")
    symbols <- xml2::xml_text(symbols)
    symbol <- symbols[1]
    symbol_aliases <- pcc(symbols[-1])

    name <- xml2::xml_find_all(prefix, ".//name")
    name <- xml2::xml_text(name)

    value <- xml2::xml_find_all(prefix, ".//value")
    value <- xml2::xml_double(value)

    data.frame(symbol, symbol_aliases, name, value)
  })

  df <- do.call(rbind.data.frame,
                c(l, stringsAsFactors = FALSE, make.row.names = FALSE))
  class(df) <- c( "tbl_df", "tbl", "data.frame")
  df
}

test_that("udunits database can be read", {
  skip_if_not_installed("xml2")
  suppressMessages(expect_s3_class(valid_udunits(), "data.frame"))
  suppressMessages(expect_s3_class(valid_udunits_prefixes(), "data.frame"))
})

#' Define or remove units
#'
#' Installing new symbols and/or names allows them to be used in \code{as_units},
#' \code{make_units} and \code{set_units}. Optionally, a relationship can be
#' defined between such symbols/names and existing ones (see details and examples).
#'
#' At least one symbol or name is expected, but multiple symbols and/or names
#' can be installed (and thus mapped to the same unit) or removed at the same
#' time. The \code{def} argument enables arbitrary relationships with existing
#' units using UDUNITS-2 syntax:
#' \tabular{llll}{
#'   \strong{String Type} \tab \strong{Using Names} \tab \strong{Using Symbols}
#'     \tab \strong{Comment}\cr
#'   Simple \tab meter \tab m \tab \cr
#'   Raised \tab meter^2 \tab m2 \tab
#'     higher precedence than multiplying or dividing\cr
#'   Product \tab newton meter \tab N.m \tab \cr
#'   Quotient \tab meter per second \tab m/s \tab \cr
#'   Scaled \tab 60 second \tab 60 s \tab \cr
#'   Prefixed \tab kilometer \tab km \tab \cr
#'   Offset \tab kelvin from 273.15 \tab K @ 273.15 \tab
#'     lower precedence than multiplying or dividing\cr
#'   Logarithmic \tab lg(re milliwatt) \tab lg(re mW) \tab
#'     "lg" is base 10, "ln" is base e, and "lb" is base 2\cr
#'   Grouped \tab (5 meter)/(30 second) \tab (5 m)/(30 s) \tab
#' }
#' The above may be combined, e.g., \code{"0.1 lg(re m/(5 s)^2) @ 50"}.
#' You may also look at the \code{<def>} elements in the units database to see
#' examples of string unit specifications.
#'
#' @param symbol a vector of symbols to be installed/removed.
#' @param def either \itemize{
#'   \item an empty definition, which defines a new base unit;
#'   \item \code{"unitless"}, which defines a new dimensionless unit;
#'   \item a relationship with existing units (see details for the syntax).
#' }
#' @param name a vector of names to be installed/removed.
#'
#' @examples
#' # define a fortnight
#' install_unit("fn", "2 week", "fortnight")
#' year <- as_units("year")
#' set_units(year, fn)        # by symbol
#' set_units(year, fortnight) # by name
#' # clean up
#' remove_unit("fn", "fortnight")
#'
#' # working with currencies
#' install_unit("dollar")
#' install_unit("euro", "1.22 dollar")
#' install_unit("yen", "0.0079 euro")
#' set_units(as_units("dollar"), yen)
#' # clean up
#' remove_unit(c("dollar", "euro", "yen"))
#'
#' # an example from microbiology
#' cfu_symbols <- c("CFU", "cfu")
#' cfu_names <- c("colony_forming_unit", "ColonyFormingUnit")
#' install_unit("cell")
#' install_unit(cfu_symbols, "3.4 cell", cfu_names)
#' cell <- set_units(2.5e5, cell)
#' vol <- set_units(500, ul)
#' set_units(cell/vol, "cfu/ml")
#' set_units(cell/vol, "CFU/ml")
#' set_units(cell/vol, "colony_forming_unit/ml")
#' set_units(cell/vol, "ColonyFormingUnit/ml")
#' # clean up
#' remove_unit(c("cell", cfu_symbols), cfu_names)
#'
#' @export
install_unit <- function(symbol=character(0), def=character(0), name=character(0)) {
  stopifnot(is.character(def), length(def) < 2)
  stopifnot(is.character(symbol), is.character(name))
  if (!length(symbol) && !length(name))
    stop("at least one symbol or name must be specified")

  # workaround for https://github.com/r-quantities/units/issues/289
  if (any(grepl("%", c(symbol, name), fixed=TRUE)))
    stop("Symbols/names with a percentage character '%' are not allowed ",
         "(see https://github.com/r-quantities/units/issues/289)")

  if (!length(def)) {
    ut_unit <- R_ut_new_base_unit()
  } else if (identical(def, "unitless")) {
    ut_unit <- R_ut_new_dimensionless_unit()
  } else {
    ut_unit <- R_ut_parse(def)
  }

  ud_map_symbols(symbol, ut_unit)
  ud_map_names(name, ut_unit)
}

#' @rdname install_unit
#' @export
remove_unit <- function(symbol=character(0), name=character(0)) {
  stopifnot(is.character(symbol), is.character(name))
  if (!length(symbol) && !length(name))
    stop("at least one symbol or name must be specified")

  ud_unmap_symbols(symbol)
  ud_unmap_names(name)
}

test_that("we can convert between units with a user-defined function", {
  expect_error(as_units("apple"))
  expect_error(as_units("orange"))
  install_unit("orange")
  oranges <- 3 * as_units("orange")
  expect_error(apples + oranges) # obviously

  install_unit("apple", "orange / 2") # one orange is worth two apples
  apples <- 2 * as_units("apple")
  expect_equal(apples + oranges, (2 + 2*3) * as_units("apple"))
  expect_equal(oranges + apples, (3 + 2/2) * as_units("orange"))
  expect_equal(oranges + apples, set_units(apples + oranges, units(oranges), mode = "standard"))
  expect_equal(apples + oranges, set_units(apples + oranges, units(apples), mode = "standard"))

  # now just checking that we get different results with a different fruit
  expect_error(as_units("banana"))
  expect_error(apples + bananas) # obviously
  expect_error(bananas + apples) # obviously

  install_unit("banana", "3 apple") # one apple is worth three bananas
  bananas <- 6 * as_units("banana")
  expect_equal(bananas + 3 * apples, (6 + 3 * 2 / 3) * as_units("banana"))

  # check dimensionless
  install_unit("person", "1")
  persons <- set_units(3, person)
  expect_true(persons == set_units(3, 1))

  # restore
  remove_unit(c("orange", "apple", "banana", "person"))
})

test_that("we can simplify via user-defined units", {
  install_unit("orange")
  install_unit("apple", "orange / 2") # one orange is worth two apples
  apples <- 4 * as_units("apple")
  oranges <- 2 * as_units("orange")
  expect_equal(apples / oranges, set_units(1))
  expect_equal(oranges / apples, set_units(1))

  # restore
  remove_unit(c("orange", "apple"))
})

test_that("removing units works", {
  expect_silent(remove_unit("foo"))
  expect_silent(install_unit("foo"))
  expect_error(install_unit("foo"))
  expect_silent(remove_unit("foo"))
})

#' Plot \code{units} objects
#'
#' Create axis label with appropriate labels.
#'
#' @param lab length one character; name of the variable to plot
#' @param u vector of class \code{units}
#' @param sep length two character vector, defaulting to \code{c("~","~")}, with
#'   the white space between unit name and unit symbols, and between subsequent
#'   symbols.
#' @param group length two character vector with grouping symbols, e.g.
#'   \code{c("(",")")} for parenthesis, or \code{c("","")} for no group symbols
#' @param parse logical; indicates whether a parseable expression should be
#'   returned (typically needed for super scripts), or a simple character string
#'   without special formatting.
#'
#' @details \link{units_options} can be used to set and change the defaults for
#'   \code{sep}, \code{group} and \code{doParse}.
#'
#' @name plot.units
#' @export
make_unit_label = function(lab, u,
  		sep = units_options("sep"),
  		group = units_options("group"),
  		parse = units_options("parse")) {

  if (parse) {
    str = paste0("group('", group[1], "',",
                 as.character(units(u), escape_units = TRUE, plot_sep = sep[2]),
                 ",'", group[2], "')")

	  if (length(grep("[^\t ]", lab)) > 0) {
	    if (is.character(lab))
	      lab <- gsub("[[:space:]]+", "~", lab)
	    str = paste0(lab, "*", sep[1], str)
	  }

    parse(text = str)

  } else {
    paste0(lab, " ", group[1], as.character(units(u)), group[2])
  }
}

.make_unit_label <- function(label, unit, default) {
  if (is.null(label))
    label <- default
  if (inherits(unit, "units"))
    label <- make_unit_label(label, unit)
  label
}

#' @description Plot method for \code{units} objects.
#'
#' @param x object of class units, to plot along the x axis, or, if y is missing, along the y axis
#' @param y object to plot along the y axis, or missing
#' @param xlab character; x axis label
#' @param ylab character; y axis label
#' @param ... other parameters, passed on to \link{plot.default}
#'
#' @examples
#' displacement = mtcars$disp * as_units("in")^3
#' units(displacement) = make_units(cm^3)
#' weight = mtcars$wt * 1000 * make_units(lb)
#' units(weight) = make_units(kg)
#' plot(weight, displacement)
#'
#' units_options(group = c("(", ")") )  # parenthesis instead of square brackets
#' plot(weight, displacement)
#'
#' units_options(sep = c("~~~", "~"), group = c("", ""))  # no brackets; extra space
#' plot(weight, displacement)
#'
#' units_options(sep = c("~", "~~"), group = c("[", "]"))
#' gallon = as_units("gallon")
#' consumption = mtcars$mpg * make_units(mi/gallon)
#' units(consumption) = make_units(km/l)
#' plot(displacement, consumption) # division in consumption
#'
#' units_options(negative_power = TRUE) # division becomes ^-1
#' plot(displacement, consumption)
#'
#' plot(1/displacement, 1/consumption)
#'
#' @export
plot.units <- function(x, y, xlab = NULL, ylab = NULL, ...) {
  # We define the axis labels if they are not already provided and then let
  # the default plotting function take over...
  xlab0 <- deparse1(substitute(x))
  ylab0 <- deparse1(substitute(y))

  if (missing(y)) { # from xy.coords:
    ylab <- .make_unit_label(ylab, x, xlab0)
    if (is.null(xlab)) xlab <- "Index"

    y <- x
    x <- seq_along(x)

    return(NextMethod("plot", x, y, xlab=xlab, ylab=ylab))
  }

  xlab <- .make_unit_label(xlab, x, xlab0)
  ylab <- .make_unit_label(ylab, y, ylab0)

  NextMethod("plot", xlab=xlab, ylab=ylab)
}

#' histogram for unit objects
#'
#' histogram for unit objects
#' @param x object of class units, for which we want to plot the histogram
#' @param xlab character; x axis label
#' @param main character; title of histogram
#' @param ... parameters passed on to \link{hist.default}
#' @export
#' @examples
#' units_options(parse = FALSE) # otherwise we break on the funny symbol!
#' u = set_units(rnorm(100), degree_C)
#' hist(u)
hist.units <- function(x, xlab = NULL, main = paste("Histogram of", xname), ...) {
  # We define the axis labels if they are not already provided and then let
  # the default plotting function take over...
  xname <- deparse1(substitute(x))
  xlab <- .make_unit_label(xlab, x, xname)
  NextMethod("hist", xlab=xlab, main=main)
}

#' boxplot for unit objects
#'
#' boxplot for unit objects
#' @param x object of class units, for which we want to plot the boxplot
#' @param ... parameters passed on to \link{boxplot.default}
#' @param horizontal logical indicating if the boxplots should be horizontal;
#' default FALSE means vertical boxes.
#' @export
#' @examples
#' units_options(parse = FALSE) # otherwise we break on the funny symbol!
#' u = set_units(rnorm(100), degree_C)
#' boxplot(u)
boxplot.units <- function(x, ..., horizontal = FALSE) {
  xlab <- ylab <- NULL
  lab <- make_unit_label(deparse1(substitute(x)), x)
  if (horizontal)
    xlab <- lab
  else ylab <- lab
  x <- drop_units(x)
  NextMethod("boxplot", xlab=xlab, ylab=ylab)
}

library(units)
oldpar = par(mar = par("mar") + c(0, .3, 0, 0))
displacement = mtcars$disp * ud_units[["in"]]^3
units(displacement) = with(ud_units, cm^3)
weight = mtcars$wt * 1000 * with(ud_units, lb)
units(weight) = with(ud_units, kg)
plot(weight, displacement)
units_options(group = c("(", ")") )  # parenthesis instead of square brackets
plot(weight, displacement)
units_options(sep = c("~~~", "~"), group = c("", ""))  # no brackets; extra space
plot(weight, displacement)
units_options(sep = c("~", "~~"), group = c("[", "]"))
gallon = as_units("gallon")
consumption = mtcars$mpg * with(ud_units, mi/gallon)
units(consumption) = with(ud_units, km/l)
plot(displacement, consumption) # division in consumption
units_options(negative_power = TRUE) # division becomes ^-1
plot(displacement, consumption)
plot(1/displacement, 1/consumption)
par(oldpar)

units_options(parse = FALSE)
n = 100
u = rnorm(1:n) * as_units("degree_C")
v = rnorm(1:n) * as_units("s")
plot(u, v)
plot(u, type = 'l')
hist(u)
boxplot(u)

do.call(units_options, units:::.default_options)

# helper function:
.as.units = function(x, value, dim) {
  if (missing(dim))
    structure(x, units = value, class = "units")
  else
    structure(x, units = value, dim = dim, class = "units")
}

#' @name units
#' @export
#'
#' @param bare_expression a bare R expression describing units. Must be valid R
#'   syntax (reserved R syntax words like \code{in} must be backticked)
#'
#' @examples
#' # The easiest way to assign units to a numeric vector is like this:
#' x <- y <- 1:4
#' units(x) <- "m/s"  # meters / second
#'
#' # Alternatively, the easiest pipe-friendly way to set units:
#' if(requireNamespace("magrittr", quietly = TRUE)) {
#'   library(magrittr)
#'   y %>% set_units(m/s)
#' }
#'
#' # these are different ways of creating the same unit:
#' # meters per second squared, i.e, acceleration
#' x1 <- make_units(m/s^2)
#' x2 <- as_units(quote(m/s^2))
#' x2 <- as_units("m/s^2")
#' x3 <- as_units("m s-2") # in product power form, i.e., implicit exponents = T
#' x4 <- set_units(1,  m/s^2) # by default, mode = "symbols"
#' x5 <- set_units(1, "m/s^2",   mode = "standard")
#' x6 <- set_units(1, x1,        mode = "standard")
#' x7 <- set_units(1, units(x1), mode = "standard")
#' x8 <- as_units("m") / as_units("s")^2
#'
#' all_identical <- function(...) {
#'   l <- list(...)
#'   for(i in seq_along(l)[-1])
#'     if(!identical(l[[1]], l[[i]]))
#'       return(FALSE)
#'   TRUE
#' }
#' all_identical(x1, x2, x3, x4, x5, x6, x7, x8)
#'
#' # Note, direct usage of these unit creation functions is typically not
#' # necessary, since coercion is automatically done via as_units(). Again,
#' # these are all equivalent ways to generate the same result.
#'
#' x1 <- x2 <- x3 <- x4 <- x5 <- x6 <- x7 <- x8 <- 1:4
#' units(x1) <- "m/s^2"
#' units(x2) <- "m s-2"
#' units(x3) <- quote(m/s^2)
#' units(x4) <- make_units(m/s^2)
#' units(x5) <- as_units(quote(m/s^2))
#' x6 <- set_units(x6, m/s^2)
#' x7 <- set_units(x7, "m/s^2", mode = "standard")
#' x8 <- set_units(x8, units(x1), mode = "standard")
#'
#' all_identical(x1, x2, x3, x4, x5, x6, x7, x8)
#'
#'
#' # Both unit names or symbols can be used. By default, unit names are
#' # automatically converted to unit symbols.
#' make_units(degree_C)
#' make_units(kilogram)
#' make_units(ohm)
#'
#' ## Arithmetic operations and units
#' # conversion between unit objects that were defined as symbols and names will
#' # work correctly, although unit simplification in printing may not always occur.
#' x <- 500 * make_units(micrograms/liter)
#' y <- set_units(200, ug/l)
#' x + y
#' x * y # numeric result is correct, but units not simplified completely
#'
#' # note, plural form of unit name accepted too ('liters' vs 'liter'), and
#' # denominator simplification can be performed correctly
#' x * set_units(5, liters)
#'
#' # unit conversion works too
#' set_units(x, grams/gallon)
#'
#' ## Creating custom, user defined units
#' # For example, a microbiologist might work with counts of bacterial cells
#' # make_units(cells/ml) # by default, throws an ERROR
#' # First define the unit, then the newly defined unit is accepted.
#' install_unit("cells")
#' make_units(cells/ml)
#'
#' # Note that install_unit() adds support for defining relationships between
#' # the newly created symbols or names and existing units.
#'
#' ## set_units()
#' # set_units is a pipe friendly version of `units<-`.
#' if(requireNamespace("magrittr", quietly = TRUE)) {
#'   library(magrittr)
#'   1:5 %>% set_units(N/m^2)
#'   # first sets to m, then converts to km
#'   1:5 %>% set_units(m) %>% set_units(km)
#' }
#'
#' # set_units has two modes of operation. By default, it operates with
#' # bare symbols to define the units.
#' set_units(1:5, m/s)
#'
#' # use `mode = "standard"` to use the value of supplied argument, rather than
#' # the bare symbols of the expression. In this mode, set_units() can be
#' # thought of as a simple alias for `units<-` that is pipe friendly.
#' set_units(1:5, "m/s", mode = "standard")
#' set_units(1:5, make_units(m/s), mode = "standard")
#'
#' # the mode of set_units() can be controlled via a global option
#' # units_options(set_units_mode = "standard")
#'
#' # To remove units use
#' units(x) <- NULL
#' # or
#' set_units(x, NULL)
#' # or
#' drop_units(y)
make_units <- function(bare_expression, check_is_valid = TRUE) {
  as_units.call(substitute(bare_expression), check_is_valid = check_is_valid)
}

#' @name units
#' @export
as_units <- function(x, ...) {
  UseMethod("as_units")
}

#' @name units
#' @export
as_units.default <- function(x, value = unitless, ...) {
  if (is.null(x)) return(x)
  units(x) <- value
  x
}

#' @name units
#' @export
as_units.units <- function(x, value, ...) {
  if(!missing(value) && !identical(units(value), units(x)))
    warning("Use set_units() to perform unit conversion. Return unit unmodified")
  x
}

#' @name units
#' @export
as_units.symbolic_units <- function(x, value, ...) {
  if(!missing(value))
    warning("supplied value ignored")
  .as.units(1L, x)
}

#' @examples
#' s = Sys.time()
#' d  = s - (s+1)
#' as_units(d)
#'
#' @name units
#' @export
as_units.difftime <- function(x, value, ...) {
  u <- attr(x, "units")
  x <- unclass(x)
  attr(x, "units") <- NULL

  # convert from difftime to udunits2:
  if (u == "secs") # secs -> s
    x <- x * symbolic_unit("s")
  else if (u == "mins") # mins -> min
    x <- x * symbolic_unit("min")
  else if (u == "hours") # hours -> h
    x <- x * symbolic_unit("h")
  else if (u == "days") # days -> d
    x <- x * symbolic_unit("d")
  else if (u == "weeks") { # weeks -> 7 days
    x <- 7 * x
    x <- x * symbolic_unit("d")
  } else
    stop(paste("unknown time units", u, "in difftime object"))

  if (!missing(value)) # convert optionally:
    units(x) <- value

  x
}

#  ----- as_units.character helpers ------

backtick <- function(x) {
  # backtick all character runs uninterupted by one of ^()*^/`- or a space
  # don't double up backticks
  x <- gsub("`?([^() \\*^/`-]+)`?", "`\\1`", x)
  gsub("`([0-9]*\\.?[0-9]+)`", "\\1", x) # unbacktick bare numbers
}

are_exponents_implicit <- function(s) {
  s <- trimws(s)
  has <- function(chr, regex = FALSE)
    grepl(chr, s, fixed = !regex, perl = regex)
  !has("^") && !has("*") && !has("/") && has("\\s|\\D.*\\d$", regex = TRUE)
}

is_udunits_time <- function(s) {
  ud_is_parseable(s) && ud_are_convertible(s, "seconds since 1970-01-01")
}

#' @name units
#' @export
#'
#' @param force_single_symbol Whether to perform no string parsing and force
#'   treatment of the string as a single symbol.
#'
#' @param implicit_exponents If the unit string is in product power form (e.g.
#'   \code{"km m-2 s-1"}). Defaults to \code{NULL}, in which case a guess is made
#'   based on the supplied string. Set to \code{TRUE} or \code{FALSE} if the guess is
#'   incorrect.
#'
#' @section Character strings:
#'
#'   Generally speaking, there are 3 types of unit strings are accepted in
#'   \code{as_units} (and by extension, \code{`units<-`}).
#'
#'   The first, and likely most common, is a "standard" format unit
#'   specification where the relationship between unit symbols or names is
#'   specified explicitly with arithmetic symbols for division \code{/},
#'   multiplication \code{*} and power exponents \code{^}, or other mathematical
#'   functions like \code{log()}. In this case, the string is parsed as an R
#'   expression via \code{parse(text = )} after backticking all unit symbols and
#'   names, and then passed on to \code{as_units.call()}. A heuristic is used to
#'   perform backticking, such that any continuous set of characters
#'   uninterrupted by one of \code{()\\*^-} are backticked (unless the character
#'   sequence consists solely of numbers \code{0-9}), with some care to not
#'   double up on pre-existing backticks. This heuristic appears to be quite
#'   robust, and works for units would otherwise not be valid R syntax. For
#'   example, percent (\code{"\%"}), feet (\code{"'"}), inches (\code{"in"}),
#'   and Tesla (\code{"T"}) are all backticked and parsed correctly.
#'
#'   Nevertheless, for certain complex unit expressions, this backticking heuristic
#'   may give incorrect results.  If the string supplied fails to parse as an R
#'   expression, then the string is treated as a single symbolic unit and
#'   \code{symbolic_unit(chr)} is used as a fallback with a warning. In that
#'   case, automatic unit simplification may not work properly when performing
#'   operations on unit objects, but unit conversion and other Math operations
#'   should still give correct results so long as the unit string supplied
#'   returns \code{TRUE} for \code{ud_is_parsable()}.
#'
#'   The second type of unit string accepted is one with implicit exponents. In
#'   this format, \code{/}, \code{*}, and \code{^}, may not be present in the
#'   string, and unit symbol or names must be separated by a space. Each unit
#'   symbol may optionally be followed by a single number, specifying the power.
#'   For example \code{"m2 s-2"} is equivalent to \code{"(m^2)*(s^-2)"}.
#'
#'   It must be noted that prepended numbers are supported too, but their
#'   interpretation slightly varies depending on whether they are separated from
#'   the unit string or not. E.g., \code{"1000 m"} is interpreted as magnitude
#'   and unit, but \code{"1000m"} is interpreted as a prefixed unit, and it is
#'   equivalent to \code{"km"} to all effects.
#'
#'   The third type of unit string format accepted is the special case of
#'   udunits time duration with a reference origin, for example \code{"hours
#'   since 1970-01-01 00:00:00"}. Note, that the handling of time and calendar
#'   operations via the udunits library is subtly different from the way R
#'   handles date and time operations. This functionality is mostly exported for
#'   users that work with udunits time data, e.g., with NetCDF files. Users are
#'   otherwise encouraged to use \code{R}'s date and time functionality provided
#'   by \code{Date} and \code{POSIXt} classes.
#'
as_units.character <- function(x,
                               check_is_valid = TRUE,
                               implicit_exponents = NULL,
                               force_single_symbol = FALSE, ...) {

  stopifnot(is.character(x), length(x) == 1)

  if (isTRUE(x == "")) return(unitless)

  if(force_single_symbol || is_udunits_time(x))
    return(symbolic_unit(x, check_is_valid = check_is_valid))

  if(is.null(implicit_exponents))
    implicit_exponents <- are_exponents_implicit(x)

  if(implicit_exponents)
    x <- convert_implicit_to_explicit_exponents(x)

  x <- backtick(x)
  o <- try(expr <- parse(text = x)[[1]], silent = TRUE)

  if(inherits(o, "try-error")) {
    warning("Could not parse expression: ", sQuote(x),          # nocov
      ". Returning as a single symbolic unit()", call. = FALSE) # nocov
    return(symbolic_unit(x, check_is_valid = check_is_valid))   # nocov
  }

  as_units.call(expr, check_is_valid = check_is_valid)
}


convert_implicit_to_explicit_exponents <- function(x) {
  if (length(grep(c("[*/]"), x)) > 0)
    stop("If 'implicit_exponents = TRUE', strings cannot contain `*' or `/'")
  x <- gsub("\\b([^\\d-]+)([-]?\\d+)\\b", "\\1^(\\2)", x, perl =TRUE)
  x <- gsub("\\s+", " * ", trimws(x), perl = TRUE)
  x
}

#  ----- as_units.call helpers ------

# from package:yasp, paste collapse with serial (oxford) comma
pc_and <- function(..., sep = "") {
  x <- paste(..., sep = sep, collapse = NULL)
  lx <- length(x)
  if(lx == 0L)
    ""
  else if (lx == 1L)
    x
  else if (lx == 2L)
    paste0(x, collapse = " and ")
  else
    paste0( paste0(x[-lx], collapse = ", "), ", and ", x[lx])
}

#`%not_in%` <- function(x, table) match(x, table, nomatch = 0L) == 0L

.msg_units_not_recognized <- function(unrecognized_symbols, full_expr) {

  if (is.language(full_expr))
    full_expr <- deparse(full_expr)

  is_are <- if (length(unrecognized_symbols) > 1L) "are" else "is"

  paste0("In ", sQuote(full_expr), ", ",
    pc_and(sQuote(unrecognized_symbols)), " ", is_are, " not recognized by udunits.\n\n",
    "See a table of valid unit symbols and names with valid_udunits().\n",
    "Custom user-defined units can be added with install_unit().\n\n",
    "See a table of valid unit prefixes with valid_udunits_prefixes().\n",
    "Prefixes will automatically work with any user-defined unit.")
}

units_eval_env <- new.env(parent = baseenv())
units_eval_env$ln <- function(x) base::log(x)
units_eval_env$lg <- function(x) base::log(x, base = 10)
units_eval_env$lb <- function(x) base::log(x, base = 2)


#' @name units
#' @export
#'
#' @param check_is_valid throw an error if all the unit symbols are not either
#'   recognized by udunits2 via \code{ud_is_parseable()}, or a custom
#'   user defined via \code{install_unit()}. If \code{FALSE}, no check
#'   for validity is performed.
#'
#' @note By default, unit names are automatically substituted with unit names
#'   (e.g., kilogram --> kg). To turn off this behavior, set
#'   \code{units_options(auto_convert_names_to_symbols = FALSE)}
#'
#' @section Expressions:
#'
#'   In \code{as_units()}, each of the symbols in the unit expression is treated
#'   individually, such that each symbol must be recognized by the udunits
#'   database (checked by \code{ud_is_parseable()}, \emph{or} be a custom,
#'   user-defined unit symbol that was defined by \code{install_unit()}. To
#'   see which symbols and names are currently recognized by the udunits
#'   database, see \code{valid_udunits()}.
#'
#' @seealso \code{\link{install_unit}}, \code{\link{valid_udunits}}
as_units.call <- function(x, check_is_valid = TRUE, ...) {

  if(missing(x) || identical(x, quote(expr =)) ||
     identical(x, 1) || identical(x, 1L))
    return(.as.units(1, unitless))

  if (is.vector(x) && !is.expression(x) && any(is.na(x)))
  	stop("a missing value for units is not allowed")

  stopifnot(is.language(x))

  vars <- all.vars(x)
  if(!length(vars))
    stop(call. = FALSE,
"No symbols found. Please supply bare expressions with this approach.
See ?as_units for usage examples.")

  if (check_is_valid) {
    valid <- vapply(vars, ud_is_parseable, logical(1L))
    if (!all(valid))
      stop(.msg_units_not_recognized(vars[!valid], x), call. = FALSE)
  }

  names(vars) <- vars
  tmp_env <- lapply(vars, symbolic_unit, check_is_valid = FALSE)

  if (dont_simplify_here <- is.na(.units.simplify())) {
  	units_options(simplify = FALSE)
  	on.exit(units_options(simplify = NA))
  }

  unit <- tryCatch( eval(x, tmp_env, units_eval_env),
    error = function(e) stop( paste0( conditionMessage(e), "\n",
          "Did you try to supply a value in a context where a bare expression was expected?"
        ), call. = FALSE ))

#  if(as.numeric(unit) %not_in% c(1, 0)) # 0 if log() used.
#    stop(call. = FALSE,
#"In ", sQuote(deparse(x)), " the numeric multiplier ", sQuote(as.numeric(unit)), " is invalid.
#Use `install_unit()` to define a new unit that is a multiple of another unit.")

  .as.units(as.numeric(unit), units(unit))
}

#' @name units
#' @export
as_units.expression <- as_units.call

#' @name units
#' @export
as_units.name       <- as_units.call

#' @name units
#' @export
as_units.POSIXt = function(x, value, ...) {
  u = as.numeric(as.POSIXct(x))
  units(u) = symbolic_unit("seconds since 1970-01-01 00:00:00 +00:00")
  if (! missing(value))
    units(u) = symbolic_unit(value)
  u
}

#' @name units
#' @export
as_units.Date = function(x, value, ...) {
  u = as.numeric(x)
  units(u) = symbolic_unit("days since 1970-01-01")
  if (!missing(value))
    units(u) = symbolic_unit(value)
  u
}


symbolic_unit <- function(chr, check_is_valid = TRUE) {

  stopifnot(is.character(chr), length(chr) == 1)

  if (check_is_valid && !ud_is_parseable(chr)) {
    msg <- paste(sQuote(chr), "is not a unit recognized by udunits or a user-defined unit")
    stop(msg, call. = FALSE)
  }

  auto_convert <- units_options("auto_convert_names_to_symbols")
  if (auto_convert && ud_is_parseable(chr)) {
    sym <- ud_get_symbol(chr)
    if (length(sym))
      chr <- sym
  }

  .as.units(1, .symbolic_units(chr))
}

test_that("we can call math functions on units", {
  x <- 1:4 - 2.1
  ux <- x * as_units("m")

  expect_equal(as.numeric(abs(ux)), abs(x))
  expect_equal(units(abs(ux)), units(ux))

  expect_equal(as.numeric(sign(ux)), sign(x))
  expect_true(!inherits(sign(ux), "units"))

  expect_equal(as.numeric(sqrt(ux^2)), sqrt(x^2))
  expect_error(sqrt(ux), "powers not divisible")

  expect_equal(as.numeric(floor(ux)), floor(x))
  expect_equal(units(floor(ux)), units(ux))
  expect_equal(as.numeric(ceiling(ux)), ceiling(x))
  expect_equal(units(ceiling(ux)), units(ux))
  expect_equal(as.numeric(trunc(ux)), trunc(x))
  expect_equal(units(trunc(ux)), units(ux))
  expect_equal(as.numeric(round(ux)), round(x))
  expect_equal(units(round(ux)), units(ux))
  expect_equal(as.numeric(signif(ux)), signif(x))
  expect_equal(units(signif(ux)), units(ux))

  expect_equal(as.numeric(cumsum(ux)), cumsum(x))
  expect_equal(units(cumsum(ux)), units(ux))
  expect_equal(as.numeric(cummax(ux)), cummax(x))
  expect_equal(units(cummax(ux)), units(ux))
  expect_equal(as.numeric(cummin(ux)), cummin(x))
  expect_equal(units(cummin(ux)), units(ux))

  expect_warning(y <- cos(ux))
  expect_equal(y, cos(x))
  expect_equal(class(y), "numeric")

  expect_equal(sin(set_units(1, rad)), set_units(sin(1)))
  expect_equal(sin(set_units(1, 1)), set_units(sin(1)))
  expect_equal(sin(set_units(90, degree)), sin(set_units(pi/2, rad)))
  expect_equal(units(acos(set_units(-1))), units(make_units(rad)))
  expect_equal(set_units(acos(set_units(-1)), degree), set_units(180, degree))
})

test_that("we can take logarithms units", {
  x <- 1:4
  ux <- x * as_units("m")

  expect_equal(as.numeric(log2(ux)), log2(x))
  expect_equal(units(log2(ux)), units(as_units("lb(re 1 m)", force_single_symbol=TRUE)))

  expect_equal(as.numeric(log(ux, 2)), log(x, 2))
  expect_equal(units(log(ux, 2)), units(as_units("lb(re 1 m)", force_single_symbol=TRUE)))

  expect_equal(as.numeric(log10(ux)), log10(x))
  expect_equal(units(log10(ux)), units(as_units("lg(re 1 m)", force_single_symbol=TRUE)))

  expect_equal(as.numeric(log(ux, 10)), log(x, 10))
  expect_equal(units(log(ux, 10)), units(as_units("lg(re 1 m)", force_single_symbol=TRUE)))

  expect_equal(as.numeric(log1p(ux)), log1p(x))
  expect_equal(units(log1p(ux)), units(as_units("ln(re 1 m)", force_single_symbol=TRUE)))

  expect_equal(as.numeric(log(ux)), log(x))
  expect_equal(units(log(ux)), units(as_units("ln(re 1 m)", force_single_symbol=TRUE)))

  expect_equal(as.numeric(log(ux, exp(2))), log(x, exp(2)))
  expect_equal(units(log(ux, exp(2))), units(as_units("0.5 ln(re 1 m)", force_single_symbol=TRUE)))
})

test_that("decibel is available", {
  expect_true(units:::ud_is_parseable("B"))
  units_options(define_bel=FALSE)
  expect_false(units:::ud_is_parseable("B"))
  units_options(define_bel=TRUE)
  expect_true(units:::ud_is_parseable("B"))

  x <- set_units(set_units(100, 1), dB)
  expect_equal(as.numeric(x), 20)
  expect_equal(as.character(units(x)), "dB")
})

test_that("we can compute summary functions on units", {
  x <- 1:4
  ux <- x * as_units("m")

  expect_equal(as.numeric(sum(ux)), sum(x))
  expect_equal(as.numeric(min(ux)), min(x))
  expect_equal(as.numeric(max(ux)), max(x))
  expect_equal(as.numeric(range(ux)), range(x))

  expect_equal(units(sum(ux)), units(ux))
  expect_equal(units(min(ux)), units(ux))
  expect_equal(units(max(ux)), units(ux))
  expect_equal(units(range(ux)), units(ux))

  expect_error(all(ux))
  expect_error(sum(ux, x))

  y <- 1:4
  uy <- y * as_units("km")
  expect_equal(as.numeric(sum(ux, uy)), sum(c(x, 1000*y)))
  expect_equal(as.numeric(min(ux, uy)), min(c(x, 1000*y)))
  expect_equal(as.numeric(max(ux, uy)), max(c(x, 1000*y)))
  expect_equal(as.numeric(range(ux, uy)), range(c(x, 1000*y)))
  z <- 1:4 * as_units("s")
  expect_error(sum(ux, z))
})

test_that("we can compute means and medians and quantiles", {
  x <- 1:4
  ux <- x * as_units("m")
  w <- x / 5:8
  uw <- w * as_units("g")

  expect_equal(as.numeric(mean(ux)), mean(x))
  expect_equal(as.numeric(median(ux)), median(x))
  expect_equal(as.numeric(quantile(ux)), quantile(x), ignore_attr = TRUE)
  expect_equal(as.numeric(weighted.mean(ux, w)), weighted.mean(x, w))
  expect_equal(as.numeric(weighted.mean(ux, uw)), weighted.mean(x, w))
})

test_that("we can format units", {
  x <- 1:4
  ux <- x * as_units("m")

  expect_equal(format(ux), c("1 [m]", "2 [m]", "3 [m]", "4 [m]"))
})
test_that("summary works with NA values", {
  x <- c(1, NA) * as_units("g")
  expect_equal(summary(x), summary(as.numeric(x)))
})

test_that("global options are respected", {
  # rt: round trip
  rt <- function(x) as.character(units(as_units(x)))

  expect_equal("g", rt("grams"))

  #op <- units_options(auto_convert_names_to_symbols = FALSE)
  units_options(auto_convert_names_to_symbols = FALSE)
  expect_equal("grams", rt("grams"))

  units_options(auto_convert_names_to_symbols = TRUE)
  expect_equal("g", rt("grams"))

  #units_options(op)
  expect_equal("g", rt("grams"))

  o = units_options("set_units_mode")
  g = set_units(1, g)
  units_options(set_units_mode = "standard")
  expect_equal(set_units(1, "g"), g)
  units_options(set_units_mode = o)

})


test_that("various forms of unit creation and destruction work", {

  #ox <- x <- 1L:4L # why should ints be preserved?
  ox <- x <- as.numeric(1:4)

  units(x) <- "m/s"
  expect_s3_class(x,"units")

  units(x) <- NULL
  expect_identical(x, ox)

  units(x) <- character(0)
  expect_identical(x, ox)

  ox <- x <- matrix(ox, ncol = 2)

  units(x) <- "m/s"
  expect_s3_class(x,"units")

  units(x) <- NULL
  expect_identical(x, ox)

  units(x) <- character(0)
  expect_identical(x, ox)

  ox <- y <- x <- as.numeric(1:4)
  units(x) <- "m/s"

  expect_identical(x, set_units(y, m/s))
  expect_identical(x, set_units(y, "m/s", mode = "standard"))

  expect_identical(ox, set_units(x, NULL))
  expect_identical(ox, set_units(x, NULL, mode = "standard"))


  ox <- x <- as.numeric(1:4)

  units(x) <- "m/s"
  expect_s3_class(x, "units")

  x <- drop_units(x)
  expect_identical(x, ox)
  expect_error(drop_units(x))


  ox <- x <- data.frame(x=ox, y=ox)

  units(x[[1]]) <- "m/s"
  expect_s3_class(x[[1]], "units")

  x <- drop_units(x)
  expect_identical(x, ox)


  meter <- units:::symbolic_unit("m")

  expect_identical(meter, make_units(m))
  expect_identical(meter, as_units("m"))
  expect_identical(meter, as_units(quote(m)))
  expect_identical(meter, as_units(expression(m)))

  meter_per_sec <- meter/units:::symbolic_unit("s")

  expect_identical(meter_per_sec, make_units(m/s))
  expect_identical(meter_per_sec, as_units("m/s"))
  expect_identical(meter_per_sec, as_units(" m / s "))
  expect_identical(meter_per_sec, as_units(quote(m/s)))
  expect_identical(meter_per_sec, as_units(expression(m/s)))
  expect_identical(meter_per_sec, as_units("m s-1"))

})

test_that("unitless objects can be created", {
  unit <- 1.0
  units(unit) <-  units:::unitless

  xul <- x <- as.numeric(2:5)
  units(xul) <-  units:::unitless


  expect_identical(unit, make_units())
  expect_identical(unit, make_units(1))

  expect_identical(unit, as_units(1))
  expect_identical(unit, as_units("1"))

  expect_identical(xul, set_units(x, 1))
  expect_identical(xul, set_units(x))
  expect_identical(xul, set_units(x, "1"))
  expect_identical(xul, set_units(x, "1", mode = "standard"))
  expect_identical(xul, set_units(x, mode = "standard"))
})

test_that("set_units default enforces NSE", {
  expect_error(set_units(1:3, as_units("m")))
  expect_error(set_units(1:3, as_units("m/s")))
  expect_error(set_units(1:3, make_units(m)))
  expect_error(set_units(1:3, make_units(m/s)))

  # is it bad if this works?
  # expect_error(set_units(1:3, "m/s"))
})

expect_symbolic <- function(u, n, d)
  expect_equal(units(as_units(u)), units:::.symbolic_units(n, d))
expect_symbolic_nocheck <- function(u, n, d)
  expect_equal(units(as_units(u, check_is_valid=FALSE)), units:::.symbolic_units(n, d))

test_that("exotic units work", {
  # check what udunits support
  # units:::R_ut_format(units:::R_ut_parse(some_string))

  expect_symbolic("2.2 m s", c("2.2", "m", "s"), character(0))
  expect_symbolic("2.2*m*s", c("2.2", "m", "s"), character(0))
  #expect_symbolic("2.2.m.s", c("2.2", "m", "s"), character(0))

  expect_symbolic("m2/s", c("m", "m"), "s")
  expect_symbolic("m20/s", rep("m", 20), "s")
  expect_symbolic("m^2/s", c("m", "m"), "s")
  expect_symbolic("m 2/s", c("m", "2"), "s")
  expect_symbolic("m-2/s", character(0), c("m", "m", "s"))
  expect_symbolic("m^-2/s", character(0), c("m", "m", "s"))
  expect_symbolic("m/s2", "m", c("s", "s"))
  expect_symbolic("m/s^2", "m", c("s", "s"))
  expect_symbolic("m/s 2", c("m", "2"), "s")
  expect_symbolic("m/s-2", c("m", "s", "s"), character(0))
  expect_symbolic("m/s^-2", c("m", "s", "s"), character(0))

  expect_symbolic("ml/min/(1.73m^2)", "ml", c("min", "1.73", "m", "m"))
  expect_symbolic("ml/min/1.73/m^2", "ml", c("min", "1.73", "m", "m"))
  expect_symbolic("ml/min/1.73m^2", "ml", c("min", "1.73", "m", "m"))
  expect_symbolic("ml/min/1.73m-2", c("ml", "m", "m"), c("min", "1.73"))

  old <- unlist(units_options(strict_tokenizer=TRUE))
  expect_symbolic("ml/min/1.73m^2", c("ml", "m", "m"), c("min", "1.73"))
  expect_symbolic("ml/min/1.73m-2", "ml", c("min", "1.73", "m", "m"))
  units_options(strict_tokenizer=old)

  expect_symbolic_nocheck("inH2O", "inH2O", character(0))
  expect_symbolic_nocheck("inH2O2", c("inH2O", "inH2O"), character(0))

  expect_symbolic("m/(g/(s/L))", c("m", "s"), c("g", "L"))
  expect_error(expect_warning(as_units("m/(g/s")))
  expect_error(expect_warning(as_units("m^m")))
})

#' Continuous scales for units data
#'
#' These are the default scales for the \code{units} class. These will usually
#' be added automatically. To override manually, use \code{scale_{type}_units}.
#'
#' @param ... arguments passed on to the corresponding continuous scale
#' (see the manual page for each \code{scale_{type}} for details).
#' @inheritParams ggplot2::scale_x_continuous
#' @param unit A unit specification to use for the guide. If given, the values
#' will be converted to this unit before plotting. An error will be thrown if
#' the specified unit is incompatible with the unit of the data.
#'
#' @name scale_units
#' @aliases NULL
#'
#' @examples
#' if (requireNamespace("ggplot2", quietly=TRUE)) {
#'
#' library(ggplot2)
#'
#' mtcars$consumption <- set_units(mtcars$mpg, mi / gallon)
#' mtcars$power <- set_units(mtcars$hp, hp)
#'
#' # Use units encoded into the data
#' ggplot(mtcars) +
#'   geom_point(aes(power, consumption))
#'
#' # Convert units on the fly during plotting
#' ggplot(mtcars) +
#'   geom_point(aes(power, consumption)) +
#'   scale_x_units(unit = "W") +
#'   scale_y_units(unit = "km/l")
#'
#' # Resolve units when transforming data
#' ggplot(mtcars) +
#'   geom_point(aes(power, 1 / consumption))
#'
#' # Reverse the y axis
#' ggplot(mtcars) +
#'   geom_point(aes(power, consumption)) +
#'   scale_y_units(transform="reverse")
#'
#' }
NULL

#' @rdname scale_units
#' @export
scale_x_units <- function(..., sec.axis = ggplot2::waiver(), unit = NULL) {
  sc <- make_scale_units(ggplot2::scale_x_continuous(...), unit)
  set_sec_axis(sec.axis, sc)
}

#' @rdname scale_units
#' @export
scale_y_units <- function(..., sec.axis = ggplot2::waiver(), unit = NULL) {
  sc <- make_scale_units(ggplot2::scale_y_continuous(...), unit)
  set_sec_axis(sec.axis, sc)
}

#' @rdname scale_units
#' @export
scale_colour_units <- function(..., unit = NULL) {
  make_scale_units(ggplot2::scale_colour_continuous(...), unit)
}

#' @rdname scale_units
#' @export
scale_color_units <- scale_colour_units

#' @rdname scale_units
#' @export
scale_fill_units <- function(..., unit = NULL) {
  make_scale_units(ggplot2::scale_fill_continuous(...), unit)
}

#' @rdname scale_units
#' @export
scale_alpha_units <- function(..., unit = NULL) {
  make_scale_units(ggplot2::scale_alpha(...), unit)
}

#' @rdname scale_units
#' @export
scale_size_units <- function(..., unit = NULL) {
  make_scale_units(ggplot2::scale_size(...), unit)
}

#' @rdname scale_units
#' @export
scale_size_area_units <- function(..., unit = NULL) {
  make_scale_units(ggplot2::scale_size_area(...), unit)
}

#' @rdname scale_units
#' @export
scale_radius_units <- function(..., unit = NULL) {
  make_scale_units(ggplot2::scale_radius(...), unit)
}

#' @rdname scale_units
#' @export
scale_linewidth_units <- function(..., unit = NULL) {
  make_scale_units(ggplot2::scale_linewidth(...), unit)
}

make_scale_units <- function(parent, unit) {
  if (!requireNamespace("ggplot2", quietly=TRUE))
    stop("package 'ggplot2' is required for this functionality", call.=FALSE)

  ggplot2::ggproto(
    paste0(class(parent)[1], "Units"),
    parent,
    units = as_units(unit),

    set_convert = function(self, x) {
      if (!inherits(x, "units"))
        return(x)

      if (is.null(self$units))
        self$units <- units(x)
      else units(x) <- as_units(1, self$units)
      drop_units(x)
    },

    map = function(self, x, limits = self$get_limits()) {
      x <- self$set_convert(x)
      ggplot2::ggproto_parent(parent, self)$map(x, limits)
    },

    train = function(self, x) {
      x <- self$set_convert(x)
      ggplot2::ggproto_parent(parent, self)$train(x)
    },

    transform = function(self, x) {
      if (!is.null(self$units))
        units(x) <- as_units(1, self$units)

      if (inherits(self$limits, "units")) {
        units(self$limits) <- units(x)
        self$limits <- drop_units(self$limits)
      }

      new_x <- ggplot2::ggproto_parent(parent, self)$transform(drop_units(x))
      as_units(new_x, units(x))
    },

    make_title = function(self, guide_title, scale_title, label_title) {
      if (missing(label_title)) { # ggplot2 <= 3.5.1
        title <- guide_title
        if (!is.null(title))
          title <- make_unit_label(title, as_units(1, self$units))
        return(title)
      }

      if (!is.null(label_title))
        label_title <- make_unit_label(label_title, as_units(1, self$units))
      ggplot2::ggproto_parent(parent, self)$make_title(
        guide_title, scale_title, label_title)
    }
  )
}

set_sec_axis <- function(sec.axis, scale) {
  if (!inherits(sec.axis, "waiver")) {
    if (inherits(sec.axis, "formula"))
      sec.axis <- ggplot2::sec_axis(sec.axis)
    if (!inherits(sec.axis, "AxisSecondary"))
      stop("Secondary axes must be specified using \"sec_axis()\"", call.=FALSE)
    sec.axis$units <- scale$units
    scale$secondary.axis <- sec.axis
  }
  scale
}

# registered in .onLoad()
scale_type.units <- function(x) {
  if (!"units" %in% .packages())
    stop("Variable of class 'units' found, but 'units' package is not attached.\n",
         "  Please, attach it using 'library(units)' to properly show scales with units.")
  c("units", "continuous")
}

utils::globalVariables("caller_env")

# registered in .onLoad()
limits.units <- function(lims, var, call = caller_env())  {
  if (length(lims) != 2)
    stop("`", var, "` must be a two-element vector", call.=FALSE)

  trans <- if (!any(is.na(lims)) && lims[1] > lims[2])
    "reverse" else "identity"
  name <- paste0("scale_", var, "_units")
  sc <- match.fun(name)(limits=lims, transform=trans)
  sc$call <- if (!is.null(call)) call else str2lang(paste0(name, "()"))
  sc
}

do.call(units_options, units:::.default_options)

test_that("base plots work as expected", {
  skip_if_not_installed("vdiffr")
  fplot <- function(...) function() plot(...)

  displacement = mtcars$disp * as_units("in")^3
  units(displacement) = make_units(cm^3)
  weight = mtcars$wt * 1000 * make_units(lb)
  units(weight) = make_units(kg)
  vdiffr::expect_doppelganger("plot default", fplot(weight, displacement))

  xlab <- "some other thing"
  vdiffr::expect_doppelganger("plot lab", fplot(weight, displacement, xlab=xlab))

  units_options(group = c("(", ")") )  # parenthesis instead of square brackets
  vdiffr::expect_doppelganger("plot parentheses", fplot(weight, displacement))

  units_options(sep = c("~~~", "~"), group = c("", ""))  # no brackets; extra space
  vdiffr::expect_doppelganger("plot nothing", fplot(weight, displacement))

  units_options(sep = c("~", "~~"), group = c("[", "]"))
  gallon = as_units("gallon")
  consumption = mtcars$mpg * make_units(mi/gallon)
  units(consumption) = make_units(km/l)
  vdiffr::expect_doppelganger("plot division", fplot(displacement, consumption))

  units_options(negative_power = TRUE) # division becomes ^-1
  vdiffr::expect_doppelganger("plot npower", fplot(displacement, consumption))
  vdiffr::expect_doppelganger("plot inverse", fplot(1/displacement, 1/consumption))

  units_options(parse = FALSE)
  n = 100
  set.seed(42)
  u = rnorm(1:n) * as_units("degree_C")
  v = rnorm(1:n) * as_units("s")
  vdiffr::expect_doppelganger("plot degree_C", fplot(u, v))

  vdiffr::expect_doppelganger("plot line", fplot(u, type = 'l'))
  vdiffr::expect_doppelganger("plot hist", function() hist(u))
  vdiffr::expect_doppelganger("plot boxplot", function() boxplot(u))
})

test_that("ggplot2 plots work as expected", {
  skip_if_not_installed("vdiffr")
  skip_if_not_installed("ggplot2", "3.5.0")
  library(ggplot2)

  iris.u <- iris
  iris.u[1:4] <- lapply(iris.u[1:4], function(x) set_units(x, cm))

  p0 <- ggplot(iris.u) + aes(Sepal.Length, Sepal.Width, color=Species) +
    geom_point() + theme_bw() + theme(legend.position="inside",
                                      legend.position.inside=c(0.6, 0.8))

  p1 <- p0 + scale_x_units(unit="m") + scale_y_units(unit="mm")
  p2 <- p0 + xlab("some other thing")
  p3 <- p0 + xlab(NULL)

  vdiffr::expect_doppelganger("ggplot2 automatic", p0)
  vdiffr::expect_doppelganger("ggplot2 manual", p1)
  vdiffr::expect_doppelganger("ggplot2 lab", p2)
  vdiffr::expect_doppelganger("ggplot2 nolab", p3)
})

test_that("axis transformations do not affect displayed units", {
  skip_if_not_installed("vdiffr")
  skip_if_not_installed("ggplot2", "3.5.0")
  library(ggplot2)

  df <- data.frame(a = set_units(1:10, "m"))

  p0 <- ggplot(df, aes(y=a, x=a)) + geom_point()
  p1 <- p0 + scale_x_units(transform='log10') + scale_y_units(transform='sqrt')
  p2 <- p0 + scale_x_units(transform="log10", unit="mm")

  vdiffr::expect_doppelganger("ggplot2 default", p0)
  vdiffr::expect_doppelganger("ggplot2 transformed", p1)
  vdiffr::expect_doppelganger("ggplot2 trans + unit", p2)
})

test_that("axis limits can be changed", {
  skip_if_not_installed("vdiffr")
  skip_if_not_installed("ggplot2", "3.5.0")
  library(ggplot2)

  df <- data.frame(a = set_units(1:10, "m"))

  p0 <- ggplot(df, aes(y=a, x=a)) + geom_point()
  p1 <- p0 + scale_x_units(limits=c(0, 20))
  p2 <- p0 + scale_x_units(limits=set_units(c(0, 20), "m"))
  p3 <- p0 + xlim(set_units(c(0, 20), "m"))
  p4 <- p0 + xlim(set_units(c(0, 20), "km"))

  vdiffr::expect_doppelganger("ggplot2 limits via scale", p1)
  vdiffr::expect_doppelganger("ggplot2 limits via scale with units", p2)
  vdiffr::expect_doppelganger("ggplot2 limits via xlim", p3)
  vdiffr::expect_doppelganger("ggplot2 limits other units", p4)
})

test_that("ranges are correctly computed", {
  skip_if_not_installed("vdiffr")
  skip_if_not_installed("ggplot2", "3.5.0")
  library(ggplot2)

  p0 <- data.frame(x = 0, y = make_units(m) * c(1, 2)) |>
    ggplot() + aes(x, y) + geom_point()

  p <- data.frame(x = 0, y = make_units(m) * 1) |>
    ggplot() + aes(x, y) + geom_point()
  p1 <- p + geom_point(data = data.frame(x = 0, y = make_units(mm) * 2000))
  p2 <- p + geom_point(data = data.frame(x = 0, y = make_units(m)  * 2))
  p3 <- p + geom_point(data = data.frame(x = 0, y = make_units(km) * 0.002))

  vdiffr::expect_doppelganger("ggplot2 range", p0)
  vdiffr::expect_doppelganger("ggplot2 range", p1)
  vdiffr::expect_doppelganger("ggplot2 range", p2)
  vdiffr::expect_doppelganger("ggplot2 range", p3)
})

test_that("additional continuous scales work as expected", {
  skip_if_not_installed("vdiffr")
  skip_if_not_installed("ggplot2", "3.5.0")
  library(ggplot2)

  dat <- data.frame(x=rnorm(100), y=rnorm(100), z=set_units(rnorm(100), m))

  p1 <- ggplot(dat) + aes(x, y, color=z) +
    geom_point() + scale_color_units(unit="in")

  p2 <- ggplot(dat) + aes(x, y, fill=z) +
    geom_point() + scale_fill_units(unit="mm")

  p3 <- ggplot(dat) + aes(x, y, alpha=z) + geom_point() +
    scale_alpha_units(unit="feet")

  p4 <- ggplot(dat) + aes(x, y, size=z) +
    geom_point() + scale_size_units(unit="km")

  p5 <- ggplot(dat) + aes(x, y, size=z) +
    geom_point() + scale_size_area_units(unit="km")

  p6 <- ggplot(dat) + aes(x, y, size=z) +
    geom_point() + scale_radius_units(unit="km")

  p7 <- ggplot(dat) + aes(x, y, linewidth=z) +
    geom_path() + scale_linewidth_units(unit="cm")

  vdiffr::expect_doppelganger("ggplot2 other color", p1)
  vdiffr::expect_doppelganger("ggplot2 other fill", p2)
  vdiffr::expect_doppelganger("ggplot2 other alpha", p3)
  vdiffr::expect_doppelganger("ggplot2 other size", p4)
  vdiffr::expect_doppelganger("ggplot2 other size area", p5)
  vdiffr::expect_doppelganger("ggplot2 other radius", p6)
  vdiffr::expect_doppelganger("ggplot2 other linewidth", p7)
})

do.call(units_options, units:::.default_options)

#' @import utils
#' @import stats
#' @import graphics
#' @importFrom Rcpp evalCpp
#' @useDynLib units
NULL

.onLoad = function(libname, pkgname) {
  load_units_xml()

  if (ud_is_parseable("B"))
    .default_options$define_bel <- FALSE
  do.call(units_options, .default_options)

  native <- if (l10n_info()[["UTF-8"]]) "utf8"
  else if (l10n_info()[["Latin-1"]]) "latin1"
  else "ascii"
  ud_set_encoding(native)

  register_all_s3_methods()
}

.onAttach <- function(libname, pkgname) {
  packageStartupMessage(.startup_msg(TRUE))
}

.onUnload = function(libpath) {
  ud_exit()
}

#' \pkg{udunits2} utilities
#'
#' Some \pkg{udunits2} utilities are exposed to the user. These functions are
#' useful for checking whether units are convertible or converting between units
#' without having to create \pkg{units} objects.
#' Arguments are recycled if necessary.
#'
#' @param from,to character vector or object of class \code{symbolic_units},
#' for the symbol(s) of the original unit(s) and the unit to convert to respectively.
#' @param ...  unused.
#'
#' @return \code{ud_are_convertible}
#' returns \code{TRUE} if both units exist and are convertible,
#' \code{FALSE} otherwise.
#'
#' @name udunits2
#' @export
ud_are_convertible <- function(from, to, ...) {
  if (length(dots <- list(...))) {
    if (exists("x", dots)) from <- dots$x
    if (exists("y", dots)) to   <- dots$y
    warning("variables `x` and `y` were unfortunate names, and are deprecated",
            "; please use `from` and `to` instead")
  }
  mapply(ud_convertible, ud_char(from), ud_char(to), USE.NAMES=FALSE)
}

#' @param x numeric vector
#'
#' @return \code{ud_convert}
#' returns a numeric vector with \code{x} converted to new unit.
#'
#' @name udunits2
#' @export
#'
#' @examples
#' ud_are_convertible(c("m", "mm"), "km")
#' ud_convert(c(100, 100000), c("m", "mm"), "km")
#'
#' a <- set_units(1:3, m/s)
#' ud_are_convertible(units(a), "km/h")
#' ud_convert(1:3, units(a), "km/h")
#'
#' ud_are_convertible("degF", "degC")
#' ud_convert(32, "degF", "degC")
ud_convert <- function(x, from, to) {
  mapply(ud_convert_doubles, x, ud_char(from), ud_char(to))
}

ud_char <- function(x) {
  if (is.character(x)) return(x)
  if (!inherits(x, "symbolic_units")) stop("not a unit")

  res <- if (length(x$numerator))
    paste(x$numerator, collapse=" ") else "1"
  if (length(x$denominator))
    res <- paste0(res, " (", paste(x$denominator, collapse=" "), ")-1")
  res
}

ud_are_same <- function(x, y) {
  get_base <- function(x)
    tail(strsplit(R_ut_format(R_ut_parse(x), definition=TRUE), " ")[[1]], 1)
  identical(get_base(x), get_base(y))
}

ud_get_symbol = function(u) {
  u <- R_ut_parse(u)
	sym = R_ut_get_symbol(u)
	if (!length(sym))
		sym = R_ut_get_name(u)
	sym
}

ud_is_parseable = function(u) {
	res <- try(R_ut_parse(u), silent = TRUE)
	! inherits(res, "try-error")
}

ud_parse <- function(u, names=FALSE, definition=FALSE, ascii=FALSE) {
  R_ut_format(R_ut_parse(u), names, definition, ascii)
}

test_that("ud_are_convertible return the expected value", {
  x <- 1:10 * as_units("m")
  expect_type(ud_are_convertible("m", "km"), "logical")
  expect_true(ud_are_convertible("m", "km"))
  expect_true(ud_are_convertible(units(x), "km"))
  expect_false(ud_are_convertible("s", "kg"))

  x <- c("m", "l")
  y <- c("km", "ml", "cm", "kg")
  conv <- c(TRUE, TRUE, TRUE, FALSE)
  expect_equal(ud_are_convertible(x, y), conv)
  expect_equal(ud_are_convertible(y, x), conv)
})

test_that("ud_convert works with simple conversions", {
  x <- 1:10 * as_units("m")
  expect_equal(ud_convert(1, "m", "km"), 1/1000)
  expect_equal(ud_convert(as.numeric(x), units(x), "km"), as.numeric(x)/1000)
  expect_equal(ud_convert(1, "km", "m"), 1000)
  expect_equal(ud_convert(32, "degF", "degC"), 0)
  expect_equal(ud_convert(0, "degC", "K"), 273.15)
})

test_that("ud_convert works with vectors", {
  expect_equal(ud_convert(1:2, c("m", "mm"), "km"), 1:2/c(1e3,1e6))
  expect_equal(ud_convert(c(32, 212), "degF", "degC"), c(0, 100))
})

test_that("ud_convert returns Error for incompatible units", {
  expect_error(ud_convert(100, "m", "kg"), "Units not convertible")
})

/*
 *  Make a program object for the OS/2 version of units
 *  Peter Weilbacher (os2@Weilbacher.org), 25Jan2003
 *
 *  Copyright (C) 1996, 1997, 1999 Free Software Foundation, Inc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *     
 */
 
call RxFuncAdd SysLoadFuncs, REXXUtil, SysLoadFuncs
call SysLoadFuncs

workdir = Directory()
classname='WPProgram'

title = 'units'
location = '<WP_DESKTOP>'
setup = 'OBJECTID=<UNITS>;'||,
	'EXENAME=CMD.EXE;'||,
	'PARAMETERS=/c mode co50,5 & units.exe [from-unit] [to-unit] & pause;'||,
	'STARTUPDIR='workdir
if SysCreateObject(classname, title, location, setup, 'U') then 
   say 'Object created.'
else
   say 'Could NOT create the object!!'

title = 'units Shell'
location = '<WP_DESKTOP>'
setup = 'OBJECTID=<UNITSSHELL>;'||,
	'EXENAME=CMD.EXE;'||,
	'PARAMETERS=/c mode co80,25 & units.exe;'||,
	'STARTUPDIR='workdir
if SysCreateObject(classname, title, location, setup, 'U') then 
   say 'Object created.'
else
   say 'Could NOT create the object!!'

# set MKS Korn shell environment variables for MS Visual Studio
# 2015/2019/2022 Community

# Revision history
#            2023-09-08: changed to work with Visual Studio 2022:
#                        replaced here document with temorary file;
#                        improved error messages
#            2022-09-04: changed "egrep" to "grep -E"
#            2021-09-26: revised to use Visual Studio 2019
#            2020-05-08: revised to prevent MKS Toolkit from prepending
#                        UTF-8 signature
#            2016-05-28: initial version
#
# Copyright (C) 2016, 2020, 2022, 2023 Free Software Foundation, Inc

# Microsoft Visual Studio requires that several environment variables
# include many directories if a build is to be done from the command
# line.  These variables are normally set by selecting 'Developer
# Command Prompt' or one of the more specific options on the Windows
# Start Menu under the appropriate Visual Studio version; the shortcut
# runs a batch file that calls several other batch files to set the
# variables before launching an instance of the Windows command
# interpreter.  This program calls the Windows command interpreter to
# run the batch file; the resulting values of the variables are echoed
# and read into the shell to set the variables in the shell.  For the
# values to persist, this program must be run in the current
# environment, i.e.,
#
#   source setvcvars
#
#       or
#
#   . setvcvars
#
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY--without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-
# 1301 USA
#
# 
# This program was written by Jeff Conrad (jeff_conrad@msn.com), and
# tested with the MKS Toolkit version 10.4 and Microsoft Visual Studio
# 2015, 2019, and 2022 on Windows 10 Professional.

tempdir="$TEMP"
tempfile="$tempdir\setvctemp$$.bat"
progname=setvcvars

# this will match progname if not invoked with source
thisprog=${0##*[/\\]}
thisprog=${thisprog%.sh}

if [[ $thisprog == $progname ]]
then
    print -r "Environment not initialized; run as 'source $thisprog'"
    exit 1
fi

setvcerrors=

umsg="Usage: source $progname [options] [platform]
Options:
  -V<vsversion>  Use Visual Studio <vsversion> (default: 2022)"

# default version; change to suit
vsversion=2022
vcverbose=

DUALCASE=1	# MKS Toolkit option to distinguish upper and lower case
while getopts :V: arg
do
    case $arg in
    v)
	vcverbose=YES ;;	# not yet used
    V)
	vsversion=$OPTARG ;;
    :)
	# OPTARG contains the option missing the argument
	print -ru2 -- "$progname: option $OPTARG requires an argument"
	setvcerrors=YES
	;;
    [?])
	# OPTARG contains the invalid option
	print -ru2 -- "$progname: unknown option $OPTARG"
	setvcerrors=YES
	;;
    esac
done
shift OPTIND-1
unset DUALCASE

if [ -n "$setvcerrors" ]
then
    print -ru2 -- "$umsg"
fi

# These values assume the default installation directory for Visual
# Studio Community; after initialization, the first part of the PATH
# should match the VSINSTALLDIR environment variable.

if [[ -z "$setvcerrors" ]]
then
    case $vsversion in
    2022)
	vsinstalldir="C:\Program Files\Microsoft Visual Studio\2022\Community"
	vcbatdir="$vsinstalldir\VC\Auxiliary\Build"
	vscommontools=VS170COMNTOOLS
	;;
    2019)
	vsinstalldir="C:\Program Files (x86)\Microsoft Visual Studio\2019\Community"
	vcbatdir="$vsinstalldir\VC\Auxiliary\Build"
	vscommontools=VS160COMNTOOLS
	;;
    2015)
	vsinstalldir="C:\Program Files (x86)\Microsoft Visual Studio 14.0"
	vcbatdir="$vsinstalldir\VC"
	vscommontools=VS140COMNTOOLS
	;;
    *)
	print -ru2 "$progname: unsupported Visual Studio version: '$vsversion'; choices are 2022/2019/2015"
	setvcerrors=YES
	;;
    esac
fi

# variables needed for Visual Studio; 'PATH' will match 'LIBPATH'
envvars="PATH=|INCLUDE=|LIB=|$vscommontools="

# batch file: this value is installation and version dependent--adjust
# as needed. It should be the shortcut on the Start Menu for 
# 'x64 Native Tools Command Prompt' or 'x64_x86 Cross Tools Command Prompt'
# for Visual Studio 2019 and 2022 or for 'x64 Native Tools Command Prompt'
# or 'x64 x86 Cross Tools Command Prompt' for Visual Studio 2015

# This assumes the compilation will be run on a 64-bit host; use the
# commented-out values for 'vcarch' to run it on a 32-bit host

if [[ -z "$setvcerrors" ]]
then
    if [[ $# -gt 0 ]]
    then
	vcarch=$1
    else
	vcarch=x64
    fi

    # valid target architectures; 64-bit host is assumed
    vcarches32='x86|x64_x86|amd64_x86|32|32bit|32-bit'
    vcarches64='x64|amd64|64|64bit|64-bit'
    
    case $vsversion in
    2019|2022)
	case $vcarch in
	x86|x64_x86|amd64_x86|32|32bit|32-bit)
	    vcbatfile="$vcbatdir\vcvars32.bat"
	    #vcarch=x86		# 32 bit target on 32 bit host
	    vcarch=x64_x86	# 32 bit target on 64 bit host
	    ;;
	x64|amd64|64|64bit|64-bit)
	    vcbatfile="$vcbatdir\vcvars64.bat"
	    #vcarch=x86_x64	# 64 bit target on 32 bit host
	    vcarch=x64		# 64 bit target on 64 bit host
	    ;;
	*)
	    print -ru2 "$progname: invalid architecture: '$vcarch'"
	    print -ru2 "$progname: 64-bit choices are $vcarches64"
	    print -ru2 "$progname: 32-bit choices are $vcarches32"
	    setvcerrors=YES
	    ;;
	esac
	;;
    2015)
	case $vcarch in
	x86|32|32bit|32-bit)
	    vcbatfile="$vcbatdir\vcvarsall.bat"
	    vcarch=amd64_x86
	    ;;
	x64|amd64|64|64bit|64-bit)
	    vcbatfile="$vcbatdir\vcvarsall.bat"
	    vcarch=amd64
	    ;;
	*)
	    print -ru2 "$progname: invalid architecture: '$vcarch'"
	    print -ru2 "$progname: 64-bit choices are $vcarches64"
	    print -ru2 "$progname: 32-bit choices are $vcarches32"
	    setvcerrors=YES
	    ;;
	esac
	;;
    esac

    if [[ -z "$setvcerrors" ]]
    then
	# don't set the variables twice, because new values are added to previous values
	if test "$VCVARSSET" != ""
	then
	    print -ru2 "$progname: Visual C $VSVERSION variables already set: $VCVARSSET bit"
	    setvcerrors=YES
	fi
	if test ! -e "$vcbatfile"
	then
	    print -ru2 "$progname: cannot find Command Prompt batch file '$vcbatfile'"
	    setvcerrors=YES
	fi
	if test -z "$ComSpec"
	then
	    print -ru2 "$progname: no path to command interpreter: ComSpec not set"
	    setvcerrors=YES
	fi
	if test ! -e "$ComSpec"
	then
	    print -ru2 "$progname: cannot find command interpreter '$ComSpec'"
	    setvcerrors=YES
	fi
	if test ! -x "$ComSpec"	# this should never happen ...
	then
	    print -ru2 "$progname: cannot run command interpreter '$ComSpec'"
	    setvcerrors=YES
	fi
	if test ! -d "$tempdir"	# this should never happen, either
	then
	    print -ru2 "$progname: cannot find TEMP dir'$tempdir'"
	    setvcerrors=YES
	fi
    fi
fi

if [ -z "$setvcerrors" ]
then
    # save this for comparison with (hopefully) modified version
    OPATH="$PATH"

    # create a batch file to initialize the environment using the Visual
    # Studio batch files and echo the results
    # As of Visual Studio 2022, this no longer works in a here document
    # with PTC MKS Toolkit 10.4

    print -r "prompt \$s
@call \"$vcbatfile\" $vcarch
@echo PATH=\"%PATH%\"; export PATH
@echo INCLUDE=\"%INCLUDE%\"; export INCLUDE
@echo LIB=\"%LIB%\"; export LIB
@echo LIBPATH=\"%LIBPATH%\"; export LIBPATH
@rem assume $vscommontools has trailing backslash
@echo $vscommontools=\"%$vscommontools%\\\"; export $vscommontools" > $tempfile

    # run the batch file and echo the results to read into the shell with
    # eval(1)
    eval "$($ComSpec /q /c $tempfile | grep -E $envvars)"
    rm -f $tempfile

    # variable initialization probably failed
    if test "$PATH" = "$OPATH"
    then
	echo "$progname: could not set Visual C variables"
    else	# prevent setting the variables more than once
	VSVERSION=$vsversion
	export VSVERSION
	case "$vcarch" in
	x64|amd64)
	    VCVARSSET=64 ;;
	*)
	    VCVARSSET=32 ;;
	esac
	export VCVARSSET

	# FIXME: show this only with vcverbose set?
	print -r "Environment initialized for Visual Studio $vsversion: '$vcarch'"
    fi
fi

# make getopts work in case a previous invocation used an invalid option
OPTIND=1

# because this is run in source mode, these will persist if not unset
unset progname umsg thisprog tempdir tempfile OPATH OPTARG
unset envvars setvcerrors vsversion vscommontools vcbatdir vcbatfile
unset vcarch vcarches32 vcarches64 vcverbose

@rem create Windows installation director hierarchy for units,
@rem   a program for units conversion
@rem
@rem   for Microsoft Windows(R) without Unix-style utilities
@rem   tested with Windows XP SP3
@rem
@rem version 1.0 12 March 2014 by Jeff Conrad (jeff_conrad@msn.com)
@rem
@rem Copyright (C) 2014
@rem Free Software Foundation, Inc
@rem
@rem This program is free software; you can redistribute it and/or modify
@rem it under the terms of the GNU General Public License as published by
@rem the Free Software Foundation; either version 3 of the License, or
@rem (at your option) any later version.
@rem
@rem This program is distributed in the hope that it will be useful,
@rem but WITHOUT ANY WARRANTY; without even the implied warranty of
@rem MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@rem GNU General Public License for more details.
@rem
@rem You should have received a copy of the GNU General Public License
@rem along with this program; if not, write to the Free Software
@rem Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
@rem    
@rem
@rem The units program was written by Adrian Mariano (adrianm@gnu.org).

@echo off

SetLocal EnableDelayedExpansion EnableExtensions
path %SystemRoot%\System32;%SystemRoot%;%SystemRoot%\System32\Wbem

@rem basename of this script
set ProgName=%~n0
set Errors=0

if "%cmdextversion%"=="" (
  echo %0: you must enable command extensions to run this batch file
  goto :eof
)

set args=%*
if not defined args (
  echo %ProgName%: missing directory
  echo Usage: %ProgName% ^<directory1^> [^<directory2^> ... ]
  goto end
)

@rem protect parens, e.g., 'Program Files (x86)'
set args=%args:(=^(%
set args=%args:)=^)%

@rem separate directory pathnames into components
for %%i in (%args%
) do (
    @rem strip quotes
    call :mkdirs %%~i
)
goto end

:mkdirs
  set installdir=%*
  if not defined installdir goto end_mkdirs
  if "%installdir:~0,1%"==":" (
    echo %ProgName%: pathname '%*': cannot begin with ':'
    goto end_mkdirs
  )
  @rem tag leading '\' so it can be preserved
  if "%installdir:~0,1%"=="\" set installdir=:#%installdir%
  @rem tag spaces so we can restore them later
  set installdir=%installdir: =:@:%
  @rem split pathname into components
  set installdir=%installdir:\= %
  @rem restore leading '\'
  set installdir=%installdir::#=\%
  call :mkpathcomp %installdir%
:end_mkdirs
goto :eof

:mkpathcomp
  set drive=%1
  set drive=%drive:~0,2%
  @rem protect silly stuff like '()'
  set drive=%drive:(=^^(%
  set drive=%drive:)=^^)%
  if "%drive:~1,1%"==":" (
    if not exist "%drive%" (
      echo %ProgName%: drive '%drive%' does not exist
      set Errors=1
      goto end_mkpathcomp
    )
  )
  set pathcomp=
  set pathargs=%*
  set pathargs=%pathargs:(=^^(%
  set pathargs=%pathargs:)=^^)%
  for %%j in (%pathargs%) do (
    set pathcomp=!pathcomp!%%j
    @rem restore spaces
    set pathcomp=!pathcomp::@:= !
    if "!pathcomp!"=="\" set pathcomp=

    if not exist !pathcomp!\ (
      mkdir "!pathcomp!"
      if errorlevel 1  (
        echo %ProgName%: cannot create directory '!pathcomp!'
        set Errors=1
        goto end_mkpathcomp
      ) else (
        echo %ProgName%: created directory '!pathcomp!'
      )
    )
    set pathcomp=!pathcomp!\
  )
:end_mkpathcomp
goto :eof

:end
if %Errors% NEQ 0 (
    echo.
    echo %ProgName%: one or more directories could not be created
)
@rem needed to handle invocation with command extension disabled
:eof
EndLocal

#! /bin/sh
# ******************************************************************************
#  elemcvt: convert NIST "Linear ASCII" table of elements to units(1) format
#  Usage: elemcvt [options] [<file>]
#  Author: Jeff Conrad
#  Date:  2024-01-06
# ******************************************************************************

# Adjust PATH to suit.
# For Windows w/MKS Toolkit, this assumes /bin is a symbolic link to
# $ROOTDIR/mksnt.

PATH=/bin

progname=${0##*[/\\]}
progname=${progname%.sh}
export TITLEBAR=$progname

umsg="Usage: $progname [options] [file]
Options:
  -d  Show elements for which no standard atomic mass is given and exit
  -v  Verbose"

show_no_std_atomic_mass=	# show elements for which no std atomic mass is given
verbose=
errors=
DUALCASE=1	# used in MKS Toolkit to make options case sensitive

while getopts :dv arg
do
    case $arg in
    d)
	show_no_std_atomic_mass=YES ;;
    v)
	verbose=YES ;;
    :)
	# OPTARG contains the option missing the argument
	print -ru2 -- "$progname: option $OPTARG requires an argument"
	errors=YES
	;;
    [?])
	# OPTARG contains the invalid option
	print -ru2 -- "$progname: unknown option $OPTARG"
	errors=YES
	;;
    esac
done
shift $((OPTIND - 1))
unset DUALCASE

if [ -n "$errors" ]
then
    print -ru2 -- "$umsg"
    exit 1
fi

awk '

function show_element_info(atomic_number, atomic_symbol, std_atomic_mass_str)
{
    printf("# %s: %s (%d)", names[atomic_number], atomic_symbol, atomic_number)
    if (std_atomic_mass_str)
	printf("  std atomic weight: %s", std_atomic_mass_str)
    print ""
}

# <name>_<atomic num>  <mass> # <mole fraction>
function show_isotope(name, num)
{
    printf("%-*s%*s%*.*f", max_isotope_len, sprintf("%s_%d", name, num),
	    sepwid, " ", isoprecis + 4, isoprecis, mass[num])
    if (composition[num])
	printf("   # %.*f", compprecis, composition[num])
    print ""
}

function show_element_name(name)
{
    printf("%-*s%*s", max_name_len, name, sepwid, " ")
}

# <mole fraction> <name>_<mass num>
function mole_fraction(atomic_number, names, mass_num,  mass_wid)
{
    if (composition[mass_num] == 1) {
	mass_wid = length(int(mass_number[n_isotopes]))
	# align with 1st digit of atomic mass
	printf("%*s%s_%d", sepwid + 2 - mass_wid, " ", names[atomic_number], mass_num)
    }
    else
	printf("%.*f %s_%d", compprecis, composition[mass_num],
		names[atomic_number], mass_num)
}

# add line continuation and '+' sign
function add_continuation()
{
    printf(" \\\n")
    printf("%-*s+ ", max_name_len + sepwid - 2, " ")
}

# <name>_<mass num> # most stable
function use_most_stable(atomic_number, mass,  mass_wid)
{
    mass_wid = length(int(mass_number[n_isotopes]))
    printf("%*s%-*s   # most stable", sepwid + 2 - mass_wid, " ",
	    isoprecis + 1 + mass_wid, sprintf("%s_%d", names[atomic_number], mass))
}

# <name>_<mass num> # standard atomic mass
function use_std_mass(atomic_number, mass,  mass_wid)
{
    mass_wid = length(int(mass_number[n_isotopes]))
    printf("%*s%-*s   # standard atomic mass", sepwid + 2 - mass_wid, " ",
	    isoprecis + 1 + mass_wid, sprintf("%s_%d", names[atomic_number], mass))
}

# show isotopes and the sum of mole fraction-mass products
function show_element_data(atomic_number, names)
{
    # isotopes and relative abundances
    for (ndx = 1; ndx <= n_isotopes; ndx++) {
	mass_num = mass_number[ndx]
	show_isotope(names[atomic_number], mass_num)
    }

    # show a value for atomic mass if one of these is available;
    # otherwise, show only isotope masses.
    if (total_composition > 0 || most_stable_mass || std_atomic_mass)
	show_element = 1
    else
	show_element = 0

    # atomic mass: sum of mole fraction-mass products
    # element name
    if (show_element)
	show_element_name(names[atomic_number])

    mass_num = mass_number[1]
    firstval = 0

    # first isotope
    if (composition[mass_num] > 0) {
	mole_fraction(atomic_number, names, mass_num)
	firstval = 1
    }
    if (n_isotopes > 1) {
	for (ndx = 2; ndx < n_isotopes; ndx++) {
	    mass_num = mass_number[ndx]
	    if (composition[mass_num] > 0) {
		if (firstval == 1)
		    add_continuation()
		mole_fraction(atomic_number, names, mass_num)
		firstval = 1
	    }
	}
	# last isotope
	mass_num = mass_number[n_isotopes]
	if (composition[mass_num] > 0) {
	    if (firstval == 1)
		add_continuation()
	    mole_fraction(atomic_number, names, mass_num)
	    print ""
	}
    }
    else
	print ""

    # options if mole fraction is not given for any isotope
    if (total_composition == 0) {
	if (most_stable_mass)
	    use_most_stable(atomic_number, most_stable_mass)
	else if (std_atomic_mass)
	    use_std_mass(atomic_number, std_atomic_mass)
    }

    if (! composition[mass_number[n_isotopes]])
	print ""
}

function output(atomic_number)
{
    sepwid = 5		# width of column separation
    compprecis = 8	# for mole fraction
    isoprecis = 10	# for isotope mass

    # NIST show H, D, and T
    if (atomic_number == 1)
	atomic_symbol = "H"

    show_element_info(atomic_number, atomic_symbol, std_atomic_mass_str)

    if (am_names[atomic_number])
	print "# IUPAC spelling"

    show_element_data(atomic_number, names)

    # show American spelling if different from IUPAC
    if (am_names[atomic_number]) {
	print "# American spelling"
	show_element_data(atomic_number, am_names)
    }
    if (show_element)
	print ""	# blank line between elements

    most_stable_mass = 0
    total_composition = 0
}

function gnu_notes()
{
    print "\
# This file is the elements database for use with GNU units, a units\n\
# conversion program by Adrian Mariano adrianm@gnu.org\n\
#\n\
# January 2024 Version 1.0\n\
#\n\
# Copyright (C) 2024\n\
# Free Software Foundation, Inc\n\
#\n\
# This program is free software; you can redistribute it and/or modify\n\
# it under the terms of the GNU General Public License as published by\n\
# the Free Software Foundation; either version 3 of the License, or\n\
# (at your option) any later version.\n\
#\n\
# This data is distributed in the hope that it will be useful,\n\
# but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
# GNU General Public License for more details.\n\
#\n\
# You should have received a copy of the GNU General Public License\n\
# along with this program; if not, write to the Free Software\n\
# Foundation, Inc., 51 Franklin Street, Fifth Floor,\n\
# Boston, MA  02110-1301  USA\n"
}

function nist_notes()
{
    print "# From https://www.nist.gov/pml/atomic-weights-and-isotopic-compositions-relative-atomic-masses\n"

    # notes from https://www.nist.gov/pml/atomic-weights-and-isotopic-compositions-column-descriptions
    print "\
#   For several elements, the standard atomic weight A_r is given as an\n\
#   atomic-weight interval with the symbol [a,b] to denote the set of\n\
#   atomic-weight values in normal materials; thus, [a <= A_r(E) <= b].\n\
#   The symbols a and b denote the lower and upper bounds of the\n\
#   interval [a,b], respectively.  The values in parentheses, following\n\
#   the last significant digit to which they are attributed, are\n\
#   uncertainties.\n\
#\n\
#   Brackets [ ] enclosing a single value indicate the mass number of\n\
#   the most stable isotope.  For radioactive elements with atomic\n\
#   numbers 95 or greater, the mass number of the most stable isotope is\n\
#   not specified, as the list of studied isotopes is still\n\
#   incomplete.\n"
}

function units_notes()
{
    print "\
# When composition mole fractions of isotopes are given, the atomic mass\n\
# of an element is given as the sum of the product(s) of mole\n\
# fraction(s) and the atomic masses of the relevant isotopes.  When composition\n\
# mole fractions are not given, the atomic mass is given as\n\
#\n\
#   * the mass of the most stable isotope, if available, or\n\
#   * the standard atomic mass of the element, if available.\n\
#\n\
# If neither the most stable isotope nore a standard atomic mass is\n\
# available, no atomic mass for the element is given; the user must\n\
# select the isotope most suitable for their purposes.\n\
#\n\
# If the standard atomic mass is a range, the value given is the\n\
# midpoint of that range, which may differ from the value determined\n\
# from the sum of the products of composition mole fraction and isotope\n\
# atomic mass.\n"
}

BEGIN {
    FS = " *= *"
    show_no_std_atomic_mass = "'"$show_no_std_atomic_mass"'"
    verbose = "'"$verbose"'"
    console = "/dev/console"

    # IUPAC spellings
    names[1] = "hydrogen"
    names[2] = "helium"
    names[3] = "lithium"
    names[4] = "beryllium"
    names[5] = "boron"
    names[6] = "carbon"
    names[7] = "nitrogen"
    names[8] = "oxygen"
    names[9] = "fluorine"
    names[10] = "neon"
    names[11] = "sodium"
    names[12] = "magnesium"
    names[13] = "aluminium"
    names[14] = "silicon"
    names[15] = "phosphorus"
    names[16] = "sulfur"
    names[17] = "chlorine"
    names[18] = "argon"
    names[19] = "potassium"
    names[20] = "calcium"
    names[21] = "scandium"
    names[22] = "titanium"
    names[23] = "vanadium"
    names[24] = "chromium"
    names[25] = "manganese"
    names[26] = "iron"
    names[27] = "cobalt"
    names[28] = "nickel"
    names[29] = "copper"
    names[30] = "zinc"
    names[31] = "gallium"
    names[32] = "germanium"
    names[33] = "arsenic"
    names[34] = "selenium"
    names[35] = "bromine"
    names[36] = "krypton"
    names[37] = "rubidium"
    names[38] = "strontium"
    names[39] = "yttrium"
    names[40] = "zirconium"
    names[41] = "niobium"
    names[42] = "molybdenum"
    names[43] = "technetium"
    names[44] = "ruthenium"
    names[45] = "rhodium"
    names[46] = "palladium"
    names[47] = "silver"
    names[48] = "cadmium"
    names[49] = "indium"
    names[50] = "tin"
    names[51] = "antimony"
    names[52] = "tellurium"
    names[53] = "iodine"
    names[54] = "xenon"
    names[55] = "caesium"
    names[56] = "barium"
    names[57] = "lanthanum"
    names[58] = "cerium"
    names[59] = "praseodymium"
    names[60] = "neodymium"
    names[61] = "promethium"
    names[62] = "samarium"
    names[63] = "europium"
    names[64] = "gadolinium"
    names[65] = "terbium"
    names[66] = "dysprosium"
    names[67] = "holmium"
    names[68] = "erbium"
    names[69] = "thulium"
    names[70] = "ytterbium"
    names[71] = "lutetium"
    names[72] = "hafnium"
    names[73] = "tantalum"
    names[74] = "tungsten"
    names[75] = "rhenium"
    names[76] = "osmium"
    names[77] = "iridium"
    names[78] = "platinum"
    names[79] = "gold"
    names[80] = "mercury"
    names[81] = "thallium"
    names[82] = "lead"
    names[83] = "bismuth"
    names[84] = "polonium"
    names[85] = "astatine"
    names[86] = "radon"
    names[87] = "francium"
    names[88] = "radium"
    names[89] = "actinium"
    names[90] = "thorium"
    names[91] = "protactinium"
    names[92] = "uranium"
    names[93] = "neptunium"
    names[94] = "plutonium"
    names[95] = "americium"
    names[96] = "curium"
    names[97] = "berkelium"
    names[98] = "californium"
    names[99] = "einsteinium"
    names[100] = "fermium"
    names[101] = "mendelevium"
    names[102] = "nobelium"
    names[103] = "lawrencium"
    names[104] = "rutherfordium"
    names[105] = "dubnium"
    names[106] = "seaborgium"
    names[107] = "bohrium"
    names[108] = "hassium"
    names[109] = "meitnerium"
    names[110] = "darmstadtium"
    names[111] = "roentgenium"
    names[112] = "copernicium"
    names[113] = "nihonium"
    names[114] = "flerovium"
    names[115] = "moscovium"
    names[116] = "livermorium"
    names[117] = "tennessine"
    names[118] = "oganesson"

    # American spellings
    am_names[13] = "aluminum"
    am_names[55] = "cesium"

    max_name_len = 0		# length of longest element name
    for (i = 1; i <= 118; i++) {
	len = length(names[i])
	if (len > max_name_len) {
	    max_name_len = len;
	    longestname = names[i]
	}
    }
    max_isotope_len = max_name_len + 4	# allow for "_xxx" suffix

    if (! show_no_std_atomic_mass) {
	gnu_notes()
	nist_notes()
	units_notes()
    }

    if (verbose)
	printf("Longest element name: %s (%d)\n\n", longestname, max_name_len)

    n_isotopes = 0
    mass_number[1] = 0
}

# begin file processing

# skip JavaScript and HTML before data
NR == 1, $0 ~ /<pre/ { next }
# skip HTML after data
$0 ~ /<\/pre>/ { exit }

# remove trailing space and unpaddable spaces
{ 
  gsub(/&nbsp;/, "")
  gsub(/ +$/, "")
}

$1 ~ /Atomic Number/ {
    last_atomic_number = atomic_number
    atomic_number = $2 + 0
    if (atomic_number != last_atomic_number && atomic_number > 1) {
	if (show_no_std_atomic_mass) {
	    if (! std_atomic_mass_str)
		print names[last_atomic_number]
	}
	else
	    output(last_atomic_number)
    }
}

$1 ~ /Atomic Symbol/ {
    atomic_symbol = $2
}

$1 ~ /Mass Number/ {
    if (atomic_number != last_atomic_number) {
	for (i = 1; i <= n_isotopes; i++)
	    delete mass_number[i]
	n_isotopes = 0
    }
    mass_number[++n_isotopes] = $2
}

$1 ~ /Relative Atomic Mass/ {
    atomic_mass = $2
    sub(/\([[:digit:]#]+\)/, "", atomic_mass)
    mass[mass_number[n_isotopes]] = atomic_mass
}

$1 ~ /Isotopic Composition/ {
    isotopic_composition = $2
    sub(/\([[:digit:]#]+\)/, "", isotopic_composition)
    composition[mass_number[n_isotopes]] = isotopic_composition
    total_composition += isotopic_composition
}

$1 ~ /Standard Atomic Weight/ {
    std_atomic_mass = std_atomic_mass_str = $2
    gsub(/\([^)]+\)/, "", std_atomic_mass)
    gsub(/[][]/, "", std_atomic_mass)
    if (std_atomic_mass ~ /,/) {
	split(std_atomic_mass, range, /,/)
	std_atomic_mass = (range[1] + range[2]) / 2
    }
    if (std_atomic_mass_str ~ /\[[[:digit:].]+\]/)
	most_stable_mass = std_atomic_mass
    last_atomic_number = atomic_number
}

END {
    if (show_no_std_atomic_mass) {
	if (! std_atomic_mass_str)
	    print names[last_atomic_number]
    }
    else
	output(last_atomic_number)
} ' $*

---
title: "Measurement units in R"
author: "Edzer Pebesma, Thomas Mailund, and James Hiebert"
output:
  rmarkdown::html_vignette:
    toc: yes
bibliography: measurement_units_in_R.bib
link-citations: yes
vignette: >
  %\VignetteIndexEntry{Measurement units in R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r echo=FALSE}
knitr::opts_chunk$set(collapse = TRUE, fig.asp = 0.7, fig.width = 7)
```

This vignette is identical to @rj, except for two changes:

* it has been synchronized with updates to the [units](https://cran.r-project.org/package=units) package
* it has been converted to R-markdown

## Abstract
We briefly review SI units, and discuss R packages that deal
with measurement units, their compatibility and conversion.
Built upon the UNIDATA udunits library, we introduce the package
[units](https://cran.r-project.org/package=units) that provides a
class for maintaining unit metadata. When used in expression, it
automatically converts units, and simplifies units of results when
possible; in case of incompatible units, errors are raised. The
class flexibly allows expansion beyond predefined units.  Using
[units](https://cran.r-project.org/package=units) may eliminate a
whole class of potential scientific programming mistakes.  We discuss
the potential and limitations of computing with explicit units.


## Introduction

Two quotes from @cobb -- _Data are not just numbers,
they are numbers with a context_ and  _in data analysis,
context provides meaning_ -- illustrate that for a data analysis
to be meaningful, knowledge of the data's context is needed.
Pragmatic aspects of this context include who collected or generated
the data, how this was done, and for which purpose [@scheider];
semantic aspects concern what the data represents: which aspect
of the world do the data refer to, when and where were they measured,
and what a value of `1` means.

R does allow for keeping some context with data, for instance

* `data.frame` columns must have and `list` elements may have names that can be used to describe context, using freetext
* `matrix` or `array` objects may have `dimnames`
* for variables of class `factor` or `ordered`, `levels` may indicate, using freetext, the categories of nominal or ordinal variables
* `POSIXt` and `Date` objects specify how numbers should be interpreted as time or date, with fixed units (second and day, respectively) and origin (Jan 1, 1970, 00:00 UTC)
* `difftime` objects specify how time duration can be represented by numbers, with flexible units (secs, mins, hours, days, weeks); [lubridate](https://cran.r-project.org/package=lubridate) [@lubridate] extends some of this functionality.

Furthermore, if spatial objects as defined in package
[sp](https://cran.r-project.org/package=sp) [@sp] have a proper
coordinate reference system set, they can be transformed to other
datums, or converted to various flat (projected) representations
of the Earth [@iliffe].

In many cases however, R drops contextual information. As an example, we
look at annual global land-ocean temperature index (from
`http://climate.nasa.gov/vital-signs/global-temperature/`) since 1960:
```{r}
temp_data = subset(read.table("647_Global_Temperature_Data_File.txt", 
	header=TRUE)[1:2], Year >= 1960)
temp_data$date = as.Date(paste0(temp_data$Year, "-01-01"))
temp_data$time = as.POSIXct(temp_data$date)
Sys.setenv(TZ="UTC")
head(temp_data, 3)
year_duration = diff(temp_data$date)
mean(year_duration)
```

Here, the time difference units are reported for the `difftime`
object `year_duration`, but if we would use it in a linear algebra operation
```{r}
year_duration %*% rep(1, length(year_duration)) / length(year_duration)
```
the unit is dropped. Similarly, for linear regression coefficients we see
```{r}
coef(lm(Annual_Mean ~ date, temp_data))
coef(lm(Annual_Mean ~ time, temp_data))
```
where the unit of change is in degrees Celsius but either _per
day_ (`date`) or _per second_ (`time`). For purely mathematical
manipulations, R often strips context from numbers when it is carried
in attributes, the linear algebra routines being a prime example.

Most variables are somehow attributed with information about
their _units_, which specify what the value `1` of this variable
represents.  This may be counts of something, e.g. `1 apple`,
but it may also refer to some _physical unit_, such as distance
in meter. This article discusses how strong unit support can be
introduced in R.

## SI

The [BIPM](https://www.bipm.org/) (Bureau International des Poids et
Mesures) is "_the intergovernmental organization through which
Member States act together on matters related to measurement science
and measurement standards.  Its recommended practical system of units
of measurement is the International System of Units (Système International 
d'Unités, with the international abbreviation SI)_
(https://www.bipm.org/en/measurement-units/)". 

@si describe the SI units, where, briefly, _SI units_

* consist of seven base units (length, mass, time \& duration, electric current, thermodynamic temperature, amount of substance, and luminous intensity), each with a name and abbreviation (see table below)
* consist of _derived units_ that are formed by products of powers of base units, such as m/s$^2$, many of which have special names and symbols (e.g. angle: 1 rad = 1 m/m; force: 1 N = 1 m kg s$^{-2}$) 
* consist of _coherent derived units_ when derived units include no numerical factors other than one (with the exception of `kg`; as a base unit, kg can be part of coherent derived units); an example of a coherent derived unit is 1 watt = 1 joule per 1 second, 
* may contain SI prefixes (k = kilo for $10^3$, m = milli for $10^{-3}$, etc.) 
* contain special quantities where units disappear (e.g., m/m) or have the nature of a count, in which cases the unit is 1.

base quantities, SI units and their symbols (from @si, p. 23):

| Base quantity |    | SI base unit |   |
| -----------|--------|----------|------|
| Name          |Symbol|Name | Symbol
| length        | $l,x,r,$ etc.| meter | m |
| mass          | $m$ | kilogram | kg |
| time, duration | $t$ | second | s |
| electric current| $I, i$ | ampere | A |
| thermodynamic temperature | $T$ | kelvin | K |
| amount of substance | $n$ | mole | mol |
| luminous intensity |  $I_v$ | candela | cd |

## Related work in R

Several R packages provide unit conversions. 
For instance, [measurements](https://cran.r-project.org/package=measurements) [@measurements] provides
a collection of tools to make working with physical measurements
easier. It converts between metric and imperial units, or calculates
a dimension's unknown value from other dimensions' measurements. It
does this by the `conv_unit` function:
```{r, eval=requireNamespace("measurements", quietly=TRUE)}
library(measurements)
conv_unit(2.54, "cm", "inch")
conv_unit(c("101 44.32","3 19.453"), "deg_dec_min", "deg_min_sec")
conv_unit(10, "cm_per_sec", "km_per_day")
```
but uses for instance `kph` instead of `km_per_hour`, and then
`m3_per_hr` for flow -- unit names seem to come from convention rather
than systematic composition.  Object
`conv_unit_options` contains all 173 supported units, categorized by
the physical dimension they describe:
```{r, eval=requireNamespace("measurements", quietly=TRUE)}
names(conv_unit_options)
conv_unit_options$volume
```
Function `conv_dim` allows for the conversion of units in products or ratios,
e.g.
```{r, eval=requireNamespace("measurements", quietly=TRUE)}
conv_dim(x = 100, x_unit = "m", trans = 3, trans_unit = "ft_per_sec", y_unit = "min")
```
computes how many minutes it takes to travel 100 meters at 3 feet per second.

Package [NISTunits](https://cran.r-project.org/package=NISTunits) [@NISTunits] provides fundamental
physical constants (Quantity, Value, Uncertainty, Unit) for SI and
non-SI units, plus unit conversions, based on the data from NIST
(National Institute of Standards and Technology). The package
provides a single function for every
unit conversion; all but 5 from its 896 functions are of the form
`NISTxxxTOyyy` where `xxx` and `yyy` refer to two
different units. For instance, converting from W m$^{-2}$ to W inch$^{-2}$
is done by
```{r, eval=requireNamespace("NISTunits", quietly=TRUE)}
library(NISTunits)
NISTwattPerSqrMeterTOwattPerSqrInch(1:5)
```
Both [measurements](https://cran.r-project.org/package=measurements) and [NISTunits](https://cran.r-project.org/package=NISTunits) are written entirely
in R.

## UNIDATA's udunits library

Udunits, developed by UCAR/UNIDATA, advertises itself on [its web
page](https://www.unidata.ucar.edu/software/udunits)
as: "_The udunits package supports units of physical
quantities. Its C library provides for arithmetic manipulation
of units and for conversion of numeric values between compatible
units. The package contains an extensive unit database, which is
in XML format and user-extendable._"

Unlike the
[measurements](https://cran.r-project.org/package=measurements)
and [NISTunits](https://cran.r-project.org/package=NISTunits),
the underlying udunits2 C library parses
units as expressions, and bases its logic upon the convertibility
of expressions, rather than the comparison of fixed strings:
```{r}
m100_a = paste(rep("m", 100), collapse = "*")
dm100_b = "dm^100"
units::ud_are_convertible(m100_a, dm100_b)
```
This has the advantage that through complex computations,
intermediate objects can have units that are arbitrarily complex,
and that can potentially be simplified later on. It also means that
the package practically supports an unlimited amount of derived units.

## Udunits versus the Unified Code for Units of Measure (UCUM)

Another set of encodings for measurement units is the Unified Code for
Units of Measure (UCUM, @ucum). A dedicated web
site\footnote{\url{http://coastwatch.pfeg.noaa.gov/erddap/convert/units.html}}
describes the details of the differences between udunits and UCUM,
and provides a conversion service between the two encoding sets. 

The UCUM website refers to some Java implementations, but some of the
links seem to be dead.  UCUM is the preferred encoding for standards
from the Open Geospatial Consortium. udunits on the other hand
is the units standard of choice by the climate science community,
and is adopted by the CF (Climate and Forecast) conventions, which
mostly uses NetCDF. NetCDF [@netcdf] is a binary data format
that is widely used for atmospheric and climate model predictions.

The udunits library is a C library that has strong support from
UNIDATA, and we decided to build our developments on this, rather
than on Java implementations of UCUM with a less clear provenance.

## Handling data with units in R: the units package

The [units](https://cran.r-project.org/package=units) package builds `units` objects from scratch,
e.g. where
```{r}
library(units)
x = set_units(1:5, m/s)
str(x)
```
represents speed values in `m/s`. The units `m` and `s` are resolved
from the udunits2 C library (but could be user-defined units).

Units can be used in arbitrary R expressions like
```{r}
set_units(1:3, m/s^2)
```

Several manipulations with `units` objects will now be illustrated.
Manipulations that do not involve unit conversion are for instance addition:
```{r}
x = set_units(1:3, m/s)
x + 2 * x
```
Explicit unit conversion is done by assigning new units:
```{r}
(x = set_units(x, cm/s))
as.numeric(x)
```
similar to the behaviour of `difftime` objects, this modifies
the numeric values without modifying their meaning (what the numbers
refer to).

When mixing units in sums, comparisons or concatenation, units are 
automatically converted to those of the first argument:
```{r}
y = set_units(1:3, km/h)
x + y
y + x
x == y
c(y, x)
```
where `c(y, x)` concatenates `y` and `x` after converting `x` to the
units of `y`. Derived units are created where appropriate:
```{r}
x * y
x^3
```
and meaningful error messages appear when units are not compatible:
```{r}
e = try(z <- x + x * y)
attr(e, "condition")[[1]]
```
The full set of methods and method groups for `units` objects is shown by
```{r}
methods(class = "units")
```
where the method groups

* `Ops` include operations that require compatible units, converting when necessary (`+`, `-`, `==`, `!=`, `<`, `>`, `<=`, `>=`), and operations that create new units (`*`, `/`, `^` and `**`),
* `Math` include `abs`, `sign`, `floor`, `ceiling`, `trunc`, `round`, `signif`, `log`, `cumsum`, `cummax`, `cummin`, and
* `Summary` include `sum`, `min`, `max` and `range`, and all convert to the unit of the first argument.

When possible, new units are simplified:
```{r}
a = set_units(1:10, m/s)
b = set_units(1:10, h)
a * b
ustr1 = paste(rep("m", 101), collapse = "*")
ustr2 = "dm^100"
as_units(ustr1) / as_units(ustr2)
```

Units are printed as simple R expressions, e.g.
```{r}
set_units(1, m^5/s^4)
```
Another way to print units commonly seen in Climate and Forecast Conventions\footnote{CF, \url{http://cfconventions.org/Data/cf-standard-names/34/build/cf-standard-name-table.html}} is `m2 s-1` for m$^2$/s. These are not R expressions, but they can be parsed by `as_units`, and created by `deparse_unit`:
```{r}
as_units("m2 s-1")
deparse_unit(set_units(1, m^2*s^-1))
```
The `plot` and `hist` methods add units to default axis labels, an
example is shown in the following figures. For [ggplot2](https://cran.r-project.org/package=ggplot2) plots
[@ggplot2], automatic unit placement in default axis label is also provided; `demo(ggplot2)` gives an example.

```{r}
library(units)
units_options(negative_power = TRUE)
# initialize variables with units:
mtcars$consumption = set_units(mtcars$mpg, mi/gallon)
# "in" is also a reserved R keyword, and so needs back-quotes:
mtcars$displacement = set_units(mtcars$disp, `in`^3)
# convert to SI:
mtcars$consumption = set_units(mtcars$consumption, km/l)
mtcars$displacement = set_units(mtcars$displacement, cm^3)
par(mar = par("mar") + c(0, .3, 0, 0))
with(mtcars, plot(1/displacement, 1/consumption))
```

```{r, eval=requireNamespace("ggplot2", quietly=TRUE)}
library(ggplot2)
ggplot(mtcars) + geom_point(aes(x = 1/displacement, y = 1/consumption))
```

Automatic conversion between `units` and `difftime` is provided: 
```{r}
(dt = diff(Sys.time() + c(0, 1, 1+60, 1+60+3600))) # class difftime
(dt.u = as_units(dt))
identical(as_difftime(dt.u), dt)
```
as well as to and from `POSIXct` or `Date`:
```{r}
(t1 <- as_units(as.POSIXct("2017-08-20 17:03:00")))
(t2 <- as_units(as.POSIXct("2017-08-20 17:03:00"), "hours since 2017-08-20"))
(d1 <- as_units(as.Date("2017-08-20")))
as.POSIXct(t1)
as.Date(d1)
```
Objects of class `units` can be used as columns in `data.frame`
objects, as well as in `tbl_df` [@tibble]. They can also be `matrix`
or `array`, with the constraint that a single unit holds for all
elements.

## Discussion and conclusions

The [units](https://cran.r-project.org/package=units) R package provides a new class, `units`,
for numeric data with associated measurement units. Operations
on objects of this class retain the unit metadata and provide
automated dimensional analysis: dimensions are taken into consideration
in computations and comparisons. Combining different units that are
compatible triggers automatic unit conversion, derived units
are automatically generated and simplified where possible, and
meaningful error messages are given when a user tries to add objects
with incompatible units. This verifies that computations are not
only syntactically and numerically allowed, but also semantically,
and in the case of physical units, physically allowed, which may
support code verification and provenance tracking.  Using this
package may eliminate a whole class of potential scientific
programming mistakes.

Where the R packages
[measurements](https://cran.r-project.org/package=measurements)
and [NISTunits](https://cran.r-project.org/package=NISTunits)
provide conversion between a fixed number of units, with
the help of the udunits2 C library and unit database, R package
[units](https://cran.r-project.org/package=units) handles
arbitrarily complex derived units. By treating units as expressions
it can derive, convert and simplify units. In addition, beyond the
SI units packaged, [units](https://cran.r-project.org/package=units)
handles user-defined units.

Data in `units` vectors can be stored as columns in
`data.frame` or `tbl_df` objects, and can be
converted to and from `difftime`.  When `units`
objects have associated time and location information,
they could be stored in spatial or spatio-temporal objects
provided by [sp](https://cran.r-project.org/package=sp) or
[spacetime](https://cran.r-project.org/package=spacetime)
[@spacetime] as these store attribute data in
`data.frame` slots, but for instance not in `zoo`
[@zoo] or `xts` [@xts] objects, as these latter
two set the class attribute of a vector or matrix.

Despite all standardization efforts, units may still be ambiguous,
or subject to interpretation. For instance for the duration of one
year [NISTunits](https://cran.r-project.org/package=NISTunits) gives
us an answer that depends on whether we want a common, leap,
Gregorian, Julian, tropical or siderial year (@lang, see
also `demo(year)`).  This illustrates that those who apply
unit conversion should be aware of possible pitfalls. Support for
calendars in udunits seems not as well developed as in R.

Future work includes extending packages that read external
data from formats, databases or interfaces with support for
measurement unit information into R, preserving the measurement
unit information. Examples would be interfaces to HDF5 (e.g.,
[h5](https://cran.r-project.org/package=h5), @h5),
[RNetCDF](https://cran.r-project.org/package=RNetCDF) [@RNetCDF]
or [sos4R](https://cran.r-project.org/package=sos4R) [@sos4R].
It would be nice to see units of measurements propagate into units
of regression coefficient estimates.

## Acknowledgements

We acknowledge three anonymous reviewers and the handling editor for
their constructive comments, and Thomas Lin Pedersen for implementing
the ggplot extensions in package `ggforce` (ported to the `units` package
since v0.8-0) that automatically add units to default ggplot axis labels.

## References

---
title: "Units of Measurement for R Vectors: an Introduction"
output:
  rmarkdown::html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{Units of Measurement for R Vectors: an Introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r echo=FALSE}
knitr::opts_chunk$set(collapse = TRUE, fig.asp = 0.7, fig.width = 7)
```

```{r echo=FALSE}
units:::units_options(negative_power = FALSE)
```

R has little support for physical measurement units.  The exception
is formed by time differences: time differences objects of class
`difftime` have a `units` attribute that can be modified:

```{r}
t1 = Sys.time() 
t2 = t1 + 3600 
d = t2 - t1
class(d)
units(d)
d
units(d) = "secs"
d
```
We see here that the `units` method is used to retrieve and modify the
unit of time differences. 

The `units` package generalizes this idea to other physical units, building upon the
[udunits2](https://www.unidata.ucar.edu/software/udunits) C library.
The `udunits2` library provides the following operations:

* validating whether an expression, such as `m/s` is a valid physical unit
* verifying whether two units such as `m/s` and `km/h` are convertible
* converting values between two convertible units 
* providing names and symbols for specific units
* handle different character encodings (utf8, ascii, iso-8859-1 and latin1)

The `units` R package uses the
[udunits2](https://www.unidata.ucar.edu/software/udunits) C library to extend
R with functionality for manipulating numeric vectors that have
physical measurement units associated with them, in a similar way as
`difftime` objects behave.

## Setting units, unit conversion

We can set units to numerical values by `set_units`:
```{r}
library(units)
(a <- set_units(runif(10),  m/s))
```
the result, e.g. 
```{r}
set_units(10, m/s)
```
literally means "10 times 1 m divided by 1 s". In writing, the "1"
values are omitted, and the multiplication is implicit.

### Unit conversion

When conversion is meaningful, such as hours to seconds or meters to kilometers, conversion can be done explicitly by setting the units of a vector
```{r}
b = a
units(b) <- make_units(km/h)
b
```

## Basic manipulations

### Arithmetic operations

Arithmetic operations verify units, and create new ones
```{r}
a + a
a * a
a ^ 2
a ** -2
```
and convert to the units of the first argument if necessary:
```{r}
a + b # m/s + km/h -> m/s
```

Currently, powers are only supported for integer powers, so using `a ** 2.5` would result in an error.


### Unit simplification
There are some basic simplification of units:
```{r}
t <- make_units(s)
a * t
```
which also work when units need to be converted before they can be simplified:
```{r}
t <- make_units(min)
a * t
```
Simplification to unit-less values gives the "1" as unit:
```{r}
m <- make_units(m)
a * t / m
```

Allowed operations that require convertible units are `+`, `-`, `==`,
`!=`, `<`, `>`, `<=`, `>=`.  Operations that lead to new units are
`*`, `/`, and the power operations `**` and `^`.


### Mathematical functions

Mathematical operations allowed are: `abs`, `sign`, `floor`,
`ceiling`, `trunc`, `round`, `signif`, `log`, `cumsum`, `cummax`, `cummin`.
```{r}
signif(a ** 2 / 3, 3)
cumsum(a)
log(a) # base defaults to exp(1)
log(a, base = 10)
log(a, base = 2)
```

### Summary functions
Summary functions `sum`, `min`, `max`, and `range` are allowed:
```{r}
sum(a)
min(a)
max(a)
range(a)
```

### Printing
Following `difftime`, printing behaves differently for length-one vectors:
```{r}
a
a[1]
```

### Subsetting
The usual subsetting rules work:
```{r}
a[2:5]
a[-(1:9)]
```

### Concatenation
```{r}
c(a,a)
```
concatenation converts to the units of the first argument, if necessary:
```{r}
c(a,b) # m/s, km/h -> m/s
c(b,a) # km/h, m/s -> km/h
```

## Conversion to/from `difftime`
From `difftime` to `units`:
```{r}
t1 = Sys.time() 
t2 = t1 + 3600 
d = t2 - t1
(du = as_units(d))
```
vice versa:
```{r}
(dt = as_difftime(du))
class(dt)
```

## units in `matrix` objects
```{r}
set_units(matrix(1:4,2,2), m/s)
set_units(matrix(1:4,2,2), m/s * m/s)
```
but
```{r}
set_units(matrix(1:4,2,2), m/s) %*% set_units(4:3, m/s)
```
strips units.

## units objects in `data.frame`s
units in `data.frame` objects are printed, but do not appear in `summary`:.
```{r}
set.seed(131)
d <- data.frame(x = runif(4), 
                    y = set_units(runif(4), s), 
                    z = set_units(1:4, m/s))
d
summary(d)
d$yz = with(d, y * z)
d
d[1, "yz"]
```
## Formatting

Units are often written in the form `m2 s-1`, for square meter per second. This
can be defined as unit, and also parsed by `as_units`:
```{r}
(x = 1:10 * as_units("m2 s-1"))
```
udunits understands such string, and can convert them
```{r}
y = 1:10 * make_units(m^2/s)
x + y
```
Printing units in this form is done by
```{r}
deparse_unit(x)
```

## Plotting

Base scatter plots and histograms support automatic unit placement
in axis labels. In the following example we first convert to
SI units. (Unit `in` needs a bit special treatment, because `in` is a
reserved word in R.)

```{r}
mar = par("mar") + c(0, .3, 0, 0)
displacement = mtcars$disp * as_units("in")^3
units(displacement) = make_units(cm^3)
weight = mtcars$wt * 1000 * make_units(lb)
units(weight) = make_units(kg)
par(mar = mar)
plot(weight, displacement)
```

We can change grouping symbols from `[ ]` into `( )`:
```{r}
units_options(group = c("(", ")") )  # parenthesis instead of square brackets
par(mar = mar)
plot(weight, displacement)
```

We can also remove grouping symbols, increase space between variable name and unit by:
```{r}
units_options(sep = c("~~~", "~"), group = c("", ""))  # no brackets; extra space
par(mar = mar)
plot(weight, displacement)
```

More complex units can be plotted either with negative powers, or as divisions,
by modifying one of `units`'s global options using `units_options`:

```{r}
gallon = as_units("gallon")
consumption = mtcars$mpg * make_units(mi/gallon)
units(consumption) = make_units(km/l)
par(mar = mar)
plot(displacement, consumption) # division in consumption
units_options(negative_power = TRUE) # division becomes ^-1
plot(displacement, consumption) # division in consumption
```

As usual, units modify automatically in expressions:

```{r}
units_options(negative_power = TRUE) # division becomes ^-1
par(mar = mar)
plot(displacement, consumption)
plot(1/displacement, 1/consumption)
```

```{r echo=FALSE}
units_options(negative_power = FALSE) # division becomes /
```

---
title: "Units of Measurement for R Vectors: an Introduction"
output:
  rmarkdown::html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{Units of Measurement for R Vectors: an Introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r echo=FALSE}
knitr::opts_chunk$set(collapse = TRUE, fig.asp = 0.7, fig.width = 7)
```

```{r echo=FALSE}
units:::units_options(negative_power = FALSE)
```

R has little support for physical measurement units.  The exception
is formed by time differences: time differences objects of class
`difftime` have a `units` attribute that can be modified:

```{r}
t1 = Sys.time() 
t2 = t1 + 3600 
d = t2 - t1
class(d)
units(d)
d
units(d) = "secs"
d
```
We see here that the `units` method is used to retrieve and modify the
unit of time differences. 

The `units` package generalizes this idea to other physical units, building upon the
[udunits2](https://www.unidata.ucar.edu/software/udunits/) C library.
The `udunits2` library provides the following operations:

* validating whether an expression, such as `m/s` is a valid physical unit
* verifying whether two units such as `m/s` and `km/h` are convertible
* converting values between two convertible units 
* providing names and symbols for specific units
* handle different character encodings (utf8, ascii, iso-8859-1 and latin1)

The `units` R package uses the
[udunits2](https://www.unidata.ucar.edu/software/udunits/) C library to extend
R with functionality for manipulating numeric vectors that have
physical measurement units associated with them, in a similar way as
`difftime` objects behave.

## Setting units, unit conversion

We can set units to numerical values by `set_units`:
```{r}
library(units)
(a <- set_units(runif(10),  m/s))
```
the result, e.g. 
```{r}
set_units(10, m/s)
```
literally means "10 times 1 m divided by 1 s". In writing, the "1"
values are omitted, and the multiplication is implicit.

### Unit conversion

When conversion is meaningful, such as hours to seconds or meters to kilometers, conversion can be done explicitly by setting the units of a vector
```{r}
b = a
units(b) <- make_units(km/h)
b
```

## Basic manipulations

### Arithmetic operations

Arithmetic operations verify units, and create new ones
```{r}
a + a
a * a
a ^ 2
a ** -2
```
and convert to the units of the first argument if necessary:
```{r}
a + b # m/s + km/h -> m/s
```

Currently, powers are only supported for integer powers, so using `a ** 2.5` would result in an error.


### Unit simplification
There are some basic simplification of units:
```{r}
t <- make_units(s)
a * t
```
which also work when units need to be converted before they can be simplified:
```{r}
t <- make_units(min)
a * t
```
Simplification to unit-less values gives the "1" as unit:
```{r}
m <- make_units(m)
a * t / m
```

Allowed operations that require convertible units are `+`, `-`, `==`,
`!=`, `<`, `>`, `<=`, `>=`.  Operations that lead to new units are
`*`, `/`, and the power operations `**` and `^`.


### Mathematical functions

Mathematical operations allowed are: `abs`, `sign`, `floor`,
`ceiling`, `trunc`, `round`, `signif`, `log`, `cumsum`, `cummax`, `cummin`.
```{r}
signif(a ** 2 / 3, 3)
cumsum(a)
log(a) # base defaults to exp(1)
log(a, base = 10)
log(a, base = 2)
```

### Summary functions
Summary functions `sum`, `min`, `max`, and `range` are allowed:
```{r}
sum(a)
min(a)
max(a)
range(a)
make_units(min(m/s, km/h)) # converts to first unit:
```

### Printing
Following `difftime`, printing behaves differently for length-one vectors:
```{r}
a
a[1]
```

### Subsetting
The usual subsetting rules work:
```{r}
a[2:5]
a[-(1:9)]
```

### Concatenation
```{r}
c(a,a)
```
concatenation converts to the units of the first argument, if necessary:
```{r}
c(a,b) # m/s, km/h -> m/s
c(b,a) # km/h, m/s -> km/h
```

## Conversion to/from `difftime`
From `difftime` to `units`:
```{r}
t1 = Sys.time() 
t2 = t1 + 3600 
d = t2 - t1
(du = as_units(d))
```
vice versa:
```{r}
(dt = as_difftime(du))
class(dt)
```

## units in `matrix` objects
```{r}
set_units(matrix(1:4,2,2), m/s)
set_units(matrix(1:4,2,2), m/s * m/s)
```
but
```{r}
set_units(matrix(1:4,2,2), m/s) %*% set_units(4:3, m/s)
```
strips units.

## units objects in `data.frame`s
units in `data.frame` objects are printed, but do not appear in `summary`:.
```{r}
set.seed(131)
d <- data.frame(x = runif(4), 
                    y = set_units(runif(4), s), 
                    z = set_units(1:4, m/s))
d
summary(d)
d$yz = with(d, y * z)
d
d[1, "yz"]
```
## Formatting

Units are often written in the form `m2 s-1`, for square meter per second. This
can be defined as unit, and also parsed by `as_units`:
```{r}
(x = 1:10 * as_units("m2 s-1"))
```
udunits understands such string, and can convert them
```{r}
y = 1:10 * make_units(m^2/s)
x + y
```
Printing units in this form is done by
```{r}
deparse_unit(x)
```

## Plotting

Base scatter plots and histograms support automatic unit placement
in axis labels. In the following example we first convert to
SI units. (Unit `in` needs a bit special treatment, because `in` is a
reserved word in R.)

```{r}
mar = par("mar") + c(0, .3, 0, 0)
displacement = mtcars$disp * as_units("in")^3
units(displacement) = make_units(cm^3)
weight = mtcars$wt * 1000 * make_units(lb)
units(weight) = make_units(kg)
par(mar = mar)
plot(weight, displacement)
```

We can change grouping symbols from `[ ]` into `( )`:
```{r}
units_options(group = c("(", ")") )  # parenthesis instead of square brackets
par(mar = mar)
plot(weight, displacement)
```

We can also remove grouping symbols, increase space between variable name and unit by:
```{r}
units_options(sep = c("~~~", "~"), group = c("", ""))  # no brackets; extra space
par(mar = mar)
plot(weight, displacement)
```

More complex units can be plotted either with negative powers, or as divisions,
by modifying one of `units`'s global options using `units_options`:

```{r}
gallon = as_units("gallon")
consumption = mtcars$mpg * make_units(mi/gallon)
units(consumption) = make_units(km/l)
par(mar = mar)
plot(displacement, consumption) # division in consumption
units_options(negative_power = TRUE) # division becomes ^-1
plot(displacement, consumption) # division in consumption
```

As usual, units modify automatically in expressions:

```{r}
units_options(negative_power = TRUE) # division becomes ^-1
par(mar = mar)
plot(displacement, consumption)
plot(1/displacement, 1/consumption)
```

```{r echo=FALSE}
units_options(negative_power = FALSE) # division becomes /
```

# pillar methods are available for units objects

    Code
      pillar::pillar(x)
    Output
      <pillar>
      [km]
         1
         2
         3
    Code
      pillar::pillar(m)
    Output
      <pillar>
      <mixed_units>
             1 [km]
             2 [km]
             3 [km]
              4 [g]
              5 [g]
              6 [g]

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.units}
\alias{plot.units}
\alias{make_unit_label}
\title{Plot \code{units} objects}
\usage{
make_unit_label(lab, u, sep = units_options("sep"),
  group = units_options("group"), parse = units_options("parse"))

\method{plot}{units}(x, y, xlab = NULL, ylab = NULL, ...)
}
\arguments{
\item{lab}{length one character; name of the variable to plot}

\item{u}{vector of class \code{units}}

\item{sep}{length two character vector, defaulting to \code{c("~","~")}, with
the white space between unit name and unit symbols, and between subsequent
symbols.}

\item{group}{length two character vector with grouping symbols, e.g.
\code{c("(",")")} for parenthesis, or \code{c("","")} for no group symbols}

\item{parse}{logical; indicates whether a parseable expression should be
returned (typically needed for super scripts), or a simple character string
without special formatting.}

\item{x}{object of class units, to plot along the x axis, or, if y is missing, along the y axis}

\item{y}{object to plot along the y axis, or missing}

\item{xlab}{character; x axis label}

\item{ylab}{character; y axis label}

\item{...}{other parameters, passed on to \link{plot.default}}
}
\description{
Create axis label with appropriate labels.

Plot method for \code{units} objects.
}
\details{
\link{units_options} can be used to set and change the defaults for
  \code{sep}, \code{group} and \code{doParse}.
}
\examples{
displacement = mtcars$disp * as_units("in")^3
units(displacement) = make_units(cm^3)
weight = mtcars$wt * 1000 * make_units(lb)
units(weight) = make_units(kg)
plot(weight, displacement)

units_options(group = c("(", ")") )  # parenthesis instead of square brackets
plot(weight, displacement)

units_options(sep = c("~~~", "~"), group = c("", ""))  # no brackets; extra space
plot(weight, displacement)

units_options(sep = c("~", "~~"), group = c("[", "]"))
gallon = as_units("gallon")
consumption = mtcars$mpg * make_units(mi/gallon)
units(consumption) = make_units(km/l)
plot(displacement, consumption) # division in consumption

units_options(negative_power = TRUE) # division becomes ^-1
plot(displacement, consumption)

plot(1/displacement, 1/consumption)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/database.R
\name{load_units_xml}
\alias{load_units_xml}
\title{Load a unit system}
\usage{
load_units_xml(path = default_units_xml())
}
\arguments{
\item{path}{a path to a valid unit system in XML format.}
}
\description{
Load an XML database containing a unit system compatible with UDUNITS2.
}
\details{
A unit system comprises a root \code{<unit-system>} and a number of children
defining prefixes (\code{<prefix>}) or units (\code{<unit>}).
See the contents of

\code{system.file("share/udunits", package="units")}

for examples.
}
\examples{
# load a new unit system
load_units_xml(system.file("share/udunits/udunits2-base.xml", package="units"))
\dontrun{
set_units(1, rad) # doesn'twork
}

# reload the default unit system
load_units_xml()
set_units(1, rad) # works again

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale_units.R
\name{scale_units}
\alias{scale_x_units}
\alias{scale_y_units}
\alias{scale_colour_units}
\alias{scale_color_units}
\alias{scale_fill_units}
\alias{scale_alpha_units}
\alias{scale_size_units}
\alias{scale_size_area_units}
\alias{scale_radius_units}
\alias{scale_linewidth_units}
\title{Continuous scales for units data}
\usage{
scale_x_units(..., sec.axis = ggplot2::waiver(), unit = NULL)

scale_y_units(..., sec.axis = ggplot2::waiver(), unit = NULL)

scale_colour_units(..., unit = NULL)

scale_color_units(..., unit = NULL)

scale_fill_units(..., unit = NULL)

scale_alpha_units(..., unit = NULL)

scale_size_units(..., unit = NULL)

scale_size_area_units(..., unit = NULL)

scale_radius_units(..., unit = NULL)

scale_linewidth_units(..., unit = NULL)
}
\arguments{
\item{...}{arguments passed on to the corresponding continuous scale
(see the manual page for each \code{scale_{type}} for details).}

\item{sec.axis}{\code{\link[ggplot2:sec_axis]{sec_axis()}} is used to specify a secondary axis.}

\item{unit}{A unit specification to use for the guide. If given, the values
will be converted to this unit before plotting. An error will be thrown if
the specified unit is incompatible with the unit of the data.}
}
\description{
These are the default scales for the \code{units} class. These will usually
be added automatically. To override manually, use \code{scale_{type}_units}.
}
\examples{
if (requireNamespace("ggplot2", quietly=TRUE)) {

library(ggplot2)

mtcars$consumption <- set_units(mtcars$mpg, mi / gallon)
mtcars$power <- set_units(mtcars$hp, hp)

# Use units encoded into the data
ggplot(mtcars) +
  geom_point(aes(power, consumption))

# Convert units on the fly during plotting
ggplot(mtcars) +
  geom_point(aes(power, consumption)) +
  scale_x_units(unit = "W") +
  scale_y_units(unit = "km/l")

# Resolve units when transforming data
ggplot(mtcars) +
  geom_point(aes(power, 1 / consumption))

# Reverse the y axis
ggplot(mtcars) +
  geom_point(aes(power, consumption)) +
  scale_y_units(transform="reverse")

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{cbind.units}
\alias{cbind.units}
\alias{rbind.units}
\title{Combine R Objects by Rows or Columns}
\usage{
\method{cbind}{units}(..., deparse.level = 1)

\method{rbind}{units}(..., deparse.level = 1)
}
\arguments{
\item{...}{(generalized) vectors or matrices.  These can be given as named
    arguments.  Other \R objects may be coerced as appropriate, or S4
    methods may be used: see sections \sQuote{Details} and
    \sQuote{Value}.  (For the \code{"data.frame"} method of \code{cbind}
    these can be further arguments to \code{\link[base]{data.frame}} such as
    \code{stringsAsFactors}.)}

\item{deparse.level}{integer controlling the construction of labels in
    the case of non-matrix-like arguments (for the default method):\cr
    \code{deparse.level = 0} constructs no labels;\cr
    the default \code{deparse.level = 1} typically and
    \code{deparse.level = 2} always construct labels from the argument
    names, see the \sQuote{Value} section below.}
}
\description{
S3 methods for \code{units} objects (see \code{\link[base]{cbind}}).
}
\examples{
x <- set_units(1, m/s)
y <- set_units(1:3, m/s)
z <- set_units(8:10, m/s)
(m <- cbind(x, y)) # the '1' (= shorter vector) is recycled
(m <- cbind(m, z)[, c(1, 3, 2)]) # insert a column
(m <- rbind(m, z)) # insert a row
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/udunits.R
\name{ud_are_convertible}
\alias{ud_are_convertible}
\title{Test if two units are convertible}
\usage{
ud_are_convertible(x, y)
}
\arguments{
\item{x}{character or object of class \code{symbolic_units}, for the symbol
of the first unit.}

\item{y}{character or object of class \code{symbolic_units}, for the symbol
of the second unit.}
}
\value{
boolean, \code{TRUE} if both units exist and are convertible.
}
\description{
Parses and checks whether units can be converted by UDUNITS-2. Units may not
be convertible either because they are different magnitudes or because one
(or both) units are not defined in the database.
}
\examples{
ud_are_convertible("m", "km")
a <- set_units(1:3, m/s)
ud_are_convertible(units(a), "km/h")
ud_are_convertible("s", "kg")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user_conversion.R
\name{install_unit}
\alias{install_unit}
\alias{remove_unit}
\title{Define or remove units}
\usage{
install_unit(symbol = character(0), def = character(0),
  name = character(0))

remove_unit(symbol = character(0), name = character(0))
}
\arguments{
\item{symbol}{a vector of symbols to be installed/removed.}

\item{def}{either \itemize{
  \item an empty definition, which defines a new base unit;
  \item \code{"unitless"}, which defines a new dimensionless unit;
  \item a relationship with existing units (see details for the syntax).
}}

\item{name}{a vector of names to be installed/removed.}
}
\description{
Installing new symbols and/or names allows them to be used in \code{as_units},
\code{make_units} and \code{set_units}. Optionally, a relationship can be
defined between such symbols/names and existing ones (see details and examples).
}
\details{
At least one symbol or name is expected, but multiple symbols and/or names
can be installed (and thus mapped to the same unit) or removed at the same
time. The \code{def} argument enables arbitrary relationships with existing
units using UDUNITS-2 syntax:
\tabular{llll}{
  \strong{String Type} \tab \strong{Using Names} \tab \strong{Using Symbols}
    \tab \strong{Comment}\cr
  Simple \tab meter \tab m \tab \cr
  Raised \tab meter^2 \tab m2 \tab
    higher precedence than multiplying or dividing\cr
  Product \tab newton meter \tab N.m \tab \cr
  Quotient \tab meter per second \tab m/s \tab \cr
  Scaled \tab 60 second \tab 60 s \tab \cr
  Prefixed \tab kilometer \tab km \tab \cr
  Offset \tab kelvin from 273.15 \tab K @ 273.15 \tab
    lower precedence than multiplying or dividing\cr
  Logarithmic \tab lg(re milliwatt) \tab lg(re mW) \tab
    "lg" is base 10, "ln" is base e, and "lb" is base 2\cr
  Grouped \tab (5 meter)/(30 second) \tab (5 m)/(30 s) \tab
}
The above may be combined, e.g., \code{"0.1 lg(re m/(5 s)^2) @ 50"}.
You may also look at the \code{<def>} elements in the units database to see
examples of string unit specifications.
}
\examples{
# define a fortnight
install_unit("fn", "2 week", "fortnight")
year <- as_units("year")
set_units(year, fn)        # by symbol
set_units(year, fortnight) # by name
# clean up
remove_unit("fn", "fortnight")

# working with currencies
install_unit("dollar")
install_unit("euro", "1.22 dollar")
install_unit("yen", "0.0079 euro")
set_units(as_units("dollar"), yen)
# clean up
remove_unit(c("dollar", "euro", "yen"))

# an example from microbiology
cfu_symbols <- c("CFU", "cfu")
cfu_names <- c("colony_forming_unit", "ColonyFormingUnit")
install_unit("cell")
install_unit(cfu_symbols, "3.4 cell", cfu_names)
cell <- set_units(2.5e5, cell)
vol <- set_units(500, ul)
set_units(cell/vol, "cfu/ml")
set_units(cell/vol, "CFU/ml")
set_units(cell/vol, "colony_forming_unit/ml")
set_units(cell/vol, "ColonyFormingUnit/ml")
# clean up
remove_unit(c("cell", cfu_symbols), cfu_names)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{deparse_unit}
\alias{deparse_unit}
\title{deparse unit to string in product power form (e.g. km m-2 s-1)}
\usage{
deparse_unit(x)
}
\arguments{
\item{x}{object of class units}
}
\value{
length one character vector
}
\description{
deparse unit to string in product power form (e.g. km m-2 s-1)
}
\examples{
u = as_units("kg m-2 s-1", implicit_exponents = TRUE)
u
deparse_unit(u)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symbolic_units.R
\docType{data}
\name{unitless}
\alias{unitless}
\title{The "unit" type for vectors that are actually dimension-less.}
\format{
An object of class \code{symbolic_units} of length 2.
}
\usage{
unitless
}
\description{
The "unit" type for vectors that are actually dimension-less.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/math.R
\name{Math.units}
\alias{Math.units}
\title{Mathematical operations for units objects}
\usage{
\method{Math}{units}(x, ...)
}
\arguments{
\item{x}{object of class units}

\item{...}{parameters passed on to the Math functions}
}
\description{
Mathematical operations for units objects
}
\details{
Logarithms receive a special treatment by the underlying \pkg{udunits2}
library. If a natural logarithm is applied to some \code{unit}, the result is
\code{ln(re 1 unit)}, which means \emph{natural logarithm referenced to
\code{1 unit}}. For base 2 and base 10 logarithms, the output \code{lb(...)}
and \code{lg(...)} respectively instead of \code{ln(...)}.

This is particularly important for some units that are typically expressed in
a logarithmic scale (i.e., \emph{bels}, or, more commonly, \emph{decibels}),
such as Watts or Volts. For some of these units, the default \pkg{udunits2}
database contains aliases: e.g., \code{BW} (bel-Watts) is an alias of
\code{lg(re 1 W)}; \code{Bm} (bel-milliWatts) is an alias of
\code{lg(re 0.001 W)}; \code{BV} is an alias of \code{lg(re 1 V)} (bel-Volts),
and so on and so forth (see the output of \code{valid_udunits()} for further
reference).

Additionally, the \pkg{units} package defines \code{B}, the \emph{bel}, by
default (because it is not defined by \pkg{udunits2}) as an alias of
\code{lg(re 1)}, unless a user-provided XML database already contains a
definition of \code{B}, or the \code{define_bel} option is set to \code{FALSE}
(see \code{help(units_options)}).
}
\examples{
# roundings, cummulative functions
x <- set_units(sqrt(1:10), m/s)
signif(x, 2)
cumsum(x)

# trigonometry
sin(x) # not meaningful
x <- set_units(sqrt(1:10), rad)
sin(x)
cos(x)
x <- set_units(seq(0, 1, 0.1), 1)
asin(x)
acos(x)

# logarithms
x <- set_units(sqrt(1:10), W)
log(x) # base exp(1)
log(x, base = 3)
log2(x)
log10(x)
set_units(x, dBW) # decibel-watts
set_units(x, dBm) # decibel-milliwatts
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{units_options}
\alias{units_options}
\title{set one or more units global options}
\usage{
units_options(..., sep, group, negative_power, parse, set_units_mode,
  auto_convert_names_to_symbols, simplify, allow_mixed, unitless_symbol,
  define_bel)
}
\arguments{
\item{...}{named options (character) for which the value is queried}

\item{sep}{character length two; default \code{c("~", "~")}; space separator between variable and units, and space separator between two different units}

\item{group}{character length two; start and end group, may be two empty strings, a parenthesis pair, or square brackets; default: square brackets.}

\item{negative_power}{logical, default \code{FALSE}; should denominators have negative power, or follow a division symbol?}

\item{parse}{logical, default \code{TRUE}; should the units be made into an expression (so we get subscripts)? Setting to \code{FALSE} may be useful if \link{parse} fails, e.g. if the unit contains symbols that assume a particular encoding}

\item{set_units_mode}{character; either \code{"symbols"} or \code{"standard"}; see \link{set_units}; default is \code{"symbols"}}

\item{auto_convert_names_to_symbols}{logical, default \code{TRUE}: should names, such as \code{degree_C} be converted to their usual symbol?}

\item{simplify}{logical, default \code{NA}; simplify units in expressions?}

\item{allow_mixed}{logical; if \code{TRUE}, combining mixed units creates a \code{mixed_units} object, if \code{FALSE} it generates an error}

\item{unitless_symbol}{character; set the symbol to use for unitless (1) units}

\item{define_bel}{logical; if \code{TRUE}, define the unit \code{B} (i.e., the \emph{bel}, widely used with the \emph{deci-} prefix as \code{dB}, \emph{decibel}) as an alias of \code{lg(re 1)}. \code{TRUE} by default, unless \code{B} is already defined in the existing XML database.}
}
\value{
in case options are set, invisibly a named list with the option values that are being set; if an option is queried, the current option value.
}
\description{
set units global options, mostly related how units are printed and plotted
}
\details{
This sets or gets units options. Set them by using named arguments, get them by passing the option name.

The default \code{NA} value for \code{simplify} means units are not simplified in \link{set_units} or \link{as_units}, but are simplified in arithmetical expressions.
}
\examples{
old = units_options(sep = c("~~~", "~"), group = c("", "")) # more space, parenthesis
old
## set back to defaults:
units_options(sep = c("~", "~"), group = c("[", "]"), negative_power = FALSE, parse = TRUE)
units_options("group")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conversion.R
\name{as_difftime}
\alias{as_difftime}
\title{convert units object into difftime object}
\usage{
as_difftime(x)
}
\arguments{
\item{x}{object of class \code{units}}
}
\description{
convert units object into difftime object
}
\examples{

t1 = Sys.time()
t2 = t1 + 3600
d = t2 - t1
du <- as_units(d)
dt = as_difftime(du)
class(dt)
dt
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conversion.R
\name{units}
\alias{units}
\alias{units<-.numeric}
\alias{units<-.units}
\alias{units<-.logical}
\alias{units.units}
\title{Set measurement units on a numeric vector}
\usage{
\method{units}{numeric}(x) <- value

\method{units}{units}(x) <- value

\method{units}{logical}(x) <- value

\method{units}{units}(x)
}
\arguments{
\item{x}{numeric vector, or object of class \code{units}}

\item{value}{object of class \code{units} or \code{symbolic_units}, or in the case of \code{set_units} expression with symbols that can be resolved in \link{ud_units} (see examples).}
}
\value{
object of class \code{units}

the units method retrieves the units attribute, which is of class \code{symbolic_units}
}
\description{
Set measurement units on a numeric vector

Convert units

retrieve measurement units from \code{units} object
}
\details{
if \code{value} is of class \code{units} and has a value unequal to 1, this value is ignored unless \code{units_options("simplifiy")} is \code{TRUE}. If \code{simplify} is \code{TRUE}, \code{x} is multiplied by this value.
}
\examples{
x = 1:3
class(x)
units(x) <- as_units("m/s")
class(x)
y = 2:5
a <- set_units(1:3, m/s)
units(a) <- with(ud_units, km/h)
a
# convert to a mixed_units object:
units(a) = c("m/s", "km/h", "km/h")
a
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{keep_units}
\alias{keep_units}
\title{Apply a function keeping units}
\usage{
keep_units(FUN, x, ..., unit = units(x))
}
\arguments{
\item{FUN}{the function to be applied.}

\item{x}{first argument of \code{FUN}, of class \code{units}.}

\item{...}{optional arguments to \code{FUN}.}

\item{unit}{symbolic unit to restore after \code{FUN}.}
}
\value{
An object of class \code{units}.
}
\description{
Helper function to apply a function to a \code{units} object and then restore
the original units.
}
\details{
Provided for incompatible functions that do not preserve units. The user is
responsible for ensuring the correctness of the output.

If \code{x} is not a \code{units} object
and \code{unit} is not provided by the user,
a warning is issued, and the output will also have no units
(see examples).
}
\examples{
x <- set_units(1:5, m)
keep_units(drop_units, x)

# An example use case is with random number generating functions:
mu <- as_units(10, "years")
keep_units(rnorm, n = 1, x = mu)

# units can be directly specified if needed; for example, with
# `rexp()`, the units of the rate parameter are the inverse of
# the units of the output:
rate <- as_units(3, "1/year")
keep_units(rexp, n = 1, x = rate, unit = units(1/rate))

# if `x` does not actually have units, a warning is issued,
# and the output has no units:
rate2 <- 3
keep_units(rexp, n = 1, x = rate2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user_conversion.R
\name{install_symbolic_unit}
\alias{install_symbolic_unit}
\alias{remove_symbolic_unit}
\title{Define new symbolic units}
\usage{
install_symbolic_unit(name, warn = TRUE, dimensionless = TRUE)

remove_symbolic_unit(name)
}
\arguments{
\item{name}{a length 1 character vector that is the unit name or symbol.}

\item{warn}{warns if the supplied unit symbol is already a valid unit symbol
recognized by udunits.}

\item{dimensionless}{logical; if \code{TRUE}, a new dimensionless unit is created, if \code{FALSE} a new base unit is created. Dimensionless units are convertible to other dimensionless units (such as \code{rad}), new base units are not convertible to other existing units.}
}
\description{
Adding a symbolic unit allows it to be used in \code{as_units},
\code{make_units} and \code{set_units}. No installation is performed if the
unit is already known by udunits.
}
\details{
\code{install_symbolic_unit} installs a new dimensionless unit; these are directly compatible to any other dimensionless unit. To install a new unit that is a scaled or shifted version of an existing unit, use \code{install_conversion_constant} or \code{install_conversion_offset} directly.
}
\examples{
install_symbolic_unit("person")
set_units(1, rad) + set_units(1, person) # that is how dimensionless units work!
}
\seealso{
\code{\link{install_conversion_constant}}, \code{\link{install_conversion_offset}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symbolic_units.R
\name{convert_to_base}
\alias{convert_to_base}
\title{Convert units to their base units}
\usage{
convert_to_base(x, simplify = TRUE, keep_fraction = TRUE)
}
\arguments{
\item{x}{object of class \code{units}.}

\item{simplify}{logical; if TRUE (default), the resulting units are simplified.}

\item{keep_fraction}{logical; if TRUE (default), the result is kept as a fraction.}
}
\value{
object of class \code{units} with units converted to base units.
}
\description{
Convert the units of a \code{units} object to their base units, as defined by
the udunits database (SI units).
}
\examples{
x <- set_units(32, mJ/g)
convert_to_base(x)
convert_to_base(x, keep_fraction=FALSE)
convert_to_base(x, simplify=FALSE)
convert_to_base(x, simplify=FALSE, keep_fraction=FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user_conversion.R
\name{install_conversion_constant}
\alias{install_conversion_constant}
\alias{install_conversion_offset}
\title{Install a conversion constant or offset between user-defined units.}
\usage{
install_conversion_constant(from, to, const)

install_conversion_offset(from, to, const)
}
\arguments{
\item{from}{String for the symbol of the unit being converted from.}

\item{to}{String for the symbol of the unit being converted to. One of \code{from} and \code{to} must be an existing unit name.}

\item{const}{The constant \eqn{\alpha} in the conversion.}
}
\description{
Tells the \code{units} package how to convert between units that
  have a linear relationship, i.e. can be related on the form \eqn{y = \alpha
  x} (constant) or \eqn{y = \alpha + x} (offset).
}
\details{
This function handles the very common case where units are related 
  through a linear function, that is, you can convert from one to the other 
  as \eqn{y = \alpha x}. Using this function, you specify that you
  can go from values of type \code{from} to values of type \code{to} by 
  multiplying by a constant, or adding a constant.
}
\examples{

# one orange is worth two apples
install_symbolic_unit("orange")
install_conversion_constant("orange", "apple", 2) # apple = 2 * orange
apples <- 2 * as_units("apple")
oranges <- 1 * as_units("orange")
apples + oranges
oranges + apples

install_conversion_offset("meter", "newmeter", 1)
m = set_units(1:3, meter)
n = set_units(1:3, newmeter)
m + n
n + m
}
\seealso{
\code{\link{install_symbolic_unit}}, \code{\link{remove_symbolic_unit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/valid_udunits.R
\docType{data}
\name{ud_units}
\alias{ud_units}
\title{List containing pre-defined units from the udunits2 package.}
\format{
An object of class \code{NULL} of length 0.
}
\usage{
ud_units
}
\description{
Lazy loaded when used
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defunct.R
\name{units-defunct}
\alias{units-defunct}
\title{Defunct functions in \pkg{units}}
\description{
These functions are no longer available.
}
\details{
\itemize{
 \item \code{ud_units}: Use \code{\link{as_units}} instead.
 \item \code{as.units}: Use \code{\link{as_units}} instead.
 \item \code{make_unit}: Use \code{\link{as_units}} instead.
 \item \code{parse_unit}: Use \code{\link{as_units}} instead.
 \item \code{as_cf}: Use \code{\link{deparse_unit}} instead.
 \item \code{install_symbolic_unit}: Use \code{\link{install_unit}} instead.
 \item \code{remove_symbolic_unit}: Use \code{\link{remove_unit}} instead.
 \item \code{install_conversion_constant}: Use \code{\link{install_unit}} instead.
 \item \code{install_conversion_offset}: Use \code{\link{install_unit}} instead.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_units.R
\name{set_units}
\alias{set_units}
\title{set_units}
\usage{
set_units(x, value, ..., mode = units_options("set_units_mode"))
}
\arguments{
\item{x}{a numeric to be assigned units, or a units object to have units
converted}

\item{value}{a \code{units} object, or something coercible to one with
\code{as_units}. Depending on \code{mode}, the unit is constructed from the
supplied bare expression or from the supplied value via standard
evaluation.}

\item{...}{passed on to \code{as_units}}

\item{mode}{if \code{"symbols"} (the default), then unit is constructed from
the expression supplied. Otherwise, if\code{mode = "standard"},
standard evaluation is used for the supplied value This argument can be set
via a global option \code{units_options(set_units_mode = "standard")}}
}
\description{
A pipe friendly version of \code{units<-}
}
\seealso{
\code{\link{as_units}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixed.R
\name{mixed_units}
\alias{mixed_units}
\alias{units<-.mixed_units}
\title{Create or convert to a mixed units list-column}
\usage{
mixed_units(x, values, ...)

\method{units}{mixed_units}(x) <- value
}
\arguments{
\item{x}{numeric, or vector of class \code{units}}

\item{values}{character vector with units encodings, or list with symbolic units of class \code{mixed_symbolic_units}}

\item{...}{ignored}

\item{value}{see values}
}
\description{
Create or convert to a mixed units list-column
}
\details{
if \code{x} is of class \code{units}, \code{values} should be missing or of class \code{mixed_symbolic_units}; if \code{x} is numeric, \code{values} should be a character vector the length of \code{x}.
}
\examples{
a <- 1:4
u <- c("m/s", "km/h", "mg/L", "g")
mixed_units(a, u)
units(a) = as_units("m/s")
mixed_units(a) # converts to mixed representation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{make_unit}
\alias{make_unit}
\alias{parse_unit}
\alias{deprecated}
\alias{as.units}
\title{Deprecated functions}
\usage{
make_unit(chr)

parse_unit(chr)

as.units(x, value = unitless)
}
\arguments{
\item{chr}{length 1 character string}

\item{x}{a numeric}

\item{value}{a units object, by default, unitless}
}
\description{
The following functions are deprecated and will be removed in a future release.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arith.R
\name{Ops.units}
\alias{Ops.units}
\title{S3 Ops Group Generic Functions for units objects}
\usage{
\method{Ops}{units}(e1, e2)
}
\arguments{
\item{e1}{object of class \code{units},
or something that can be coerced to it by \code{as_units(e1)}}

\item{e2}{object of class \code{units},
or something that can be coerced to it by \code{as_units(e2)},
or in case of power a number (integer n or 1/n)}
}
\value{
object of class \code{units}
}
\description{
Ops functions for units objects, including comparison, product and divide,
add, subtract.
}
\details{
Users are advised against performing arithmetical operations with
temperatures in different units. The \pkg{units} package ensure that results
1) are arithmetically correct, and 2) satisfy dimensional analysis, but could
never ensure that results are physically meaningful. Temperature units are
special because there is an absolute unit, Kelvin, and relative ones, Celsius
and Fahrenheit degrees. Arithmetic operations between them are meaningless
from the physical standpoint. Users are thus advised to convert all
temperatures to Kelvin before operating.
}
\examples{
a <- set_units(1:3, m/s)
b <- set_units(1:3, m/s)
a + b
a * b
a / b
a <- as_units("kg m-3")
b <- set_units(1, kg/m/m/m)
a + b
a = set_units(1:5, m)
a \%/\% a
a \%/\% set_units(2)
set_units(1:5, m^2) \%/\% set_units(2, m)
a \%\% a
a \%\% set_units(2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/udunits.R
\name{udunits2}
\alias{udunits2}
\alias{ud_are_convertible}
\alias{ud_convert}
\title{\pkg{udunits2} utilities}
\usage{
ud_are_convertible(from, to, ...)

ud_convert(x, from, to)
}
\arguments{
\item{from, to}{character vector or object of class \code{symbolic_units},
for the symbol(s) of the original unit(s) and the unit to convert to respectively.}

\item{...}{unused.}

\item{x}{numeric vector}
}
\value{
\code{ud_are_convertible}
returns \code{TRUE} if both units exist and are convertible,
\code{FALSE} otherwise.

\code{ud_convert}
returns a numeric vector with \code{x} converted to new unit.
}
\description{
Some \pkg{udunits2} utilities are exposed to the user. These functions are
useful for checking whether units are convertible or converting between units
without having to create \pkg{units} objects.
Arguments are recycled if necessary.
}
\examples{
ud_are_convertible(c("m", "mm"), "km")
ud_convert(c(100, 100000), c("m", "mm"), "km")

a <- set_units(1:3, m/s)
ud_are_convertible(units(a), "km/h")
ud_convert(1:3, units(a), "km/h")

ud_are_convertible("degF", "degC")
ud_convert(32, "degF", "degC")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conversion.R, R/make_units.R
\name{as_units}
\alias{as_units}
\alias{as_units.default}
\alias{as_units.difftime}
\alias{make_units}
\alias{as_units.character}
\alias{as_units.call}
\title{convert object to a units object}
\usage{
as_units(x, ...)

\method{as_units}{default}(x, value = unitless, ...)

\method{as_units}{difftime}(x, value, ...)

make_units(bare_expression, check_is_valid = TRUE)

\method{as_units}{character}(x, check_is_valid = TRUE,
  implicit_exponents = NULL, force_single_symbol = FALSE, ...)

\method{as_units}{call}(x, check_is_valid = TRUE, ...)
}
\arguments{
\item{x}{object of class units}

\item{...}{passed on to other methods}

\item{value}{an object of class units, or something coercible to one with
\code{as_units}}

\item{bare_expression}{a bare R expression describing units. Must be valid R
syntax (reserved R syntax words like \code{in} must be backticked)}

\item{check_is_valid}{throw an error if all the unit symbols are not either
recognized by udunits2 via \code{ud_is_parseable()}, or a custom
user defined via \code{install_symbolic_unit()}. If \code{FALSE}, no check
for validity is performed.}

\item{implicit_exponents}{If the unit string is in product power form (e.g.
\code{"km m-2 s-1"}). Defaults to \code{NULL}, in which case a guess is made
based on the supplied string. Set to \code{TRUE} or \code{FALSE} if the guess is
incorrect.}

\item{force_single_symbol}{Whether to perform no string parsing and force
treatment of the string as a single symbol.}
}
\value{
A new unit object that can be used in arithmetic, unit conversion or
  unit assignment.
}
\description{
A number of functions are provided for creating unit objects. 
\itemize{
    \item \code{as_units}, a generic with methods for a
    character string and for quoted language. Note, direct usage of this function
    by users is typically not necessary, as coercion via \code{as_units} is
    automatically done with \code{`units<-`} and \code{set_units()}.
   
    \item \code{make_units()}, constructs units from bare expressions.
    \code{make_units(m/s)} is equivalent to \code{as_units(quote(m/s))}
   
    \item \code{set_units()}, a pipe_friendly version of \code{`units<-`}. By
    default it operates with bare expressions like \code{make_unit}, but this
    behavior can be disabled by a specifying \code{mode = "standard"} or setting 
    \code{units_options(set_units_mode = "standard")}.
}
}
\note{
By default, unit names are automatically substituted with unit names
  (e.g., kilogram --> kg). To turn off this behavior, set
  \code{units_options(auto_convert_names_to_symbols = FALSE)}
}
\section{Character strings}{


  Generally speaking, there are 3 types of unit strings are accepted in
  \code{as_units} (and by extension, \code{`units<-`}).

  The first, and likely most common, is a "standard" format unit
  specification where the relationship between unit symbols or names is
  specified explicitly with arithmetic symbols for division \code{/},
  multiplication \code{*} and power exponents \code{^}, or other mathematical
  functions like \code{log()}. In this case, the string is parsed as an R
  expression via \code{parse(text = )} after backticking all unit symbols and
  names, and then passed on to \code{as_units.call()}. A heuristic is used to
  perform backticking, such that any continuous set of characters
  uninterrupted by one of \code{()\\*^-} are backticked (unless the character
  sequence consists solely of numbers \code{0-9}), with some care to not
  double up on pre-existing backticks. This heuristic appears to be quite
  robust, and works for units would otherwise not be valid R syntax. For
  example, percent (\code{"\%"}), feet (\code{"'"}), inches (\code{"in"}),
  and Tesla (\code{"T"}) are all backticked and parsed correctly.

  Nevertheless, for certain complex unit expressions, this backticking heuristic
  may give incorrect results.  If the string supplied fails to parse as an R
  expression, then the string is treated as a single symbolic unit and
  \code{symbolic_unit(chr)} is used as a fallback with a warning. In that
  case, automatic unit simplification may not work properly when performing
  operations on unit objects, but unit conversion and other Math operations
  should still give correct results so long as the unit string supplied
  returns \code{TRUE} for \code{ud_is_parsable()}.

  The second type of unit string accepted is one with implicit exponents. In
  this format, \code{/}, \code{*}, and \code{^}, may not be present in the
  string, and unit symbol or names must be separated by a space. Each unit
  symbol may optionally be followed by a single number, specifying the power.
  For example \code{"m2 s-2"} is equivalent to \code{"(m^2)*(s^-2)"}.

  The third type of unit string format accepted is the special case of
  udunits time duration with a reference origin, for example \code{"hours
  since 1970-01-01 00:00:00"}. Note, that the handling of time and calendar
  operations via the udunits library is subtly different from the way R
  handles date and time operations. This functionality is mostly exported for
  users that work with udunits time data, e.g., with NetCDF files. Users are
  otherwise encouraged to use \code{R}'s date and time functionality provided
  by \code{Date} and \code{POSIXt} classes.
}

\section{Expressions}{


  In \code{as_units()}, each of the symbols in the unit expression is treated
  individually, such that each symbol must be recognized by the udunits
  database (checked by \code{ud_is_parseable()}, \emph{or} be a custom,
  user-defined unit symbol that was defined either by
  \code{install_symbolic_unit()} or \code{install_conversion_constant()}. To
  see which symbols and names are currently recognized by the udunits
  database, see \code{udunits_symbols()}.
}

\examples{
s = Sys.time()
d  = s - (s+1)
as_units(d)
# The easiest way to assign units to a numeric vector is like this: 
x <- y <- 1:4
units(x) <- "m/s"  # meters / second

# Alternatively, the easiest pipe-friendly way to set units:
if(requireNamespace("magrittr", quietly = TRUE)) {
  library(magrittr)
  y \%>\% set_units(m/s)
} 

# these are different ways of creating the same unit:
# meters per second squared, i.e, acceleration
x1 <- make_units(m/s^2)
x2 <- as_units(quote(m/s^2))
x2 <- as_units("m/s^2")
x3 <- as_units("m s-2") # in product power form, i.e., implicit exponents = T
x4 <- set_units(1,  m/s^2) # by default, mode = "symbols"
x5 <- set_units(1, "m/s^2",   mode = "standard")
x6 <- set_units(1, x1,        mode = "standard")
x7 <- set_units(1, units(x1), mode = "standard")
x8 <- as_units("m") / as_units("s")^2

all_identical <- function(...) {
  l <- list(...)
  for(i in seq_along(l)[-1])
    if(!identical(l[[1]], l[[i]]))
      return(FALSE)
  TRUE
}
all_identical(x1, x2, x3, x4, x5, x6, x7, x8)

# Note, direct usage of these unit creation functions is typically not 
# necessary, since coercion is automatically done via as_units(). Again, 
# these are all equivalent ways to generate the same result.

x1 <- x2 <- x3 <- x4 <- x5 <- x6 <- x7 <- x8 <- 1:4
units(x1) <- "m/s^2"
units(x2) <- "m s-2"
units(x3) <- quote(m/s^2)
units(x4) <- make_units(m/s^2)
units(x5) <- as_units(quote(m/s^2))
x6 <- set_units(x6, m/s^2)
x7 <- set_units(x7, "m/s^2", mode = "standard")
x8 <- set_units(x8, units(x1), mode = "standard")

all_identical(x1, x2, x3, x4, x5, x6, x7, x8)


# Both unit names or symbols can be used. By default, unit names are
# automatically converted to unit symbols.
make_units(degree_C)
make_units(kilogram)
make_units(ohm)
# Note, if the printing of non-ascii characters is garbled, then you may
# need to specify the encoding on your system manually like this:
# ud_set_encoding("latin1")
# not all unit names get converted to symbols under different encodings

## Arithmetic operations and units
# conversion between unit objects that were defined as symbols and names will
# work correctly, although unit simplification in printing may not always occur.
x <- 500 * make_units(micrograms/liter)
y <- set_units(200, ug/l)
x + y
x * y # numeric result is correct, but units not simplified completely

# note, plural form of unit name accepted too ('liters' vs 'liter'), and
# denominator simplification can be performed correctly
x * set_units(5, liters)

# unit conversion works too
set_units(x, grams/gallon)

## Creating custom, user defined units
# For example, a microbiologist might work with counts of bacterial cells
# make_units(cells/ml) # by default, throws an ERROR
# First define the unit, then the newly defined unit is accepted.
install_symbolic_unit("cells")
make_units(cells/ml) 

# Note, install_symbolic_unit() does not add any support for unit
# conversion, or arithmetic operations that require unit conversion. See
# ?install_conversion_constant for defining relationships between user 
# defined units.

## set_units()
# set_units is a pipe friendly version of `units<-`. 
if(requireNamespace("magrittr", quietly = TRUE)) {
  library(magrittr)
  1:5 \%>\% set_units(N/m^2)
  # first sets to m, then converts to km
  1:5 \%>\% set_units(m) \%>\% set_units(km)
}

# set_units has two modes of operation. By default, it operates with 
# bare symbols to define the units.
set_units(1:5, m/s)

# use `mode = "standard"` to use the value of supplied argument, rather than
# the bare symbols of the expression. In this mode, set_units() can be
# thought of as a simple alias for `units<-` that is pipe friendly.
set_units(1:5, "m/s", mode = "standard")
set_units(1:5, make_units(m/s), mode = "standard")

# the mode of set_units() can be controlled via a global option
# units_options(set_units_mode = "standard")

# To remove units use
units(x) <- NULL
# or
set_units(x, NULL)
# or
drop_units(y)
}
\seealso{
\code{\link{valid_udunits}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{seq.units}
\alias{seq.units}
\title{seq method for units objects}
\usage{
\method{seq}{units}(from, to, by = ((to - from)/(length.out - 1)),
  length.out = NULL, along.with = NULL, ...)
}
\arguments{
\item{from}{see \link[base]{seq}}

\item{to}{see \link[base]{seq}}

\item{by}{see \link[base]{seq}}

\item{length.out}{see \link[base]{seq}}

\item{along.with}{see \link[base]{seq}}

\item{...}{see \link[base]{seq}}
}
\description{
seq method for units objects
}
\details{
arguments with units are converted to have units of the first argument (which is either \code{from} or \code{to})
}
\examples{
seq(to = set_units(10, m), by = set_units(1, m), length.out = 5)
seq(set_units(10, m), by = set_units(1, m), length.out = 5)
seq(set_units(10, m), set_units(19, m))
seq(set_units(10, m), set_units(.1, km), set_units(10000, mm))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/valid_udunits.R
\name{valid_udunits}
\alias{valid_udunits}
\alias{valid_udunits_prefixes}
\title{Get information about valid units}
\usage{
valid_udunits(quiet = FALSE)

valid_udunits_prefixes(quiet = FALSE)
}
\arguments{
\item{quiet}{logical, defaults \code{TRUE} to give a message about the location of
the udunits database being read.}
}
\value{
a data frame with columns \code{symbol} , \code{symbol_aliases} ,
  \code{name_singular} , \code{name_singular_aliases} , \code{name_plural} ,
  or \code{name_plural_aliases} , \code{def} , \code{definition} ,
  \code{comment} , \code{dimensionless} and \code{source_xml}
}
\description{
These functions require the \pkg{xml2} package, and return data frames with
complete information about pre-defined units from UDUNITS2. Inspect this data
frames to determine what inputs are accepted by \code{as_units} (and the
other functions it powers: \code{as_units}, \code{set_units} , \code{units<-}).
}
\details{
Any entry listed under \code{symbol} , \code{symbol_aliases} , \code{
name_singular} , \code{name_singular_aliases} , \code{name_plural} , or
\code{name_plural_aliases} is valid. Additionally, any entry under
\code{symbol} or \code{symbol_aliases} may can also contain a valid prefix,
as specified by \code{valid_udunits_prefixes()} .

Note, this is primarily intended for interactive use, the exact format of the
returned data frames may change in the future.
}
\examples{
if (requireNamespace("xml2", quietly = TRUE)) {
  valid_udunits()
  valid_udunits_prefixes()
  if(interactive())
    View(valid_udunits())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arith.R
\name{matrixOps.units}
\alias{matrixOps.units}
\title{S3 matrixOps Group Generic Functions for units objects}
\usage{
\method{matrixOps}{units}(x, y)
}
\arguments{
\item{x}{object of class \code{units},
or something that can be coerced to it by \code{as_units(x)}}

\item{y}{object of class \code{units},
or something that can be coerced to it by \code{as_units(y)}}
}
\value{
object of class \code{units}
}
\description{
matrixOps functions for units objects.
}
\examples{
a = set_units(1:5, m)
a \%*\% a
a \%*\% t(a)
a \%*\% 1:5
1:5 \%*\% a
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{hist.units}
\alias{hist.units}
\title{histogram for unit objects}
\usage{
\method{hist}{units}(x, xlab = NULL, main = paste("Histogram of", xname),
  ...)
}
\arguments{
\item{x}{object of class units, for which we want to plot the histogram}

\item{xlab}{character; x axis label}

\item{main}{character; title of histogram}

\item{...}{parameters passed on to \link{hist.default}}
}
\description{
histogram for unit objects
}
\examples{
units_options(parse = FALSE) # otherwise we break on the funny symbol!
u = set_units(rnorm(100), degree_C)
hist(u)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conversion.R, R/mixed.R
\name{drop_units}
\alias{drop_units}
\alias{drop_units.units}
\alias{drop_units.data.frame}
\alias{drop_units.mixed_units}
\title{Drop Units}
\usage{
drop_units(x)

\method{drop_units}{units}(x)

\method{drop_units}{data.frame}(x)

\method{drop_units}{mixed_units}(x)
}
\arguments{
\item{x}{an object with units metadata.}
}
\value{
the numeric without any units attributes, while preserving other
attributes like dimensions or other classes.
}
\description{
Drop units attribute and class.
}
\details{
Equivalent to \code{units(x) <- NULL}, or the pipe-friendly version
\code{set_units(x, NULL)}, but \code{drop_units} will fail if the object has
no units metadata. Use the alternatives if you want this operation to succeed
regardless of the object type.

A \code{data.frame} method is also provided, which checks every column and
drops units if any.
}
\examples{
x <- 1
y <- set_units(x, m/s)

# this succeeds
drop_units(y)
set_units(y, NULL)
set_units(x, NULL)

\dontrun{
# this fails
drop_units(x)
}

df <- data.frame(x=x, y=y)
df
drop_units(df)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{boxplot.units}
\alias{boxplot.units}
\title{boxplot for unit objects}
\usage{
\method{boxplot}{units}(x, ..., horizontal = FALSE)
}
\arguments{
\item{x}{object of class units, for which we want to plot the boxplot}

\item{...}{parameters passed on to \link{boxplot.default}}

\item{horizontal}{logical indicating if the boxplots should be horizontal;
default FALSE means vertical boxes.}
}
\description{
boxplot for unit objects
}
\examples{
units_options(parse = FALSE) # otherwise we break on the funny symbol!
u = set_units(rnorm(100), degree_C)
boxplot(u)
}
